#pragma once

#include "codegen/logic.h"

#include "platform.h"
#include "defs.h"
#include "json.h"

#include <vector>

#define REGISTER_LOGIC(classname, idname) classname s_##classname(idname)

namespace engine
{
    class Scene;

    class GameState
    {
    public:

        class MemberIterator : public std::iterator
            <std::input_iterator_tag,
            GameState,
            GameState,
            const GameState*,
            GameState&>
        {
        public:
            explicit MemberIterator(GameState* ptr) { m_ptr = ptr; }
            MemberIterator& operator++() { m_ptr = (GameState*)((intptr_t)m_ptr + sizeof(GameState)); return *this; }
            bool operator==(MemberIterator other) const { return other.m_ptr == m_ptr; }
            bool operator!=(MemberIterator other) const { return other.m_ptr != m_ptr; }
            reference operator*() const { return *m_ptr; };
        protected:
            GameState* m_ptr = nullptr;
        };

        MemberIterator begin() const { return MemberIterator((GameState*)m_data.m_data); }
        MemberIterator end() const { return MemberIterator((GameState*)((intptr_t)m_data.m_data + sizeof(GameState) * m_count)); }

        GameState& operator[](size_t i)
        {
            ASSERT(m_type == e_gameStateType::array);
            return ((GameState*)m_data.m_data)[i];
        }

        GameState& operator[](const std::string& name)
        {
            ID id = hash(name);

            ASSERT(m_type == e_gameStateType::unset || m_type == e_gameStateType::object);

            m_type = e_gameStateType::object;

            // search for member
            for(size_t i = 0; i < m_count; ++i)
            {
                GameState& s = ((GameState*)m_data.m_data)[i];
                if (s.m_id == id)
                {
                    return s;
                }
            }

            // add new member if not found
            if (m_count == 0)
            {
                m_count = 1;
                m_data.m_size = sizeof(GameState);
                m_data.m_data = memAlloc(m_data.m_size);
            }
            else
            {
                m_data.m_size = sizeof(GameState) * (++m_count);
                m_data.m_data = memRealloc(m_data.m_data, m_data.m_size);
            }

            // zero new member
            GameState& s = ((GameState*)m_data.m_data)[m_count - 1];
            s = {};
            s.m_id = id;
            s.m_name = (char*)memAlloc(name.size() + 1);
            strcpy_s(s.m_name, name.size() + 1, name.c_str());
            s.m_name[name.size()] = '\0';
            return s;
        }

        template<typename T>
        void setUints(const T& value)
        {
            set<T, u64>(value, e_gameStateValueType::uint);
        }

        template<typename T>
        void setInts(const T& value)
        {
            set<T, int>(value, e_gameStateValueType::sint);
        }

        template<typename T>
        void setFloats(const T& value)
        {
            set<T, float>(value, e_gameStateValueType::sfloat);
        }

        void setString(const std::string& value)
        {
            m_type = e_gameStateType::value;
            m_valueType = e_gameStateValueType::string;
            m_count = 1;

            if (m_data.m_data)
            {
                memFree(m_data.m_data);
            }
            m_data.m_data = memAlloc(value.size() + 1);
            strcpy_s((char*)m_data.m_data, value.size() + 1, value.c_str());
            ((char*)m_data.m_data)[value.size()] = '\0';
        }

        template<typename T>
        T value() const
        {
            ASSERT(m_type == e_gameStateType::value || m_type == e_gameStateType::array);

            if (m_type == e_gameStateType::array)
            {
                GameState* children = (GameState*)m_data.m_data;

                T value = {};
                size_t size = children[0].m_data.m_size;
                size_t iterations = sizeof(T) / size;
                void* dst = &value;
                for (size_t i = 0; i < iterations && i < m_count; ++i)
                {
                    memcpy(dst, children[i].m_data.m_data, size);
                    dst = (void*)((intptr_t)dst + size);
                }
                return value;
            }
            else
            {
                ASSERT(sizeof(T) >= m_data.m_size);
                if (sizeof(T) == m_data.m_size)
                {
                    return *(T*)m_data.m_data;
                }
                else if (sizeof(T) > m_data.m_size)
                {
                    // if T is larger than data splat the stored data across return value
                    T value = {};
                    size_t iterations = sizeof(T) / m_data.m_size;
                    void* dst = &value;
                    for (size_t i = 0; i < iterations; ++i)
                    {
                        memcpy(dst, m_data.m_data, m_data.m_size);
                        dst = (void*)((intptr_t)dst + m_data.m_size);
                    }
                    return value;
                }
            }

            T defaultReturn = {};
            return defaultReturn;
        }

        std::string string() const 
        {
            ASSERT(m_type == e_gameStateType::value && m_valueType == e_gameStateValueType::string);
            return std::string((char*)m_data.m_data);
        }

        void push(const GameState& state)
        {
            ASSERT(m_type == e_gameStateType::unset || m_type == e_gameStateType::array);
            
            m_type = e_gameStateType::array;

            if (m_count == 0)
            {
                m_data.m_size = sizeof(GameState);
                m_data.m_data = memAlloc(m_data.m_size);
            }
            else
            {
                m_data.m_size = sizeof(GameState) * (m_count + 1);
                m_data.m_data = memRealloc(m_data.m_data, m_data.m_size);
            }
            ((GameState*)m_data.m_data)[m_count++] = state;
        }

        void clear()
        {
            if (m_type == e_gameStateType::unset)
                return;

            ASSERT(m_type == e_gameStateType::array || m_type == e_gameStateType::object);

            if (m_count == 0)
                return;

            GameState* state = (GameState*)m_data.m_data;
            for (size_t i = 0; i < m_count; ++i)
            {
                state[i].erase();
            }
            memFree(m_data.m_data);
            m_count = 0;
        }


        void erase(size_t i)
        {
            ASSERT(m_type == e_gameStateType::array);
            sharedErase(i);
        }

        void erase(const std::string& name)
        {
            ASSERT(m_type == e_gameStateType::object);

            ID id = hash(name);

            GameState* members = (GameState*)m_data.m_data;
            for (size_t i = 0; i < m_count;  ++i)
            {
                if (members[i].m_id == id)
                {
                    sharedErase(i);
                    break;
                }
            }
        }

        bool contains(const std::string& name) const
        {
            if (m_type != e_gameStateType::object)
                return false;

            ID id = hash(name);

            for (size_t i = 0; i < m_count; ++i)
            {
                if (((GameState*)m_data.m_data)[i].m_id == id) 
                {
                    return true;
                }
            }
            return false;
        }

        size_t              count() const { return m_count; }
        std::string         name() const { return (m_name != nullptr) ? std::string(m_name) : ""; }
        GameStateType       type() const { return m_type; }
        GameStateValueType  valueType()  const { return m_valueType; }

    protected:

        template<typename T, typename D>
        void set(const T& value, GameStateValueType valueType)
        {
            size_t count = sizeof(T) / sizeof(D);
            if (count > 1)
            {
                m_type = e_gameStateType::array;
                bool initialise = (m_data.m_data == nullptr);

                D* data = (D*)&value;
                for (size_t i = 0; i < count; ++i)
                {
                    if (initialise)
                    {
                        GameState s = {};
                        s.set<D, D>(data[i], valueType);
                        push(s);
                    }
                    else
                    {
                        GameState& s = ((GameState*)m_data.m_data)[i];
                        s.set<D, D>(data[i], valueType);
                    }
                }
            }
            else
            {
                m_type = e_gameStateType::value;
                m_valueType = valueType;
                if (m_data.m_size == 0)
                {
                    m_data.m_size = sizeof(T);
                    m_data.m_data = memAlloc(sizeof(T));
                }
                *((T*)m_data.m_data) = value;
            }
        }

        void erase()
        {
            if (m_count > 0)
            {
                GameState* children = (GameState*)m_data.m_data;
                for (size_t i = 0; i < m_count; ++i)
                {
                    children[i].erase();
                }
            }
            memFree(m_data.m_data);

            if (m_name)
                memFree(m_name);
        }

        void sharedErase(size_t i)
        {
            GameState* s = (GameState*)m_data.m_data;
            s[i].erase();

            void* old = m_data.m_data;

            if (m_count > 1)
            {
                intptr_t before = (intptr_t)m_data.m_data;
                intptr_t target = before + sizeof(GameState) * i;
                intptr_t after = target + sizeof(GameState);
                intptr_t end = (intptr_t)m_data.m_data + m_count * sizeof(GameState);

                m_data.m_data = memAlloc(sizeof(GameState) * (m_count - 1));

                if (before < target)
                {
                    memcpy(m_data.m_data, (void*)before, sizeof(GameState) * i);
                }
                
                if (after < end)
                {
                    memcpy((void*)((intptr_t)m_data.m_data + (sizeof(GameState) * i)), (void*)after, sizeof(GameState) * (m_count - (i + 1)));
                }
            }

            memFree(old);
            m_count--;
        }

        GameStateType       m_type = e_gameStateType::unset;
        GameStateValueType  m_valueType = e_gameStateValueType::uint;
        size_t              m_count = 0;
        ID                  m_id = 0;
        char*               m_name = nullptr;
        Mem                 m_data;
    };

    namespace e_arithmeticType
    {
        enum arithmeticType
        {
            add,
            subtract,
            multiply,
            divide
        };
    }
    typedef e_arithmeticType::arithmeticType ArithmeticType;

    template<typename T>
    struct TriggerVar
    {
        T               m_value = {};
        TriggerVar<T>*  m_left = nullptr;
        TriggerVar<T>*  m_right = nullptr;
        std::string     m_stateName = "";
        ID              m_entityID = 0;
        ID              m_entityPropertyID = 0;
        size_t          m_arrayIndex = -1;
        TriggerVarType  m_type = e_triggerVarType::raw;
        ArithmeticType  m_arithmeticType = e_arithmeticType::add;

        TriggerVar()
        {
            m_value = {};
        }

        TriggerVar(T val)
        {
            m_value = val;
        }

        ~TriggerVar()
        {
            if (m_left)
            {
                memDelete(m_left);
                m_left = nullptr;
            }

            if (m_right)
            {
                memDelete(m_right);
                m_right = nullptr;
            }
        }

        template<typename D>
        void    set(const rapidjson::Value& j, const char* key, D arrayDefault);
        
        T       get();
    };

    class TriggerCondition
    {
    public:
        TriggerCondition() {};
        TriggerCondition(const char* name);
        
        virtual TriggerCondition*   create(const rapidjson::Value& j, Scene* s) = 0;
        virtual void                processInput(std::vector<utils::InputEvent>& events) {};
        virtual bool                update(float dt) = 0;
        virtual void                normalise(u32 width, u32 height, bool inverse) {};

        ID                  m_id = 0;
        TriggerVar<size_t>  m_target = -1;
        bool                m_entityLinked = false;
        bool                m_not = false;

    protected:
    };

    class TriggerAction
    {
    public:
        TriggerAction() {};
        TriggerAction(const char* name);

        virtual TriggerAction*  create(const rapidjson::Value& j, Scene* s) = 0;
        virtual void            reset() = 0;
        virtual bool            update(float dt) = 0;
        virtual void            normalise(u32 width, u32 height, bool inverse) {};

        ID                  m_id = 0;
        int                 m_priority = 0;
        TriggerVar<size_t>  m_target = -1;
        bool                m_entityLinked = false;
        bool                m_firing = false;

    };

    namespace e_triggerLogicFireType
    {
        enum triggerLogicFireType
        {
            cantFire,
            fireOnce,
            canFire
        };
    }
    typedef e_triggerLogicFireType::triggerLogicFireType TriggerLogicFireType;

    struct TriggerLogic
    {
        TriggerLogicFireType            m_fire = e_triggerLogicFireType::canFire;
        bool                            m_simultaneousActions = true;
        size_t                          m_actionIndex = (size_t)-1;
        std::vector<TriggerCondition*>  m_conditions;
        std::vector<TriggerAction*>     m_actions;
        std::vector<TriggerAction*>     m_elseActions;
    };

    void registerLogic();
}
