#pragma once

#include "gfx_defs.h"
#include "gfx_utils.h"
#include "defs.h"
#include "json.h"

#include <unordered_map>
#include <set>
#include <functional>

#define BIND_FUNCTION(type, func) std::bind(&type::func, this, std::placeholders::_1)

#define BUFFER_SLOT_GENERIC_VIEW 0
#define BUFFER_SLOT_VIEW_UNIFORMS 1
#define BUFFER_SLOT_CAMERA_UNIFORMS 2
#define BUFFER_SLOT_ENTITY_UNIFORMS 3
#define BUFFER_SLOT_ENTITY_MATERIAL 4

#define MAX_MRT_TARGETS 4
#define MAX_PIPELINE_VARIATIONS 4

namespace engine
{
    struct View;
    typedef std::function<void(View*)> ViewRenderFunc;
    typedef std::function<void(std::vector<View*>&)> ViewSetupFunc;

    struct GenericViewUniforms
    {
        float m_dimensions[4];
    };

    struct UniformBuffer
    {
        gfx::Buffer m_buffer = nullptr;
        size_t      m_size = 0;
    };

    struct TextureSamplers
    {
        ID              m_textureID[4] = { 0, 0, 0, 0 };
        ID              m_samplerID[4] = { 0, 0, 0, 0 };
        gfx::Texture    m_texture[4] = { nullptr, nullptr, nullptr, nullptr };
        gfx::Sampler    m_sampler[4] = { nullptr, nullptr, nullptr, nullptr };
    };

    struct TextureTarget
    {
        gfx::Texture    m_texture = nullptr;
        ID              m_textureID = 0;
        gfx::Clear      m_clear = {};
        gfx::LoadOp     m_loadOp = gfx::e_loadOpType::load;
    };

    struct PipelineVariations
    {
        ID m_variations[MAX_PIPELINE_VARIATIONS] = { 0 };
    };

    namespace e_resolveType
    {
        enum resolveType
        {
            none,
            standalone,
            after
        };
    }
    typedef e_resolveType::resolveType ResolveType;

    struct View
    {
        std::string                 m_name = "";
        ID                          m_idName = 0;
        ID                          m_linkID = 0;
        ID                          m_pipeline = 0;
        ID                          m_pipelineVariation = 0;
        gfx::RenderPass             m_pass = nullptr;
        gfx::Semaphore              m_sync;
        gfx::Buffer                 m_genericViewUniformBuffer = nullptr;
        gfx::Buffer                 m_uniformBuffer = nullptr;
        Mem                         m_uniformBufferMem;
        u64                         m_components = 0;
        TextureSamplers             m_textures;
        gfx::Viewport               m_viewport;
        TextureTarget               m_targetTextures[MAX_MRT_TARGETS];
        TextureTarget               m_depthTarget;
        size_t                      m_numTargets = 0;
        size_t                      m_numClears = 0;
        ResolveType                 m_resolve = e_resolveType::none;
        std::set<View*>             m_dependancies;
        ViewRenderFunc              m_renderFunction = nullptr;
    };

    struct PartialPipeline
    {
        gfx::Shader                 m_shaders[6];
        u64                         m_shaderStageFlags = 0;
        gfx::VertexInputDesc        m_vertexInput;
        gfx::InputAssemblyDesc      m_input;
        gfx::RasterDesc             m_raster;
        gfx::MultisampleDesc        m_multisample;
        std::vector<gfx::BlendDesc> m_blends;
        gfx::DepthStencilDesc       m_depthStencil;
        gfx::DynamicDesc            m_dynamic;
    };

    namespace e_renderTargetUsage
    {
        enum renderTargetUsage
        {
            unused,
            read,
            write,
            resolve
        };
    }
    typedef e_renderTargetUsage::renderTargetUsage RenderTargetUsage;

    struct RenderTarget
    {
        gfx::Texture            m_texture[2];
        RenderTargetUsage       m_usage = e_renderTargetUsage::unused;
        RenderTargetUsage       m_viewUsage = e_renderTargetUsage::unused;
        bool                    m_outputRes = false;
        bool                    m_variableSamples = false;
    };

    class Renderer
    {
    public:
        void                    setup();
        void                    shutdown();
        void                    update(float dt);
        void                    render();

        void                    loadConfig(const char* filename);
        View                    setupView(const View& v);
        void                    createViews();
        void                    registerRenderFunc(ViewRenderFunc func, const std::string& name);
        void                    registerSetupFunc(ViewSetupFunc func, const std::string& name);
        void                    unregisterRenderFunc(const std::string& name);
        void                    unregisterSetupFunc(const std::string& name);
        void                    resize(u32 width, u32 height);

        View*                   getView(ID id, ID linkID);
        void                    getViews(ID id, ID linkID, std::vector<View*>& views);
        std::vector<View*>      getViews() { return m_views; }

        gfx::Texture            getRenderTarget(ID id, RenderTargetUsage usage);
        gfx::RenderPipeline     getPipelineForPass(ID id, gfx::RenderPass pass, ID variation = 0);
        gfx::Sampler            getSampler(ID id);

        u32                     getMSAASamples() { return m_msaaSamples; }
        void                    setMSAASamples(u32 samples);

    protected:
        std::unordered_map<ID, PartialPipeline>             m_partialPipelines;
        std::unordered_map<ID, gfx::RenderPipeline>         m_pipelines;
        std::unordered_map<ID, RenderTarget>                m_renderTargets;
        std::unordered_map<ID, ViewRenderFunc>              m_renderFuncs;
        std::unordered_map<ID, ViewSetupFunc>               m_setupFuncs;
        std::unordered_map<ID, gfx::Sampler>                m_samplers;
        std::vector<View*>                                  m_views;
        rapidjson::Document                                 m_cfg;
        u32                                                 m_msaaSamples = 1;
    };
}
