#pragma once

#include "defs.h"
#include "gfx_defs.h"
#include "gfx_utils.h"
#include "audio.h"
#include "platform.h"
#include "font.h"

#include <unordered_map>

namespace engine
{
    namespace e_resourceLoadFlags
    {
        enum resourceLoadFlags
        {
            none = 0,
            async = 1 << 0
        };
    }
    typedef u32 ResourceLoadFlags;

    namespace e_resourceFlags
    {
        enum resourceFlags
        {
            none = 0,
            asyncLoading = 1 << 0,
            asyncLoaded = 1 << 1,
            loaded = 1 << 2,
            failed = 1 << 3
        };
    }
    typedef u32 ResourceFlags;

    class Resource
    {
    public:
        void*           m_internal = nullptr;
        void*           m_asyncData = nullptr;
        void*           m_desc = nullptr;
        ID              m_id = 0;
        ResourceFlags   m_flags = 0;
        int             m_refCount = 0;
    };

    template<typename T>
    class ResourceRef
    {
    public:
        void takeRef(Resource* res)
        {
            if(m_resource)
                m_resource->m_refCount--;

            m_resource = res;

            if(m_resource)
                m_resource->m_refCount++;
        }

        void operator = (Resource* res)
        {
            takeRef(res);
        }

        void operator=(const ResourceRef& other)
        {
            takeRef(other.m_resource);
        }

        ResourceRef()
        {
            m_resource = nullptr;
        }

        ResourceRef(const ResourceRef& other)
        {
            takeRef(other.m_resource);
        }

        ~ResourceRef()
        {
            if (m_resource)
            {
                m_resource->m_refCount--;
                ASSERT(m_resource->m_refCount > 0); // need to call unload on a ref to safely unload resources
            }
        }

        Resource*   getResource() { return m_resource; }
        T           getTypedResource() { if (std::is_pointer<T>::value) return *(T*)&m_resource->m_internal; else return *((T*)(m_resource->m_internal)); }
        void*       getDesc() { return m_resource->m_desc; }

        Resource*   m_resource = nullptr;
    };

    class ResourceManager
    {
    public:
        template<typename D>
        ID getIDFromDesc(const D& desc);

        template<typename T, typename D>
        ResourceRef<T> load(const D& desc, ResourceLoadFlags flags)
        {
            ResourceRef<T> ref;

            ID id = getIDFromDesc(desc);
            if (m_resources.find(id) == m_resources.end())
            {
                Resource* newResource = memNew<Resource>();
                newResource->m_desc = memNew<D>();
                newResource->m_refCount = 0;
                newResource->m_id = id;
                *((D*)newResource->m_desc) = desc;
                m_resources[id] = newResource;
            }

            Resource* res = m_resources.at(id);

            // do the 'async' section of the loading - can happen on main thread if async hasn't been started elsewhere
            if (!(res->m_flags & e_resourceFlags::asyncLoaded) && !(res->m_flags & e_resourceFlags::asyncLoading))
            {
                res->m_flags |= e_resourceFlags::asyncLoading;
                if (loadAsync<D>(desc, res))
                    res->m_flags |= e_resourceFlags::asyncLoaded;
                else
                    res->m_flags |= e_resourceFlags::failed;
                res->m_flags &= ~e_resourceFlags::asyncLoading;
            }

            // if loading on main thread wait for async load to complete
            if (!(res->m_flags & e_resourceFlags::loaded) && !(flags & e_resourceLoadFlags::async))
            {
                // wait for async load to complete
                while (res->m_flags & e_resourceFlags::asyncLoading)
                    utils::sleepMs(1);

                ASSERT(!(res->m_flags & e_resourceFlags::failed));  // check async load hasn't failed

                // do the main thread portion of the load
                if (load<D>(desc, res))
                {
                    res->m_flags |= e_resourceFlags::loaded;
                }
                else
                {
                    ASSERT(0); // resource failed main load
                }

                // destroy the async load data
            }

            ref = res;
            return ref;
        }

        template<typename T, typename D>
        void load(ResourceRef<T>& ref, ResourceLoadFlags flags)
        {
            if (!ref.getDesc())
                return;

            load<D>(*(D*)ref.getDesc(), flags);
        }

        template<typename T>
        void unload(ResourceRef<T>& ref)
        {
            Resource* res = ref.getResource();
            if(res)
            {
                ref = nullptr;
                if(res->m_refCount < 1)
                {
                    unload<T>(res);
                    m_resources.erase(res->m_id);
                }
            }
        }


        static ResourceManager* get();

    protected:

        template<typename D>
        bool load(const D& desc, Resource* resource);

        template<typename D>
        bool loadAsync(const D& desc, Resource* resource);

        template<typename T>
        void unload(Resource* resource);

        std::unordered_map<ID, Resource*> m_resources;
    };

    struct TextureLoadDesc
    {
        std::string m_filename = "";
        u32         m_mips = 1;
    };
    typedef ResourceRef<gfx::Texture> TextureRef;

    struct SoundLoadDesc
    {
        std::string                 m_filename = "";
        audio::AudioStreamingType   m_streamingType = audio::e_audioStreamingType::nonStreaming;
        audio::AudioLoopType        m_loopType = audio::e_audioLoopType::none;
        float                       m_startTime = 0.0f;
        float                       m_loopStartTime = 0.0f;
        float                       m_loopEndTime = -1.0f;
    };
    typedef ResourceRef<audio::Sound> SoundRef;

    typedef ResourceRef<gfx::Font*> FontRef;
}
