#pragma once

#include "world.h"
#include "renderer.h"
#include "logic.h"
#include "animations.h"

#include <atomic>

namespace engine
{
    namespace e_layerFlags
    {
        enum layerFlags
        {
            none = 0,
            normalised = 1 << 0
        };
    }
    typedef u64 LayerFlags;

    struct Layer
    {
        LayerFlags          m_flags = e_layerFlags::none;
        size_t              m_cameraEntity = (size_t)-1;
        ID                  m_cameraID = 0;
        u32                 m_width = 0;
        u32                 m_height = 0;
        ID                  m_blendID = 0;
        bool                m_deferredRender = false;
        std::vector<View>   m_preBlendPostProcessViews;
        std::vector<View>   m_postBlendPostProcessViews;
        glm::vec4           m_ambientLight = glm::vec4(0.0f);
        gfx::Buffer         m_ambientLightUniformBuffer = nullptr;
    };

    namespace e_sceneFlags
    {
        enum sceneFlags
        {
            none = 0,
            loading = 1 << 0,
            loaded = 1 << 1,
            activated = 1 << 2,
            transition = 1 << 3
        };
    }
    typedef u64 SceneFlags;

    class Scene
    {
    public:
        virtual void    load(const char* fp);
        virtual void    activate();
        virtual void    shutdown();
        virtual void    processInput(std::vector<utils::InputEvent>& events);
        virtual void    update(float dt);
        virtual void    render(View* view);
        virtual void    renderLayer(View* view);
        virtual void    renderLayerPostProcess(View* view);
        virtual void    renderLayerDeferred(View* view);
        virtual void    setupLayerViews(std::vector<View*>& views);
        virtual void    registerViewFunctions();
        virtual void    resize(u32 width, u32 height);
        virtual void    normalise(bool inverse);
        virtual void    loadEntity(size_t e, const rapidjson::Value& entity);

        void            updateLogic(float dt);
        void            updateTransforms(float dt);
        void            updateText(float dt);
        void            updateRendering(float dt);

        bool            isLoaded() const;
        bool            isActive() const;
        bool            shouldTransition() const;

        ID                                      m_id = 0;

        World*                                  m_world = nullptr;
        Renderer*                               m_renderer = nullptr;
        AnimationManager*                       m_animManager = nullptr;

        ModelRef                                m_ndcQuadRef;

        Scene*                                  m_next = nullptr;
        std::atomic<u64>                        m_flags = { 0 };

        std::vector<TriggerLogic>               m_triggers;
        std::vector<std::vector<TriggerLogic>>  m_entityTriggers;
        std::vector<TriggerAction*>             m_currentActions;

        std::vector<ID>                         m_layerOrder;
        std::unordered_map<ID, Layer>           m_layers;
    };

    rapidjson::Value mergeEntityAndPrefab(const rapidjson::Value& entity, const rapidjson::Value& prefab, rapidjson::Document::AllocatorType& allocator);
    rapidjson::Value loadPrefabs(const rapidjson::Value& prefabJson, rapidjson::Document::AllocatorType& allocator);
    Layer            loadLayer(const rapidjson::Value& layerJson);
}
