#pragma once

#include "codegen/world.h"

#include "maths_utils.h"
#include "renderer.h"
#include "resources.h"
#include "model.h"

typedef u64 ULID;

#define COMPONENT_ARRAY(type, name) ComponentArray<type> m_##name = ComponentArray<type>(hash(std::string(#name)))

namespace engine
{
    struct UniformData
    {
        glm::mat4 m_world;
        glm::vec4 m_userData0;
        glm::vec4 m_userData1;
    };

    struct OrthoCamera
    {
        bool    m_matchOutput = true;
        bool    m_invertY = false;
        float   m_left = 0.0f;
        float   m_right = 1.0f;
        float   m_bottom = 0.0f;
        float   m_top = 1.0f;
    };

    struct PerspectiveCamera
    {
        float m_fov = 90.0f;
        float m_aspect = 1.778f;
    };

    struct Camera
    {
        CameraType              m_type = e_cameraType::orthographic;
        OrthoCamera             m_ortho;
        PerspectiveCamera       m_perspective;
        glm::mat4               m_view = glm::mat4(1.0f);
        glm::mat4               m_proj = glm::mat4(1.0f);
        glm::mat4               m_matrix = glm::mat4(1.0f); 
        gfx::Buffer             m_matrixBuffer = nullptr;
        float                   m_near = 0.1f;
        float                   m_far = 1000.0f;
        maths::Frustum          m_cullingFrustum = {};
    };

    struct LightUniformData
    {
        glm::mat4 m_world;
        glm::mat4 m_view;
        glm::mat4 m_proj;
        glm::vec4 m_diffuse;
        glm::vec4 m_user;
    };

    struct Light
    {
        LightType               m_type = e_lightType::directional;
        LightUniformData        m_renderCpuData = {};
        gfx::Texture            m_shadowMap = nullptr;
        gfx::Buffer             m_renderBuffer = nullptr;
        gfx::Buffer             m_shadowMapBuffer = nullptr;
        maths::Frustum   m_cullingFrustum = {};
    };

    struct Text
    {
        std::string m_text = "";
    };

    class GenericComponentArray
    {
        friend class WorldBase;
    public:
        void grow(size_t size);
        void destroy();

        void set(size_t i, void* src, size_t offset = 0, size_t size = (size_t)-1)
        {
            intptr_t dst = (intptr_t)m_data + i * m_elementSize + offset;
            memcpy((void*)dst, src, (size == (size_t)-1) ? m_elementSize : size);
        }

        void get(size_t i, void* dst, size_t offset = 0, size_t size = (size_t)-1)
        {
            intptr_t src = (intptr_t)m_data + i * m_elementSize + offset;
            memcpy(dst, (void*)src, (size == (size_t)-1) ? m_elementSize : size);
        }

        void* data()
        {
            return m_data;
        }

        size_t elementSize()
        {
            return m_elementSize;
        }

    protected:
        void*   m_data = nullptr;
        size_t  m_capacity = 0;
        size_t  m_elementSize = 0;
        ID      m_alias = 0;
    };

    template<typename T>
    class ComponentArray : public GenericComponentArray
    {
    public:
        ComponentArray(ID alias)
        {
            m_elementSize = sizeof(T);
            m_alias = alias;
        }

        T& operator[](size_t i)
        {
            return ((T*)m_data)[i];
        }
    };

    class WorldBase
    {
    protected:
        size_t                  m_size = 0;
        size_t                  m_capacity = 0;
        GenericComponentArray*  m_last = nullptr;

    public:

        COMPONENT_ARRAY(u64,    components);
        COMPONENT_ARRAY(u64,    entityFlags);
        COMPONENT_ARRAY(u64,    inheritedEntityFlags);
        COMPONENT_ARRAY(size_t, parent);

        GenericComponentArray& getArray(ID alias)
        {
            // loop over all component arrays to find
            GenericComponentArray* arr = &m_components;
            for (;;)
            {
                if (arr->m_alias == alias)
                {
                    break;
                }
                else if (arr == m_last)
                {
                    arr = nullptr;
                    break;
                }
                arr = (GenericComponentArray*)(((intptr_t)arr) + sizeof(GenericComponentArray));
            }
            ASSERT(arr != nullptr); // alias not present!
            return *arr;
        }

        void            grow(size_t size);
        virtual size_t  newEntity();
        virtual void    destroyEntity(size_t e) = 0;
        size_t          capacity() { return m_capacity; }
        size_t          size() { return m_size; }
        void            clear();
        virtual u64     enumNameToComponentFlagValue(const std::string& name) = 0;
        virtual u64     enumNameToEntityFlagValue(const std::string& name) = 0;
    };

    class World : public WorldBase
    {
    public:
        COMPONENT_ARRAY(ULID,               ulid);
        COMPONENT_ARRAY(ID,                 layer);
        COMPONENT_ARRAY(u64,                cullingMask);
        COMPONENT_ARRAY(ID,                 triggerID);
        COMPONENT_ARRAY(float,              time);
        COMPONENT_ARRAY(glm::vec3,          pos);
        COMPONENT_ARRAY(glm::vec3,          rot);
        COMPONENT_ARRAY(glm::vec3,          scale);
        COMPONENT_ARRAY(glm::mat4,          localMatrix);
        COMPONENT_ARRAY(glm::mat4,          worldMatrix);
        COMPONENT_ARRAY(maths::AABB,        localAabb);
        COMPONENT_ARRAY(maths::OBB,         worldObb);
        COMPONENT_ARRAY(glm::vec3,          forward);
        COMPONENT_ARRAY(glm::vec3,          right);
        COMPONENT_ARRAY(ID,                 pipeline);
        COMPONENT_ARRAY(PipelineVariations, pipelineVariations);
        COMPONENT_ARRAY(gfx::Geometry*,     geometry);
        COMPONENT_ARRAY(UniformData,        uniformData);
        COMPONENT_ARRAY(UniformBuffer*,     uniforms);
        COMPONENT_ARRAY(UniformBuffer*,     material);
        COMPONENT_ARRAY(TextureSamplers,    textures);
        COMPONENT_ARRAY(TextureRef,         textureRef);
        COMPONENT_ARRAY(ModelRef,           modelRef);
        COMPONENT_ARRAY(size_t,             meshIndex);
        COMPONENT_ARRAY(size_t,             materialIndex);
        COMPONENT_ARRAY(Camera*,            camera);
        COMPONENT_ARRAY(SoundRef,           soundRef);
        COMPONENT_ARRAY(FontRef,            fontRef);
        COMPONENT_ARRAY(Text*,              text);
        COMPONENT_ARRAY(Light*,             light);

        World()
        {
            m_last = &m_light;
        }

        virtual void    destroyEntity(size_t e) override;
        virtual u64     enumNameToComponentFlagValue(const std::string& name) override;
        virtual u64     enumNameToEntityFlagValue(const std::string& name) override;
    };
}
