#include "animations.h"

namespace
{
}

namespace engine
{
	void AnimWorld::destroyEntity(size_t e)
	{
		m_entityFlags[e] = 0;
		m_components[e] = 0;
	}

	u64 AnimWorld::enumNameToComponentFlagValue(const std::string& name)
	{
		return e_animCmp::nameToValue(name);
	}

	u64 AnimWorld::enumNameToEntityFlagValue(const std::string& name)
	{
		return e_animEntityFlags::nameToValue(name) | e_baseEntityFlags::nameToValue(name);
	}

	void AnimationManager::setup(World* w)
	{
		m_world = w;

		// grow to an initial fixed size
		m_animWorld.grow(128);
	}

	void AnimationManager::shutdown()
	{
		m_animWorld.clear();
	}

	void wrapAnimation(AnimWorld& w, size_t i, float& t)
	{
		switch (w.m_loopType[i])
		{
		case e_animLoopType::clamp:
		{
			t =  glm::clamp(t, 0.0f, 1.0f);
		}
		break;
		case e_animLoopType::wrap:
		{
			if (t < 0.0f)
				t = 1.0f + t;
			else if (t > 1.0f)
				t = t - 1.0f;
		}
		break;
		case e_animLoopType::mirror:
		{
			if (t > 1.0f)
			{
				t = 2.0f - t;
				if (w.m_entityFlags[i] & e_animEntityFlags::mirroring)
					w.m_entityFlags[i] &= ~e_animEntityFlags::mirroring;
				else
					w.m_entityFlags[i] |= e_animEntityFlags::mirroring;
			}
			else if (t < 0.0f)
			{
				t = -t;
				if (w.m_entityFlags[i] & e_animEntityFlags::mirroring)
					w.m_entityFlags[i] &= ~e_animEntityFlags::mirroring;
				else
					w.m_entityFlags[i] |= e_animEntityFlags::mirroring;
			}
		}
		break;
		}
	}

	void AnimationManager::update(float dt)
	{
		AnimWorld& w = m_animWorld;
		
		for (size_t i = 0, n = w.size(); i < n; ++i)
		{
			if (!(w.m_components[i] & e_animCmp::animation))
				continue;

			if (w.m_entityFlags[i] & e_baseEntityFlags::disabled)
				continue;

			float t = w.m_time[i] * w.m_duration[i];
			switch (w.m_driverType[i])
			{
			case e_animDriverType::time:
			{
				// drive by delta time
				if (w.m_entityFlags[i] & e_animEntityFlags::mirroring)
					t = t - (dt * w.m_speed[i]);
				else
					t = t + (dt * w.m_speed[i]);
			}
			break;
			case e_animDriverType::entity:
			{
				// drive by property of an entity in the world the animation is applied to
			}
			break;
			case e_animDriverType::anim:
			{
				// drive by the value of another animation
			}
			break;
			}

			t = t / w.m_duration[i];
			wrapAnimation(w, i, t);
			w.m_time[i] = t;

			// interpolate the data based on current time
			size_t a = i + 1;
			for (size_t j = i + 2; j < n; ++j)
			{
				if (!(w.m_components[j] & e_animCmp::point))
					break;

				if (w.m_time[j] >= w.m_time[i])
					break;

				// clamp to final point
				if (j + 1 >= n || !(w.m_components[j + 1] & e_animCmp::point))
					break;

				a = j;
			}

			size_t b = a + 1;
			float interpT = glm::clamp((w.m_time[i] - w.m_time[a]) / (w.m_time[b] - w.m_time[a]), 0.0f, 1.0f);

			switch (w.m_interpType[i])
			{
			case e_animInterpType::linear:
			{
				for (size_t j = 0; j < w.m_dataCount[i]; ++j)
				{
					float da = w.m_data[a].m_floats[j];
					float db = w.m_data[b].m_floats[j];

					w.m_data[i].m_floats[j] = da + (db - da) * interpT;
				}
			}
			break;
			}

			// apply the data to the target world
			for (size_t j = 0; j < w.m_mapping[i].m_numGroups; ++j)
			{
				GenericComponentArray& arr = m_world->getArray(w.m_mapping[i].m_targetAlias[j]);
				u8* byteArr = (u8*)arr.data();
				size_t elemSize = arr.elementSize();
				float* floats = (float*)&(byteArr[elemSize * w.m_targetIndex[i]]);
				for (size_t k = 0; k < w.m_dataCount[i]; ++k)
				{
					if (w.m_mapping[i].m_groupIndex[k] == j)
					{
						floats[w.m_mapping[i].m_targetOffset[k]] = w.m_data[i].m_floats[k];
					}
				}
			}
		}
	}

	void AnimationManager::addAnimation(size_t targetIndex, ID animID, const rapidjson::Value& anim)
	{
		AnimWorld& w = m_animWorld;

		// create parent animation entity
		size_t p = w.newEntity();
		w.m_components[p] |= e_animCmp::animation;
		w.m_targetIndex[p] = targetIndex;
		w.m_id[p] = animID;

		// load per animation properties
		if (!getJsonValueOrDefault(anim, "start_activated", false))
		{
			w.m_entityFlags[p] |= e_baseEntityFlags::disabled;
		}

		w.m_duration[p] = getJsonValueOrDefault(anim, "duration", 1.0f);
		w.m_speed[p] = getJsonValueOrDefault(anim, "speed", 1.0f);
		w.m_driverType[p] = e_animDriverType::nameToValue(getJsonValueOrDefault<const char*>(anim, "driver_type", "time"));
		w.m_loopType[p] = e_animLoopType::nameToValue(getJsonValueOrDefault<const char*>(anim, "loop_type", "clamp"));
		w.m_interpType[p] = e_animInterpType::nameToValue(getJsonValueOrDefault<const char*>(anim, "interp_type", "linear"));

		ASSERT(anim.HasMember("mappings"));
		AnimMapping& mapping = w.m_mapping[p];
		for (auto mit = anim["mappings"].MemberBegin(); mit != anim["mappings"].MemberEnd(); ++mit)
		{
			const rapidjson::Value& j = mit->value;

			ID targetAlias = hash<std::string>(getJsonValueOrDefault<const char*>(j, "target", ""));

			// group data targeting the same alias together
			size_t group = (size_t)-1;
			for (size_t k = 0; k < w.m_dataCount[p]; ++k)
			{
				if (mapping.m_targetAlias[k] == targetAlias)
				{
					group = mapping.m_groupIndex[k];
					break;
				}
			}
			if (group == (size_t)-1)
				group = mapping.m_numGroups++;

			size_t i = w.m_dataCount[p]++;
			mapping.m_id[i] = hash<std::string>(mit->name.GetString());
			mapping.m_targetAlias[i] = targetAlias;
			mapping.m_targetOffset[i] = getJsonValueOrDefault<size_t>(j, "offset", 0);
			mapping.m_groupIndex[i] = group;
			mapping.m_groupAlias[group] = targetAlias;
		}
		
		// create the individual points
		ASSERT(anim.HasMember("data"));
		size_t numAnimPoints = anim["data"].Size();
		for (size_t n = 0; n < numAnimPoints; ++n)
		{
			const rapidjson::Value& j = anim["data"][n];

			size_t e = w.newEntity();
			w.m_components[e] |= e_animCmp::point;
			w.m_dataCount[e] = w.m_dataCount[p];
			w.m_id[e] = animID;

			w.m_time[e] = getJsonValueOrDefault(j, "time", 0.0f);

			for (auto mit = j.MemberBegin(); mit != j.MemberEnd(); ++mit)
			{
				ID id = hash<std::string>(mit->name.GetString());
				for (size_t i = 0; i < w.m_dataCount[e]; ++i)
				{
					if (w.m_mapping[p].m_id[i] != id)
						continue;

					w.m_data[e].m_floats[i] = getJsonValueOrDefault(j, mit->name.GetString(), 0.0f);
				}
			}
		}
	}
}