#include "engine.h"

#include "gfx.h"
#include "audio.h"
#include "imgui.h"

#include <mutex>


namespace
{
    namespace e_sceneThreadRequestType
    {
        enum sceneThreadRequestType
        {
            preload,
            load
        };
    }
    typedef e_sceneThreadRequestType::sceneThreadRequestType SceneThreadRequestType;

    ID getSceneID()
    {

    }
}

namespace engine
{
    Engine* Engine::m_self = nullptr;

    struct SceneThreadRequest
    {
        SceneThreadRequestType  m_type = e_sceneThreadRequestType::load;
        SceneDesc               m_desc;
    };

    void* loadThread(void* data);
        
    Scene* createGenericScene(void* data)
    {
        Scene* scene = memNew<Scene>();
        return scene;
    }

    Scene* Engine::createScene(const SceneDesc& desc)
    {
        ASSERT(m_sceneCreateFuncs.find(desc.m_funcID) != m_sceneCreateFuncs.end());
        SceneCreateFunc createFunc = m_sceneCreateFuncs.at(desc.m_funcID);
        return createFunc(nullptr);
    }

    void Engine::registerTriggerCondition(TriggerCondition* condition)
    {
        m_triggerConditions[condition->m_id] = condition;
    }

    void Engine::registerTriggerAction(TriggerAction* action)
    {
        m_triggerActions[action->m_id] = action;
    }

    class EngineInternal : public Engine
    {
    public:
        Renderer                        m_renderer;
        audio::AudioManager*            m_audioManager = nullptr;
        ModelRef                        m_ndcQuadRef;
        Scene*                          m_currentScene = nullptr;

        // load thread data
        Thread                          m_loadThread = nullptr;
        std::atomic_bool                m_closeLoadThread = false;
        RingBuffer<SceneThreadRequest>  m_requests;

        std::unordered_map<ID, Scene*>  m_scenes;
        std::recursive_mutex            m_scenesMutex;

        EngineInternal()
        {
            Engine::m_self = this;
            registerLogic();
        }

        void setup() override
        {
            m_audioManager = audio::AudioManager::get();
            m_audioManager->setup();
            
            m_loadThread = utils::createThread(loadThread, this, utils::e_threadFlags::detached);

            registerSceneCreateFunc(createGenericScene, "generic");

            auto ctx = gfx::getMainContext();

            m_renderer.setup();
            m_renderer.registerRenderFunc(BIND_FUNCTION(EngineInternal, renderFullscreen), "render_fullscreen");
            m_renderer.registerRenderFunc(BIND_FUNCTION(EngineInternal, renderClear), "render_clear");
            m_renderer.registerSetupFunc(BIND_FUNCTION(EngineInternal, setupResolve), "setup_resolve");
            m_renderer.loadConfig("data/cfg/renderer.json");
            m_renderer.createViews();

            // create ndc fullscreen quad for fullscreen draws
            ModelLoadDesc md = {};
            md.m_genData.m_size = sizeof(GeometryCpuData);
            md.m_genData.m_data = memNew<GeometryCpuData>();

            GeometryCpuData& ndcData = *(GeometryCpuData*)md.m_genData.m_data;
            ndcData = {};
            float vertexData[] =
            {
                -1.0f, -1.0f, 0.0f, 1.0f, 0.0f, 0.0f,
                 1.0f,  1.0f, 0.0f, 1.0f, 1.0f, 1.0f,
                 1.0f, -1.0f, 0.0f, 1.0f, 1.0f, 0.0f,
                -1.0f, -1.0f, 0.0f, 1.0f, 0.0f, 0.0f,
                -1.0f,  1.0f, 0.0f, 1.0f, 0.0f, 1.0f,
                 1.0f,  1.0f, 0.0f, 1.0f, 1.0f, 1.0f
            };
            ndcData.m_numVertices = 6;
            ndcData.m_vertexData.m_size = sizeof(vertexData);
            ndcData.m_vertexData.m_data = memAlloc(sizeof(vertexData));
            memcpy(ndcData.m_vertexData.m_data, vertexData, sizeof(vertexData));

            u32 indexData[] = { 0, 1, 2, 3, 4, 5 };
            ndcData.m_numIndices = 6;
            ndcData.m_indexData.m_size = sizeof(indexData);
            ndcData.m_indexData.m_data = memAlloc(sizeof(indexData));
            memcpy(ndcData.m_indexData.m_data, indexData, sizeof(indexData));

            m_ndcQuadRef = ResourceManager::get()->load<ModelData, ModelLoadDesc>(md, e_resourceLoadFlags::none);
        }

        void shutdown() override
        {

        }

        void renderFullscreen(View* view)
        {
            auto ctx = gfx::getMainContext();

            ModelData md = m_ndcQuadRef.getTypedResource();
            gfx::Geometry geo = md.m_meshes[0];
            ctx->bindBuffer(geo.m_vertexBuffer, 0, gfx::e_bindFlags::vertexBuffer);
            ctx->drawIndexed(geo.m_numIndices, geo.m_indexBuffer, gfx::e_indexType::uint32);
        }

        void renderClear(View* view)
        {

        }

        void setupResolve(std::vector<View*>& views)
        {
            // get the sample count of the engine output texture and see if we need to resolve it
            auto ctx = gfx::getMainContext();

            gfx::Texture tex = m_renderer.getRenderTarget(hash<std::string>("output"), e_renderTargetUsage::resolve);
            gfx::TextureDesc td = ctx->getTextureDesc(tex);
            if (td.m_samples > 1)
            {
                View* view = memNew<View>();
                view->m_name = "resolve_output";
                *view = m_renderer.setupView(*view);
                views.push_back(view);
            }
        }

        void processInput(std::vector<utils::InputEvent>& events) override
        {
            if (m_currentScene && m_currentScene->isActive())
            {
                m_currentScene->processInput(events);
            }
        }

        void update(float dt) override
        {
            m_gameState["frame_rate"].setFloats(dt);

            m_audioManager->update(dt);

            if (m_currentScene)
            {
                if (m_currentScene->isLoaded())
                {
                    if (m_currentScene->shouldTransition() && m_currentScene->m_next)
                    {
                        // move to next scene
                        Scene* prev = m_currentScene;
                        m_currentScene = prev->m_next;
                        m_currentScene->m_renderer = &m_renderer;

                        // unload previous scene
                        prev->shutdown();

                        // remove froms scenes tracking
                        {
                            std::lock_guard<std::recursive_mutex> lock(m_scenesMutex);
                            m_scenes.erase(prev->m_id);
                        }

                        memDelete(prev);
                    }

                    if (!m_currentScene->isActive())
                    {
                        m_currentScene->activate();
                        m_currentScene->registerViewFunctions();
                        m_renderer.createViews();
                    }

                    m_currentScene->update(dt);

                }
                else
                {
                    // display loading screen
                }
            }

            m_renderer.update(dt);
        }

        void render() override
        {
            m_renderer.render();
        }

        void resize(u32 width, u32 height) override
        {
            m_renderer.resize(width, height);
            if (m_currentScene)
            {
                m_currentScene->resize(width, height);
            }
        }

        void preloadScene(const SceneDesc& desc) override
        {
            SceneThreadRequest request = {};
            request.m_type = e_sceneThreadRequestType::preload;
            request.m_desc = desc;
        }

        void loadScene(const SceneDesc& desc) override
        {
            SceneThreadRequest request = {};
            request.m_type = e_sceneThreadRequestType::load;
            request.m_desc = desc;
            m_requests.put(request);
        }

        Scene* getCurrentScene() override
        {
            return m_currentScene;
        }

        Scene* preloadSceneInternal(const SceneDesc& desc)
        {
            Scene* scene = nullptr;
            {
                std::lock_guard<std::recursive_mutex>lg(m_scenesMutex);

                ID sceneID = hash(desc.m_filename);
                sceneID = hashCombine(sceneID, desc.m_funcID);
                if (m_scenes.find(sceneID) == m_scenes.end())
                {
                    scene = createScene(desc);
                    scene->m_renderer = &m_renderer;
                    scene->m_id = sceneID;
                    m_scenes[sceneID] = scene;
                }
                else
                {
                    scene = m_scenes.at(sceneID);
                }
            }

            if (!scene->isLoaded())
            {
                scene->load(desc.m_filename.c_str());
            }

            return scene;
        }

        Renderer& getRenderer() override
        {
            return m_renderer;
        }
    };
    
    Engine* createEngine(const EngineDesc& desc)
    {
        return memNew<EngineInternal>();
    }

    void* loadThread(void* data)
    {
        EngineInternal* engine = (EngineInternal*)data;
        while (!engine->m_closeLoadThread.load())
        {
            SceneThreadRequest request;
            while (engine->m_requests.get(request))
            {
                switch (request.m_type)
                {
                case e_sceneThreadRequestType::preload:
                {
                    engine->preloadSceneInternal(request.m_desc);
                }
                break;
                case e_sceneThreadRequestType::load:
                {
                    Scene* scene = engine->preloadSceneInternal(request.m_desc);
                    Scene* current = engine->getCurrentScene();
                    if (!current)
                    {
                        engine->m_currentScene = scene;
                    }
                    else
                    {
                        current->m_next = scene;
                        current->m_flags |= e_sceneFlags::transition;
                    }
                }
                break;
                }
            }

            utils::sleepMs(1);
        }

        return nullptr;
    }
}
