#include "engine_app.h"

#include <chrono>
#include <imgui.h>

namespace
{

}

namespace engine
{
    void plotDependanciesRecursive(const std::vector<View*>& views, std::map<View*, u32>& depths, View* v, u32 depth)
    {
        depths[v] = std::max(depths[v], depth);

        for (auto& view : views)
        {
            if (view->m_dependancies.find(v) != view->m_dependancies.end())
            {
                plotDependanciesRecursive(views, depths, view, depth + 1);
            }
        }
    }

    void EngineApp::setup()
    {
        EngineDesc desc = {};
        m_engine = createEngine(desc);
        m_engine->setup();
        DevUI::setup();
    }

    void EngineApp::shutdown()
    {
        DevUI::shutdown();
    }

    void EngineApp::processInput(std::vector<utils::InputEvent>& events)
    {
        m_engine->processInput(events);

        // dev ui open/close
        for (auto& e : events)
        {
            if (e.m_type == utils::e_inputEventType::key
                && e.m_key.m_key == utils::e_key::ret
                && e.m_key.m_state == utils::e_keyState::up)
            {
                m_devUIOpen = !m_devUIOpen;
            }
        }
    }

    void EngineApp::update()
    {
        // frame timing
        static auto start = std::chrono::high_resolution_clock::now();
        static auto prevTime = 0.0f;
        auto now = std::chrono::high_resolution_clock::now();
        auto currentTime = std::chrono::duration_cast<std::chrono::milliseconds>(now - start).count() / 1000.0f;
        float dt = currentTime - prevTime;

        u32 width = utils::getWindowWidth();
        u32 height = utils::getWindowHeight();

        DevUI::update(dt, width, height);
        m_engine->update(dt);

        // engine app dev ui update
        if (m_devUIOpen && ImGui::Begin("Dev UI", &m_devUIOpen, ImGuiWindowFlags_AlwaysAutoResize))
        {
            if (ImGui::BeginTabBar("tabs"))
            {
                if (ImGui::BeginTabItem("settings"))
                {
                    if (ImGui::Checkbox("Fullscreen", &m_fullscreen))
                    {
                        utils::setWindowFullscreen(m_fullscreen);
                    }

                    Renderer& renderer = m_engine->getRenderer();
                    u32 currentSamples = renderer.getMSAASamples();
                    u32 samplesOptions[] = { 1, 2, 4, 8 };

                    if (ImGui::BeginCombo("MSAA Samples", std::to_string(currentSamples).c_str()))
                    {
                        for (size_t i = 0; i < ARRAY_SIZE(samplesOptions); ++i)
                        {
                            bool selected = currentSamples == samplesOptions[i];
                            if (ImGui::Selectable(std::to_string(samplesOptions[i]).c_str(), &selected))
                            {
                                renderer.setMSAASamples(samplesOptions[i]);
                            }
                        }
                        ImGui::EndCombo();
                    }
                    ImGui::EndTabItem();
                }
                if (ImGui::BeginTabItem("renderer"))
                {
                    Renderer& renderer = m_engine->getRenderer();

                    std::vector<View*> views = renderer.getViews();
                    std::map<View*, u32> dependancyDepths;

                    for (intptr_t i = views.size() - 1; i >= 0; --i)
                    {
                        if (dependancyDepths.find(views[i]) != dependancyDepths.end())
                            continue;

                        plotDependanciesRecursive(views, dependancyDepths, views[i], 0);
                    }
                    ImGui::EndTabItem();
                }
                ImGui::EndTabBar();
            }
            ImGui::End();
        }

        prevTime = currentTime;
    }

    void EngineApp::render()
    {
        m_engine->render();
        DevUI::render();
    }

    void EngineApp::resize(u32 width, u32 height)
    {
        if (m_engine)
        {
            m_engine->resize(width, height);
        }
    }
}