#include "logic.h"
#include "engine.h"

#include "pystring.h"

namespace
{
    const char* k_triggerArithmeticSymbols[] =
    {
        "+",
        "-",
        "*",
        "/"
    };

    const char* k_comparisonSymbols[] =
    {
        "==",
        "!=",
        "<",
        "<=",
        ">",
        ">="
    };

    namespace e_comparisonType
    {
        enum comparisonType
        {
            equal,
            notEqual,
            lessThan,
            lessThanOrEqual,
            greaterThan,
            greaterThanOrEqual
        };
    }
    typedef e_comparisonType::comparisonType ComparisonType;
}

namespace engine
{
    std::vector<TriggerAction*> s_actions;
    std::vector<TriggerCondition*> s_conditions;

    TriggerCondition* Engine::createTriggerCondition(ID id, const rapidjson::Value& j, Scene* s, size_t index)
    {
        ASSERT(m_triggerConditions.find(id) != m_triggerConditions.end());
        TriggerCondition* t = m_triggerConditions.at(id)->create(j, s);
        // set index to provided value if not overriden during creation
        if (t->m_target.m_value == (size_t)-1 && t->m_target.m_type == e_triggerVarType::raw)
        {
            if (j.HasMember("target"))
            {
                t->m_target.set<size_t>(j, "target", -1);
            }
            else if (index != (size_t)-1)
            {
                t->m_target.m_value = index;
            }
        }
        t->m_not = getJsonValueOrDefault(j, "not", false);
        return t;
    }

    TriggerAction* Engine::createTriggerAction(ID id, const rapidjson::Value& j, Scene* s, size_t index)
    {
        ASSERT(m_triggerActions.find(id) != m_triggerActions.end());
        TriggerAction* t = m_triggerActions.at(id)->create(j, s);
        t->m_priority = getJsonValueOrDefault<int>(j, "priority", 0);
        // set index to provided value if not overriden during creation
        if (t->m_target.m_value == (size_t)-1 && t->m_target.m_type == e_triggerVarType::raw)
        {
            if (j.HasMember("target"))
            {
                t->m_target.set<size_t>(j, "target", -1);
            }
            else if (index != (size_t)-1)
            {
                t->m_target.m_value = index;
            }
        }
        return t;
    }

    template<typename T>
    template<typename D>
    void TriggerVar<T>::set(const rapidjson::Value& j, const char* key, D defaultValue)
    {
        // if not an object just load the raw value
        if (!j.HasMember(key) || !j[key].IsObject())
        {
            m_type = e_triggerVarType::raw;
            size_t count = sizeof(T) / sizeof(D);
            if (count > 1)
            {
                getJsonArrayOrDefault<D>(j, key, (D*)&m_value, count, defaultValue);
            }
            else
            {
                m_value = (T)getJsonValueOrDefault<D>(j, key, defaultValue);
            }
            return;
        }

        const rapidjson::Value& var = j[key];

        m_type = e_triggerVarType::nameToValue(getJsonValueOrDefault<const char*>(var, "type", "raw"));

        switch (m_type)
        {
        case e_triggerVarType::raw:
        {
            size_t count = sizeof(T) / sizeof(D);

            if (count > 1)
            {
                getJsonArrayOrDefault<D>(var, "value", (D*)&m_value, count, defaultValue);
            }
            else
            {
                m_value = (T)getJsonValueOrDefault<D>(var, "value", defaultValue);
            }
        }
        break;
        case e_triggerVarType::entity:
        {
            std::string str = var["value"].GetString();
            int split = pystring::find(str, "/");
            if (split != -1)
            {
                std::string entityID = str.substr(0, split);
                std::string propertyID = str.substr(split + 1, (str.size() - 1) - split);
                m_entityID = hash(entityID);
                m_entityPropertyID = hash(propertyID);
            }
            else
            {
                m_entityID = hash(str);
            }
        }
        break;
        case e_triggerVarType::state:
        {
            m_stateName = var["value"].GetString();
        }
        break;
        case e_triggerVarType::arithmetic:
        {
            m_arithmeticType = enumNameToValue<ArithmeticType>(k_triggerArithmeticSymbols, ARRAY_SIZE(k_triggerArithmeticSymbols), getJsonValueOrDefault<const char*>(var, "operation", "+"));

            m_left = memNew<TriggerVar<T>>();
            m_left->set<D>(var, "left", defaultValue);

            m_right = memNew<TriggerVar<T>>();
            m_right->set<D>(var, "right", defaultValue);
        }
        break;
        }        
    }

    template<typename T>
    T TriggerVar<T>::get()
    {
        switch (m_type)
        {
        case e_triggerVarType::raw:
        {
            return m_value;
        }
        break;
        case e_triggerVarType::state:
        {
            // find the game state this variable is tied to
            Engine* e = Engine::get();
            GameState& s = e->getGameState();
            if (s.contains(m_stateName))
            {
                if (m_arrayIndex != (size_t)-1)
                    return s[m_stateName][m_arrayIndex].value<T>();
                else
                    return s[m_stateName].value<T>();
            }
        }
        break;
        case e_triggerVarType::entity:
        {
            // find the entity this variable is tied to
            Engine* e = Engine::get();
            Scene* s = e->getCurrentScene();
            if (s)
            {
                World& w = *s->m_world;
                for (size_t i = 0, n = w.size(); i < n; ++i)
                {
                    if (w.m_ulid[i] != m_entityID)
                        continue;

                    // if no property was selected return the entity index
                    if (m_entityPropertyID == 0)
                        return (T)i;

                    ComponentArray<T>& arr = *(ComponentArray<T>*)&w.getArray(m_entityPropertyID);
                    return arr[i];
                }
            }
        }
        break;
        case e_triggerVarType::arithmetic:
        {
            switch (m_arithmeticType)
            {
            case e_arithmeticType::add:
            {
                return (T)(m_left->get() + m_right->get());
            }
            break;
            case e_arithmeticType::subtract:
            {
                return (T)(m_left->get() - m_right->get());
            }
            break;
            case e_arithmeticType::multiply:
            {
                return (T)(m_left->get() * m_right->get());
            }
            break;
            case e_arithmeticType::divide:
            {
                return (T)(m_left->get() / m_right->get());
            }
            break;
            default:
            {
                return m_value;
            }
            break;
            }
        }
        break;
        }

        return m_value;
    }

    TriggerAction::TriggerAction(const char* name)
    {
        m_id = hash<std::string>(name);
        s_actions.push_back(this);
    }

    TriggerCondition::TriggerCondition(const char* name)
    {
        m_id = hash<std::string>(name);
        s_conditions.push_back(this);
    }

    class TriggerConditionCheckEntityFlags : public TriggerCondition
    {
        using TriggerCondition::TriggerCondition;

        u64     m_flags = 0;

        TriggerCondition* create(const rapidjson::Value& j, Scene* s) override
        {
            TriggerConditionCheckEntityFlags* t = memNew<TriggerConditionCheckEntityFlags>();
            const rapidjson::Value& flagNames = j["flags"].GetArray();
            for (size_t i = 0; i < flagNames.Size(); ++i)
            {
                std::string flagName = flagNames[i].GetString();
                t->m_flags |= s->m_world->enumNameToEntityFlagValue(flagName);
            }
            return t;
        }

        bool update(float dt) override
        {
            Engine* e = Engine::get();
            Scene* s = e->getCurrentScene();
            auto& w = *s->m_world;

            size_t index = m_target.get();
            if (index == (size_t)-1)
            {
                for (size_t i = 0, n = w.size(); i < n; ++i)
                {
                    if (w.m_entityFlags[i] & m_flags)
                        return true;
                }
            }
            else
            {
                return w.m_entityFlags[index] & m_flags;
            }
            return false;
        }
    };
    REGISTER_LOGIC(TriggerConditionCheckEntityFlags, "check_entity_flags");
   
    class TriggerConditionCheckEntityComponents : public TriggerCondition
    {
        using TriggerCondition::TriggerCondition;

        u64     m_components = 0;

        TriggerCondition* create(const rapidjson::Value& j, Scene* s) override
        {
            TriggerConditionCheckEntityComponents* t = memNew<TriggerConditionCheckEntityComponents>();
            const rapidjson::Value& componentNames = j["components"].GetArray();
            for (size_t i = 0; i < componentNames.Size(); ++i)
            {
                std::string flagName = componentNames[i].GetString();
                t->m_components |= s->m_world->enumNameToComponentFlagValue(flagName);
            }
            return t;
        }

        bool update(float dt) override
        {
            Engine* e = Engine::get();
            Scene* s = e->getCurrentScene();
            auto& w = *s->m_world;

            size_t index = m_target.get();
            if (index == (size_t)-1)
            {
                for (size_t i = 0, n = w.size(); i < n; ++i)
                {
                    if (w.m_components[i] & m_components)
                        return true;
                }
            }
            else
            {
                return w.m_components[index] & m_components;
            }
            return false;
        }
    };
    REGISTER_LOGIC(TriggerConditionCheckEntityComponents, "check_entity_components");


    class TriggerConditionKeyState : public  TriggerCondition
    {
        using TriggerCondition::TriggerCondition;

        TriggerVar<utils::Key>      m_key = utils::e_key::esc;
        TriggerVar<utils::KeyState> m_state = utils::e_keyState::pressed;
        bool                        m_triggered = false;

        TriggerCondition* create(const rapidjson::Value& j, Scene* s) override
        {
            TriggerConditionKeyState* t = memNew<TriggerConditionKeyState>();
            
            std::unordered_map<std::string, utils::Key> keyMappings =
            {
                { "escape",    utils::e_key::esc, },
                { "return",    utils::e_key::ret, },
                { "tab",       utils::e_key::tab,},
                { "shift",     utils::e_key::shift, },
                { "control", utils::e_key::ctrl, },
                { "delete",    utils::e_key::del, },
                { "space",     utils::e_key::space, },
                { "up",        utils::e_key::up, },
                { "down",      utils::e_key::down, },
                { "left",      utils::e_key::left, },
                { "right",     utils::e_key::right, },
            };

            // setup some key mappings
            for (u8 i = 0x30; i < 0x3A; ++i)
            {
                std::string k = "";
                k += (char)('0' + (i - 0x30));
                keyMappings[k] = (utils::Key)(utils::e_key::zero + (i - 0x30));
            }

            for (u8 i = 0x41; i < 0x5B; ++i)
            {
                std::string k = "";
                k += (char)('a' + (i - 0x41));
                keyMappings[k] = (utils::Key)(utils::e_key::a + (i - 0x41));
            }

            std::string keyName = getJsonValueOrDefault<const char*>(j, "key", "escape");
            if (keyMappings.find(keyName) != keyMappings.end())
            {
                t->m_key = keyMappings.at(keyName);
            }
            
            t->m_state = (utils::KeyState)e_keyState::nameToValue(getJsonValueOrDefault<const char*>(j, "state", "pressed"));
            return t;
        }

        void processInput(std::vector<utils::InputEvent>& events) override
        {
            bool prevTriggered = m_triggered;
            m_triggered = false;

            utils::KeyState current = utils::KeyState::count;
            for (auto& e : events)
            {
                if (e.m_type != utils::e_inputEventType::key)
                    continue;

                if (e.m_key.m_key != m_key.get())
                    continue;

                current = e.m_key.m_state;
            }

            utils::KeyState state = m_state.get();
            switch (state)
            {
            case utils::e_keyState::down:
            {
                if (prevTriggered)
                {
                    if (current == utils::e_keyState::down
                        || current == utils::e_keyState::count)
                    {
                        m_triggered = true;
                    }
                }
                else if (current == utils::e_keyState::down)
                {
                    m_triggered = true;
                }
            }
            break;
            case utils::e_keyState::up:
            case utils::e_keyState::pressed:
            {
                m_triggered = (current == state);
            }
            break;
            default:
                break;
            }
        }

        bool update(float dt) override
        {
            return m_triggered;
        }
    };
    REGISTER_LOGIC(TriggerConditionKeyState, "key_state");
    
    class TriggerConditionComparisonInternal
    {
    public:
        ComparisonVarType   m_type = e_comparisonVarType::sfloat;
        ComparisonType      m_comparison = e_comparisonType::equal;

        virtual bool equals() = 0;
        virtual bool notEquals() = 0;
        virtual bool lessThan() = 0;
        virtual bool lessThanOrEquals() = 0;
        virtual bool greaterThan() = 0;
        virtual bool greaterThanOrEquals() = 0;
    };

    class TriggerConditionComparisonInt : public TriggerConditionComparisonInternal
    {
    public:
        TriggerVar<int> m_left;
        TriggerVar<int> m_right;

        bool equals() override
        {
            return m_left.get() == m_right.get();
        }

        bool notEquals() override
        {
            return m_left.get() != m_right.get();
        }

        bool lessThan() override
        {
            return m_left.get() < m_right.get();
        }

        bool lessThanOrEquals() override
        {
            return m_left.get() <= m_right.get();
        }

        bool greaterThan() override
        {
            return m_left.get() > m_right.get();
        }

        bool greaterThanOrEquals() override
        {
            return m_left.get() >= m_right.get();
        }
    };

    class TriggerConditionComparisonUint : public TriggerConditionComparisonInternal
    {
    public:
        TriggerVar<u64> m_left;
        TriggerVar<u64> m_right;

        bool equals() override
        {
            return m_left.get() == m_right.get();
        }

        bool notEquals() override
        {
            return m_left.get() != m_right.get();
        }

        bool lessThan() override
        {
            return m_left.get() < m_right.get();
        }

        bool lessThanOrEquals() override
        {
            return m_left.get() <= m_right.get();
        }

        bool greaterThan() override
        {
            return m_left.get() > m_right.get();
        }

        bool greaterThanOrEquals() override
        {
            return m_left.get() >= m_right.get();
        }
    };

    class TriggerConditionComparisonFloat : public TriggerConditionComparisonInternal
    {
    public:
        TriggerVar<float> m_left;
        TriggerVar<float> m_right;

        bool equals() override
        {
            return m_left.get() == m_right.get();
        }

        bool notEquals() override
        {
            return m_left.get() != m_right.get();
        }

        bool lessThan() override
        {
            return m_left.get() < m_right.get();
        }

        bool lessThanOrEquals() override
        {
            return m_left.get() <= m_right.get();
        }

        bool greaterThan() override
        {
            return m_left.get() > m_right.get();
        }

        bool greaterThanOrEquals() override
        {
            return m_left.get() >= m_right.get();
        }
    };

    class TriggerConditionComparison : public TriggerCondition
    {
        TriggerConditionComparisonInternal* m_internal = nullptr;

        using TriggerCondition::TriggerCondition;
    public:
        ~TriggerConditionComparison()
        {
            memDelete(m_internal);
        }

        TriggerCondition* create(const rapidjson::Value& j, Scene* s) override
        {
            TriggerConditionComparison* t = memNew<TriggerConditionComparison>();

            ComparisonVarType varType = e_comparisonVarType::nameToValue(getJsonValueOrDefault<const char*>(j, "var_type", "sfloat"));
            
            switch (varType)
            {
            case e_comparisonVarType::sint:
            {
                TriggerConditionComparisonInt* ti = memNew<TriggerConditionComparisonInt>();
                ti->m_left.set(j, "left", 0);
                ti->m_right.set(j, "right", 0);
                t->m_internal = ti;
            }
            break;
            case e_comparisonVarType::uint:
            {
                TriggerConditionComparisonUint* ti = memNew<TriggerConditionComparisonUint>();
                ti->m_left.set<uint64_t>(j, "left", 0);
                ti->m_right.set<uint64_t>(j, "right", 0);
                t->m_internal = ti;
            }
            break;
            case e_comparisonVarType::sfloat:
            {
                TriggerConditionComparisonFloat* ti = memNew<TriggerConditionComparisonFloat>();
                ti->m_left.set(j, "left", 0.0f);
                ti->m_right.set(j, "right", 0.0f);
                t->m_internal = ti;
            }
            break;
            case e_comparisonVarType::string:
            {
            }
            break;
            }

            t->m_internal->m_comparison = enumNameToValue<ComparisonType>(k_comparisonSymbols, ARRAY_SIZE(k_comparisonSymbols), getJsonValueOrDefault<const char*>(j, "operation", "=="));
            t->m_internal->m_type = varType;

            return t;
        }

        bool update(float dt) override
        {
            switch (m_internal->m_comparison)
            {
            case e_comparisonType::equal:
            {
                return m_internal->equals();
            }
            break;
            case e_comparisonType::notEqual:
            {
                return m_internal->notEquals();
            }
            break;
            case e_comparisonType::lessThan:
            {
                return m_internal->lessThan();
            }
            break;
            case e_comparisonType::lessThanOrEqual:
            {
                return m_internal->lessThanOrEquals();
            }
            break;
            case e_comparisonType::greaterThan:
            {
                return m_internal->greaterThan();
            }
            break;
            case e_comparisonType::greaterThanOrEqual:
            {
                return m_internal->greaterThanOrEquals();
            }
            break;
            }
            return true;
        }
    };
    REGISTER_LOGIC(TriggerConditionComparison, "comparison");

    class TriggerActionSetGameState : public TriggerAction
    {
        using TriggerAction::TriggerAction;

        std::string         m_name = "";
        TriggerVar<float>   m_value = 0.0f;
        
        TriggerAction* create(const rapidjson::Value& j, Scene* s) override
        {
            TriggerActionSetGameState* t = memNew<TriggerActionSetGameState>();
            t->m_name = getJsonValueOrDefault<const char*>(j, "state_name", "");
            t->m_value.set(j, "value", 0.0f);
            return t;
        }

        void reset() override
        {

        }

        bool update(float dt) override
        {
            Engine* e = Engine::get();
            GameState& s = e->getGameState();
            s[m_name].setFloats(m_value.get());
            return true;
        }
    };
    REGISTER_LOGIC(TriggerActionSetGameState, "set_game_state");

    class TriggerActionSetEntityFlags : public TriggerAction
    {
        using TriggerAction::TriggerAction;

        u64     m_flags = 0;
        bool    m_unset = false;

        TriggerAction* create(const rapidjson::Value& j, Scene* s) override
        {
            TriggerActionSetEntityFlags* t = memNew<TriggerActionSetEntityFlags>();
            size_t numFlags = j["flags"].Size();
            for (size_t i = 0; i < numFlags; ++i)
            {
                t->m_flags |= s->m_world->enumNameToEntityFlagValue(j["flags"][i].GetString());
            }
            t->m_unset = getJsonValueOrDefault(j, "unset", false);
            return t;
        }

        void reset() override
        {

        }

        bool update(float dt) override
        {
            Engine* e = Engine::get();
            Scene* s = e->getCurrentScene();
            auto& w = *s->m_world;
            size_t i = m_target.get();

            if (m_unset)
            {
                w.m_entityFlags[i] &= ~m_flags;
            }
            else
            {
                w.m_entityFlags[i] |= m_flags;
            }
            return true;
        }
    };
    REGISTER_LOGIC(TriggerActionSetEntityFlags, "set_entity_flags");

    class TriggerActionPlayEntitySound : public TriggerAction
    {
        using TriggerAction::TriggerAction;

        TriggerAction* create(const rapidjson::Value& j, Scene* s) override
        {
            TriggerActionPlayEntitySound* t = memNew<TriggerActionPlayEntitySound>();
            return t;
        }

        void reset() override
        {

        }

        bool update(float dt) override
        {
            Engine* e = Engine::get();
            Scene* s = e->getCurrentScene();
            auto& w = *s->m_world;
            size_t i = m_target.get();
            if (w.m_components[i] & e_cmp::soundRef)
            {
                audio::AudioManager* am = audio::AudioManager::get();
                am->playSound(w.m_soundRef[i].getTypedResource());
            }
            return true;
        }
    };
    REGISTER_LOGIC(TriggerActionPlayEntitySound, "play_entity_sound");


    class TriggerActionSetEntityTransform : public TriggerAction
    {
        using TriggerAction::TriggerAction;

        TriggerVar<glm::vec3>   m_pos = glm::vec3(0.0f);
        TriggerVar<glm::vec3>   m_rot = glm::vec3(0.0f);
        TriggerVar<glm::vec3>   m_scale = glm::vec3(1.0f);
        bool                    m_setPos = false;
        bool                    m_setRot = false;
        bool                    m_setScale = false;

        TriggerAction* create(const rapidjson::Value& j, Scene* s) override
        {
            TriggerActionSetEntityTransform* t = memNew<TriggerActionSetEntityTransform>();
            t->m_pos.set(j, "pos", 0.0f);
            t->m_rot.set(j, "rot", 0.0f);
            t->m_scale.set(j, "scale", 1.0f);

            // only override values that have actually been specified
            t->m_setPos = j.FindMember("pos") != j.MemberEnd();
            t->m_setRot = j.FindMember("rot") != j.MemberEnd();
            t->m_setScale = j.FindMember("scale") != j.MemberEnd();

            return t;
        }

        void reset() override
        {

        }

        bool update(float dt) override
        {
            Engine* e = Engine::get();
            Scene* s = e->getCurrentScene();
            auto& w = *s->m_world;
            size_t i = m_target.get();
            if(m_setPos)
                w.m_pos[i] = m_pos.get();
            if(m_setRot)
                w.m_rot[i] = m_rot.get();
            if(m_setScale)
                w.m_scale[i] = m_scale.get();
            return true;
        }
    };
    REGISTER_LOGIC(TriggerActionSetEntityTransform, "set_entity_transform");

    class TriggerActionSceneTransition : public TriggerAction
    {
        using TriggerAction::TriggerAction;

        SceneDesc m_sceneDesc;

        TriggerAction* create(const rapidjson::Value& j, Scene* s) override
        {
            TriggerActionSceneTransition* t = memNew<TriggerActionSceneTransition>();
            t->m_sceneDesc.m_funcID = hash<std::string>(getJsonValueOrDefault(j, "func", "generic"));
            t->m_sceneDesc.m_filename = getJsonValueOrDefault(j, "filename", "");
            return t;
        }

        void reset() override
        {

        }

        bool update(float dt) override
        {
            Engine* e = Engine::get();
            e->loadScene(m_sceneDesc);
            return true;
        }
    };
    REGISTER_LOGIC(TriggerActionSceneTransition, "scene_transition");

    class TriggerActionPlayAnimation : public TriggerAction
    {
        using TriggerAction::TriggerAction;

        TriggerVar<float> m_sectionBegin = 0.0f;
        TriggerVar<float> m_sectionEnd = 1.0f;
        ID                m_animTargetID = 0;
        ID                m_animID = 0;
    
        bool              m_started = false;
        float             m_start = 0.0f;
        float             m_end = 1.0f;

        TriggerAction* create(const rapidjson::Value& j, Scene* s) override
        {
            TriggerActionPlayAnimation* t = memNew<TriggerActionPlayAnimation>();
            t->m_sectionBegin.set(j, "start", 0.0f);
            t->m_sectionEnd.set(j, "end", 1.0f);
            t->m_animTargetID = hash<std::string>(getJsonValueOrDefault<const char*>(j, "anim_target_name", ""));
            t->m_animID = hash<std::string>(getJsonValueOrDefault<const char*>(j, "anim_name", ""));
            return t;
        }

        void reset() override
        {
            m_started = false;
            m_start = 0.0f;
            m_end = 1.0f;
        }

        bool update(float dt) override
        {
            // check progress of animation
            Engine* e = Engine::get();
            Scene* s = e->getCurrentScene();
            AnimationManager* am = s->m_animManager;
            
            World& ew = *s->m_world;
            AnimWorld& w = am->getWorld();

            size_t anim = -1;
            for (size_t i = 0, n = w.size(); i < n; ++i)
            {
                if (!(w.m_components[i] & e_animCmp::animation))
                    continue;

                if (m_animID != w.m_id[i])
                    continue;

                if (m_animTargetID != ew.m_ulid[w.m_targetIndex[i]])
                    continue;

                anim = i;
                break;
            }

            if (anim == (size_t)-1)
                return true;

            if (!m_started)
            {
                m_start = m_sectionBegin.get();
                m_end = m_sectionEnd.get();

                w.m_entityFlags[anim] &= ~e_baseEntityFlags::disabled;

                if (m_end < m_start)
                    w.m_speed[anim] = -abs(w.m_speed[anim]);
                else
                    w.m_speed[anim] = abs(w.m_speed[anim]);

                w.m_time[anim] = m_start;

                m_started = true;
            }

            float t = w.m_time[anim];
            // check current progress of the animation
            if ((m_end > m_start && t >= m_end) 
                || (m_end < m_start && t <= m_end))
            {
                    w.m_time[anim] = m_end;
                    w.m_entityFlags[anim] |= e_baseEntityFlags::disabled;
                    return true;
            }
            
            return false;
        }
    };
    REGISTER_LOGIC(TriggerActionPlayAnimation, "play_animation");



    void registerLogic()
    {
        Engine* e = Engine::get();
        for (auto& a : s_actions)
            e->registerTriggerAction(a);

        for (auto& c : s_conditions)
            e->registerTriggerCondition(c);
    }
}
