#include "model.h"

#include "platform.h"
#include "world.h"
#include "gfx.h"

#include <unordered_set>
#include <sstream>
#include <pystring.h>
#include <rapidjson/document.h>
#include <stb_image.h>
#include <tiny_obj_loader.h>
#include <glm/gtx/transform.hpp>

namespace
{
    const size_t k_numVertFloats = 9;
    struct membuf : std::stringbuf
    {
        membuf(char* begin, char* end)
        {
            this->setg(begin, begin, end);
        }
    };

    // implementation of tinyobj material reader that uses readFileToBuffer
    struct InternalMaterialReader : tinyobj::MaterialReader
    {
        InternalMaterialReader(const char* basedir)
        {
            m_baseDir = basedir;
        }

        virtual ~InternalMaterialReader() override {}

        virtual bool operator()(const std::string& matId,
            std::vector<tinyobj::material_t>* materials,
            std::map<std::string, int>* matMap, std::string* warn,
            std::string* err) override
        {
            std::string matpath = pystring::os::path::join(m_baseDir, matId);
            if (!utils::fileExists(matpath.c_str()))
                return false;

            void* filebuf;
            size_t filesize;
            utils::readFileToBuffer(matpath.c_str(), filebuf, filesize);
            membuf sbuf((char*)filebuf, (char*)filebuf + filesize);
            std::istream in(&sbuf);

            tinyobj::LoadMtl(matMap, materials, &in, warn, err);

            utils::freeFileBuffer(filebuf);

            return true;
        }

        std::string m_baseDir = "";
    };

    struct ModelIndices
    {
        u32     index;
        float   values[3];

        bool operator==(const ModelIndices& other) const
        {
            return memcmp(this->values, other.values, sizeof(values)) == 0;
        }
    };
}

namespace std
{
    template<> struct hash<ModelIndices>
    {
        size_t operator()(ModelIndices const& indices) const
        {
            return (hash<float>()(indices.values[0]) ^ (hash<float>()(indices.values[1]) << 1) ^ (hash<float>()(indices.values[2]) << 2));
        }
    };
}

namespace engine
{
    void generatePlaneVerts(const PlaneGenData& genData, std::vector<VertexTexturedNormal>& verts, std::vector<u32>& inds, glm::vec4& aabbMin, glm::vec4& aabbMax)
    {
        // generate a plane in the xy axis and then rotate the vertices to match the correct normal
        float step = 1.0f / genData.m_subdivisions;
        
        glm::vec4 position = glm::vec4(-0.5f, -0.5f, 0.0f, 1.0f);
        glm::vec2 texcoord = glm::vec2(0.0f, 0.0f);
        glm::vec3 normal = glm::vec3(genData.m_normal[0], genData.m_normal[1], genData.m_normal[2]);
        
        size_t firstVert = verts.size();
        for (float y = 0.0f; y < genData.m_subdivisions; ++y)
        {
            for (float x = 0.0f; x < genData.m_subdivisions; ++x)
            {
                u32 idx0 = (u32)verts.size();

                VertexTexturedNormal v0 = {};
                v0.m_position = position;
                v0.m_texcoord = glm::vec2(texcoord.x, texcoord.y + step);
                v0.m_normal = normal;
                verts.push_back(v0);

                VertexTexturedNormal v1 = {};
                v1.m_position = glm::vec4(position.x + step, position.y, 0.0f, 1.0f);
                v1.m_texcoord = glm::vec2(texcoord.x + step, texcoord.y + step);
                v1.m_normal = normal;
                verts.push_back(v1);

                VertexTexturedNormal v2 = {};
                v2.m_position = glm::vec4(position.x + step, position.y + step, 0.0f, 1.0f);
                v2.m_texcoord = glm::vec2(texcoord.x + step, texcoord.y);
                v2.m_normal = normal;
                verts.push_back(v2);

                VertexTexturedNormal v3 = {};
                v3.m_position = glm::vec4(position.x, position.y + step, 0.0f, 1.0f);
                v3.m_texcoord = glm::vec2(texcoord.x, texcoord.y);
                v3.m_normal = normal;
                verts.push_back(v3);

                inds.push_back(idx0);
                inds.push_back(idx0 + 1);
                inds.push_back(idx0 + 2);
                inds.push_back(idx0);
                inds.push_back(idx0 + 2);
                inds.push_back(idx0 + 3);

                position.x += step;
                texcoord.x += step;
            }

            position.x = -0.5f;
            texcoord.x = 0.0f;

            position.y += step;
            texcoord.y += step;
        }

        // create a rotation + translation matrix from the parameters and apply to all vertices
        glm::mat4 t = glm::translate(glm::vec3(genData.m_offset[0], genData.m_offset[1], genData.m_offset[2]));

        glm::vec3 a = glm::vec3(0.0f, 0.0f, -1.0f);
        glm::vec3 b = normal;
        glm::mat4 r = maths::createVectorAlignMatrix(a, b);

        glm::mat4 tr = t * r;

        for (size_t i = firstVert; i < verts.size(); ++i)
        {
            VertexTexturedNormal& v = verts[i];
            v.m_position = tr * v.m_position;

            for (int i = 0; i < 3; ++i)
            {
                if (v.m_position[i] < aabbMin[i])
                    aabbMin[i] = v.m_position[i];

                if (v.m_position[i] > aabbMax[i])
                    aabbMax[i] = v.m_position[i];
            }
        }
    }

    template<typename T>
    GeometryCpuData createGeometryCpuData(const std::vector<T>& verts, const std::vector<u32>& indices, glm::vec4& aabbMin, glm::vec4& aabbMax, size_t vertexFloatStepRate = 1)
    {
        GeometryCpuData cpuData;
        cpuData.m_numVertices = (u32)(verts.size() / vertexFloatStepRate);
        cpuData.m_numIndices = (u32)indices.size();
        cpuData.m_vertexData.m_data = new T[verts.size()];
        cpuData.m_vertexData.m_size = sizeof(T) * verts.size();
        memcpy(cpuData.m_vertexData.m_data, verts.data(), cpuData.m_vertexData.m_size);
        cpuData.m_indexData.m_data = new u32[indices.size()];
        cpuData.m_indexData.m_size = sizeof(u32) * indices.size();
        memcpy(cpuData.m_indexData.m_data, indices.data(), cpuData.m_indexData.m_size);
        memcpy(cpuData.m_minVert, &aabbMin.x, sizeof(float) * 4);
        memcpy(cpuData.m_maxVert, &aabbMax.x, sizeof(float) * 4);
        return cpuData;
    }

    ModelAsyncData generatePlaneGeometry(void* data)
    {
        auto ctx = gfx::getMainContext();

        PlaneGenData genData = *(PlaneGenData*)data;

        GeometryCpuData geometry = {};

        std::vector<VertexTexturedNormal> verts;
        std::vector<u32> inds;
        glm::vec4 minVert = { FLT_MAX, FLT_MAX, FLT_MAX, 0.0f };
        glm::vec4 maxVert = { -FLT_MAX, -FLT_MAX, -FLT_MAX, 0.0f };
        generatePlaneVerts(genData, verts, inds, minVert, maxVert);

        ModelAsyncData out = {};
        out.m_cpuBuffers.push_back(createGeometryCpuData<VertexTexturedNormal>(verts, inds, minVert, maxVert));
        return out;
    }

    ModelAsyncData generateCubeGeometry(void* data)
    {
        auto ctx = gfx::getMainContext();

        glm::vec3 normals[6] =
        {
            glm::vec3( 0.0f,  0.0f, -1.0f),	// front
            glm::vec3( 0.0f,  0.0f,  1.0f),	// back
            glm::vec3( 0.0f,  1.0f,  0.0f),	// top
            glm::vec3( 0.0f, -1.0f,  0.0f),	// bottom
            glm::vec3(-1.0f,  0.0f,  0.0f),	// left
            glm::vec3( 1.0f,  0.0f,  0.0f)	// right
        };

        glm::vec3 offsets[6] =
        {
            glm::vec3(0.0f, 0.0f, -0.5f),
            glm::vec3(0.0f, 0.0f, 0.5f),
            glm::vec3(0.0f, 0.5f, 0.0f),
            glm::vec3(0.0f, -0.5f, 0.0f),
            glm::vec3(-0.5f, 0.0f, 0.0f),
            glm::vec3(0.5f, 0.0f, 0.0f)
        };

        std::vector<VertexTexturedNormal> verts;
        std::vector<u32> inds;

        glm::vec4 minVert = { FLT_MAX, FLT_MAX, FLT_MAX, 0.0f };
        glm::vec4 maxVert = { -FLT_MAX, -FLT_MAX, -FLT_MAX, 0.0f };

        for (int i = 0; i < 6; ++i)
        {
            PlaneGenData gen = {};
            gen.m_subdivisions = 1.0f;
            gen.m_normal[0] = normals[i].x;
            gen.m_normal[1] = normals[i].y;
            gen.m_normal[2] = normals[i].z;
            gen.m_offset[0] = offsets[i].x;
            gen.m_offset[1] = offsets[i].y;
            gen.m_offset[2] = offsets[i].z;
            generatePlaneVerts(gen, verts, inds, minVert, maxVert);
        }

        ModelAsyncData out = {};
        out.m_cpuBuffers.push_back(createGeometryCpuData<VertexTexturedNormal>(verts, inds, minVert, maxVert));
        return out;
    }

    template<>
    bool ResourceManager::loadAsync<ModelLoadDesc>(const ModelLoadDesc& desc, Resource* resource)
    {
        // if using a gen function skip async loading
        if (desc.m_genFunc)
        {
            resource->m_internal = memNew<ModelData>();
            resource->m_asyncData = memNew<ModelAsyncData>();
            *(ModelAsyncData*)(resource->m_asyncData) = desc.m_genFunc(desc.m_genData.m_data);
            return true;
        }

        // quick work around to allow passing arbitrary vertex/index data
        if (desc.m_filename.empty())
        {   
            resource->m_internal = memNew<ModelData>();
            resource->m_asyncData = memNew<ModelAsyncData>();
            
            GeometryCpuData data = *(GeometryCpuData*)desc.m_genData.m_data;
            ((ModelAsyncData*)resource->m_asyncData)->m_cpuBuffers.push_back(data);
            return true;
        }

        if (!utils::fileExists(desc.m_filename.c_str()))
            return false;

        void* filebuf;
        size_t filesize;
        utils::readFileToBuffer(desc.m_filename.c_str(), filebuf, filesize);
        tinyobj::attrib_t attrib;
        std::vector<tinyobj::shape_t> shapes;
        std::vector<tinyobj::material_t> mats;
        std::string err, warn;
        std::string materialPath = pystring::os::path::dirname(desc.m_filename) + "/";
        InternalMaterialReader matReader(materialPath.c_str());

        membuf sbuf((char*)filebuf, (char*)filebuf + filesize);
        std::istream in(&sbuf);

        bool success = tinyobj::LoadObj(&attrib, &shapes, &mats, &err, &warn, &in, &matReader);
        utils::freeFileBuffer(filebuf);
        if (!success)
        {
            printf("Failed to load obj: %s\n", desc.m_filename.c_str());
            return false;
        }

        resource->m_internal = memNew<ModelData>();
        resource->m_asyncData = memNew<ModelAsyncData>();
        ModelData* md = (ModelData*)(resource->m_internal);
        ModelAsyncData* ad = (ModelAsyncData*)(resource->m_asyncData);

        // load materials
        ID defaultSamplerID = hash<std::string>("wrap_linear");
        for (auto& mat : mats)
        {
            MaterialLoadDesc matDesc = {};
            matDesc.m_params = mat;
            matDesc.m_defaultSamplerID = defaultSamplerID;
            ad->m_materials.push_back(load<MaterialData, MaterialLoadDesc>(matDesc, e_resourceLoadFlags::async));
        }

        // load meshes as cpu buffers
        auto& verts = attrib.vertices;
        auto& norms = attrib.normals;
        auto& texs = attrib.texcoords;
        auto& colors = attrib.colors;

        for (auto& shape : shapes)
        {
            std::vector<float> vertData;
            std::vector<u32> indexData;
            std::unordered_set<ModelIndices> uniqueIndices;
            int prevMat = -1;
            glm::vec4 minVert = { FLT_MAX, FLT_MAX, FLT_MAX, 0.0f };
            glm::vec4 maxVert = { -FLT_MAX, -FLT_MAX, -FLT_MAX, 0.0f };
            
            auto& inds = shape.mesh.indices;
            for (size_t i = 0, n = inds.size(); i < n; ++i)
            {
                size_t faceIndex = i / 3;

                ModelIndices indices = {};
                indices.values[0] = (float)inds[i].vertex_index;
                indices.values[1] = (float)inds[i].texcoord_index;
                indices.values[2] = (float)inds[i].normal_index;

                // split shapes additionally per material
                int mat = shape.mesh.material_ids[faceIndex];
                if (prevMat >= 0 && mat != prevMat)
                {
                    ad->m_cpuBuffers.push_back(createGeometryCpuData<float>(vertData, indexData, minVert, maxVert, k_numVertFloats));
                    md->m_materialIndices.push_back((size_t)prevMat);

                    vertData.clear();
                    indexData.clear();
                    uniqueIndices.clear();
                    for (int n = 0; i < 3; ++i)
                    {
                        minVert[n] = FLT_MAX;
                        maxVert[n] = -FLT_MAX;
                    }
                }
                prevMat = mat;

                auto unique = uniqueIndices.find(indices);
                if (unique == uniqueIndices.end())
                {
                    indices.index = (u32)(vertData.size() / k_numVertFloats);

                    // add a new vertex
                    // position
                    glm::vec4 pos = glm::vec4(0.0f, 0.0f, 0.0f, 1.0f);
                    if (indices.values[0] >= 0)
                    {
                        pos = glm::vec4(
                            verts[3 * indices.values[0] + 0],
                            verts[3 * indices.values[0] + 1],
                            verts[3 * indices.values[0] + 2],
                            1.0f
                        );
                    }
                    vertData.push_back(pos.x);
                    vertData.push_back(pos.y);
                    vertData.push_back(pos.z);
                    vertData.push_back(pos.w);
                        
                    // calculate aabb
                    for (int n = 0; n < 3; ++n)
                    {
                        if (pos[n] < minVert[n])
                            minVert[n] = pos[n];

                        if (pos[n] > maxVert[n])
                            maxVert[n] = pos[n];
                    }

                    // texcoord
                    glm::vec2 tex = glm::vec2(0.0f);
                    if (indices.values[1] >= 0)
                    {
                        tex = glm::vec2(
                            texs[2 * indices.values[1] + 0],
                            1.0f - texs[2 * indices.values[1] + 1]
                        );
                    }
                    vertData.push_back(tex.x);
                    vertData.push_back(tex.y);

                    // normal
                    glm::vec3 normal = glm::vec3(0.0f);
                    if (indices.values[2] >= 0)
                    {
                        normal = glm::vec3(
                            norms[3 * indices.values[2] + 0],
                            norms[3 * indices.values[2] + 1],
                            norms[3 * indices.values[2] + 2]
                        );
                    }
                    vertData.push_back(normal.x);
                    vertData.push_back(normal.y);
                    vertData.push_back(normal.z);
                        
                    uniqueIndices.insert(indices);
                    indexData.push_back(indices.index);
                }
                else
                {
                    indexData.push_back(unique->index);
                }

                // reverse the winding on the vertices
                if (indexData.size() % 3 == 0)
                {
                    u32 i1 = indexData[indexData.size() - 2];
                    u32 i2 = indexData[indexData.size() - 1];
                    indexData[indexData.size() - 1] = i1;
                    indexData[indexData.size() - 2] = i2;
                }
            }
            ad->m_cpuBuffers.push_back(createGeometryCpuData<float>(vertData, indexData, minVert, maxVert, k_numVertFloats));
            md->m_materialIndices.push_back((size_t)prevMat);
        }
        md->m_numUniques = ad->m_cpuBuffers.size();
        return true;
    }

    template<>
    bool ResourceManager::load<ModelLoadDesc>(const ModelLoadDesc& desc, Resource* resource)
    {
        auto ctx = gfx::getMainContext();

        ModelData* md = (ModelData*)(resource->m_internal);
        ModelAsyncData* ad = (ModelAsyncData*)(resource->m_asyncData);

        // complete the load of materials and textures
        for (auto& mat : ad->m_materials)
        {
            const MaterialLoadDesc& mld = *((MaterialLoadDesc*)(mat.getDesc()));
            load<MaterialLoadDesc>(mld, e_resourceLoadFlags::none);
            md->m_materials.push_back(mat);
        }

        // create gpu buffers for vertex/index data
        for (auto& geom : ad->m_cpuBuffers)
        {
            gfx::Geometry g = {};
            g.m_numVertices = geom.m_numVertices;
            g.m_numIndices = geom.m_numIndices;
            
            gfx::BufferDesc vertDesc = {};
            vertDesc.m_data = geom.m_vertexData;
            vertDesc.m_usage = gfx::e_bufferUsageFlags::vertex;
            vertDesc.m_access = gfx::e_accessFlags::none;
            g.m_vertexBuffer = ctx->createBuffer(vertDesc);


            gfx::BufferDesc indexDesc = {};
            indexDesc.m_data = geom.m_indexData;
            indexDesc.m_usage = gfx::e_bufferUsageFlags::index;
            indexDesc.m_access = gfx::e_accessFlags::none;
            g.m_indexBuffer = ctx->createBuffer(indexDesc);


            memcpy(g.m_minAABB, geom.m_minVert, sizeof(float) * 4);
            memcpy(g.m_maxAABB, geom.m_maxVert, sizeof(float) * 4);

            md->m_meshes.push_back(g);

            // delete the cpu buffers
            delete[] geom.m_vertexData.m_data;
            geom.m_vertexData.m_data = nullptr;

            delete[] geom.m_indexData.m_data;
            geom.m_indexData.m_data = nullptr;
        }

        // delete async data
        memDelete(resource->m_asyncData);
        resource->m_asyncData = nullptr;

        return true;
    }

    template<>
    void ResourceManager::unload<ModelData>(Resource* resource)
    {
        auto ctx = gfx::getMainContext();

        ModelData* md = (ModelData*)(resource->m_internal);
        
        ModelLoadDesc& mld = *(ModelLoadDesc*)resource->m_desc;
        if (mld.m_genData.m_data)
        {
            memDelete(mld.m_genData.m_data);
            mld.m_genData.m_data = nullptr;
        }

        // unload materials
        for (auto& m : md->m_materials)
        {
            unload<MaterialData>(m);
        }
        md->m_materials.clear();

        // delete buffers
        for (auto& m : md->m_meshes)
        {
            ctx->destroyBuffer(m.m_vertexBuffer);
            ctx->destroyBuffer(m.m_indexBuffer);
        }
        md->m_meshes.clear();
    }

    template<>
    ID ResourceManager::getIDFromDesc<ModelLoadDesc>(const ModelLoadDesc& desc)
    {
        if (!desc.m_filename.empty())
        {
            return hash(desc.m_filename);
        }
        else
        {
            // hash the bytes of the gen data
            ID id = 0;
            if (desc.m_genData.m_data)
            {
                u8* genDataBytes = (u8*)desc.m_genData.m_data;
                for (size_t i = 0; i < desc.m_genData.m_size; ++i)
                {
                    id = hashCombine(id, genDataBytes[i]);
                }
            }
            if (desc.m_genFunc)
                id = hashCombine(id, desc.m_genFunc);
            ASSERT(id); // don't have anything to hash for this model
            return id;
        }
    }

    template<>
    bool ResourceManager::loadAsync<MaterialLoadDesc>(const MaterialLoadDesc& desc, Resource* resource)
    {
        resource->m_internal = memNew<MaterialData>();
        resource->m_asyncData = memNew<gfx::MaterialBuffer>();
        MaterialData& material = *(MaterialData*)resource->m_internal;
        gfx::MaterialBuffer& data = *(gfx::MaterialBuffer*)(resource->m_asyncData);

        const tinyobj::material_t& params = desc.m_params;

        memcpy(data.m_ambient, params.ambient, sizeof(float) * 3);
        data.m_ambient[3] = 1.0f;
        memcpy(data.m_diffuse, params.diffuse, sizeof(float) * 3);
        data.m_diffuse[3] = 1.0f;
        memcpy(data.m_specular, params.specular, sizeof(float) * 3);
        data.m_specular[3] = 1.0f;
        data.m_shininess = params.shininess;
        data.m_ior = params.ior;
        data.m_dissolve = params.dissolve;
        data.m_illum = (float)params.illum;

        material.m_flags = 0;

        // start loading textures
        if (!params.diffuse_texname.empty())
        {
            material.m_textures.push_back(load<gfx::Texture, TextureLoadDesc>({ params.diffuse_texname, 8 }, e_resourceLoadFlags::async));
            material.m_flags |= gfx::e_materialFlags::diffuseMap;
        }

        if (!params.specular_texname.empty())
        {
            material.m_textures.push_back(load<gfx::Texture, TextureLoadDesc>({ params.specular_texname, 8 }, e_resourceLoadFlags::async));
            material.m_flags |= gfx::e_materialFlags::specularMap;
        }

        if (!params.bump_texname.empty())
        {
            material.m_textures.push_back(load<gfx::Texture, TextureLoadDesc>({ params.bump_texname, 8 }, e_resourceLoadFlags::async));
            material.m_flags |= gfx::e_materialFlags::bumpMap;
        }

        if (!params.alpha_texname.empty())
        {
            material.m_textures.push_back(load<gfx::Texture, TextureLoadDesc>({ params.alpha_texname, 1 }, e_resourceLoadFlags::async));
            material.m_flags |= gfx::e_materialFlags::alphaMap;
        }

        if (desc.m_textureFilenames.size() > 0)
        {
            material.m_flags = (u32)-1;
            for (auto& s : desc.m_textureFilenames)
            {
                material.m_textures.push_back(load<gfx::Texture, TextureLoadDesc>({ s, 8 }, e_resourceLoadFlags::async));
            }
        }

        material.m_defaultSamplerID = desc.m_defaultSamplerID;

        return true;
    }

    template<>
    bool ResourceManager::load<MaterialLoadDesc>(const MaterialLoadDesc& desc, Resource* resource)
    {
        auto ctx = gfx::getMainContext();

        MaterialData& material = *(MaterialData*)resource->m_internal;
        gfx::MaterialBuffer& data = *(gfx::MaterialBuffer*)(resource->m_asyncData);

        // complete texture loads
        for (auto& t : material.m_textures)
        {
            const TextureLoadDesc& tld = *(TextureLoadDesc*)t.getDesc();
            load<TextureLoadDesc>(tld, e_resourceLoadFlags::none);
        }

        // load material gpu buffer
        gfx::BufferDesc bd = {};
        bd.m_usage = gfx::e_bufferUsageFlags::uniform;
        bd.m_access = gfx::e_accessFlags::none;
        bd.m_data.m_size = sizeof(gfx::MaterialBuffer);
        bd.m_data.m_data = resource->m_asyncData;
        material.m_buffer = ctx->createBuffer(bd);

        // clean up async data
        memDelete(resource->m_asyncData);

        return true;
    }

    template<>
    void ResourceManager::unload<MaterialData>(Resource* resource)
    {

    }

    template<>
    ID ResourceManager::getIDFromDesc<MaterialLoadDesc>(const MaterialLoadDesc& desc)
    {
        ID id = hash(desc.m_filename);
        id = hashCombine(id, desc.m_params.name);
        return id;
    }
}
