#include "renderer.h"

#include "gfx.h"
#include "platform.h"
#include "json.h"

#include <pystring.h>

namespace
{
    const char* k_blendFactorNames[] =
    {
        "zero",
        "one",
        "srcColor",
        "oneMinusSrcColor",
        "dstColor",
        "oneMinusDstColor",
        "srcAlpha",
        "oneMinusSrcAlpha",
        "dstAlpha",
        "oneMinusDstAlpha",
        "constantColor",
        "oneMinusConstantColor",
        "constantAlpha",
        "oneMinusConstantAlpha",
        "src1Color",
        "oneMinusSrc1Color",
        "src1Alpha",
        "oneMinusSrc1Alpha"
    };

    const char* k_blendOpNames[] = 
    {
        "add",
        "subtract",
        "reverseSubtract",
        "min",
        "max"
    };
    
    const char* k_compareOpNames[] = 
    {
        "never",
        "less",
        "equal",
        "lessOrEqual",
        "greater",
        "notEqual",
        "greatorOrEqual",
        "always"
    };

    const char* k_stencilOpNames[] =
    {
        "keep",
        "zero",
        "replace",
        "incrementClamp",
        "decrementClamp",
        "invert",
        "incrementWrap",
        "decrementWrap"
    };    

    const char* k_typenameNames[] =
    {
         "unorm",
         "snorm",
         "uscaled",
         "sscaled",
         "uint",
         "int",
         "float"
    };

    const char* k_filterNames[] =
    {
        "nearest",
        "linear"
    };

    const char* k_mipmapModes[] =
    {
        "nearest",
        "linear"
    };

    const char* k_addressModeNames[] =
    {
        "repeat",
        "mirroredRepeat",
        "clampToEdge",
        "clampToBorder",
        "mirrorClampToEdge"
    };

    const char* k_borderColourNames[] =
    {
        "transparentBlack",
        "opaqueBlack",
        "opaqueWhite"
    };

    const char* k_cullModeNames[] = 
    {
        "front",
        "back"
    };
}

namespace engine
{
    bool findExistingDependanciesRecursive(View* current, View* dependant)
    {
        for (auto d : current->m_dependancies)
        {
            if (d == dependant)
            {
                return true;
            }
            else
            {
                return findExistingDependanciesRecursive(d, dependant);
            }
        }
        return false;
    }

    void Renderer::setup()
    {

    }

    void Renderer::shutdown()
    {

    }

    gfx::RenderPass createViewRenderPass(View& view)
    {
        auto ctx = gfx::getMainContext();

        gfx::RenderPassDesc rpd = {};
        std::vector<gfx::TextureDesc*> tds;

        // main rendering subpass
        gfx::SubpassDesc spd = {};
        spd.m_prevSubpassStageDependency = gfx::e_pipelineStageType::colorOutput;
        spd.m_subpassWaitStage = gfx::e_pipelineStageType::colorOutput;
        for(int i = 0; i < view.m_numTargets; ++i)
        {
            gfx::Texture tex = view.m_targetTextures[i].m_texture;

            gfx::TextureDesc* td = memNew<gfx::TextureDesc>();
            if (tex)
                *td = ctx->getTextureDesc(tex);
            else
                *td = ctx->getBackbufferTextureDesc();

            tds.push_back(td);

            view.m_viewport.x = 0.0f;
            view.m_viewport.y = 0.0f;
            view.m_viewport.m_width = (float)td->m_width;
            view.m_viewport.m_height = (float)td->m_height;
            view.m_viewport.m_minDepth = 0.0f;
            view.m_viewport.m_maxDepth = 1.0f;

            gfx::AttachmentDesc ad = {};
            ad.m_texture = td;
            ad.m_storeOp = gfx::e_storeOpType::store;
            ad.m_stencilStoreOp = gfx::e_storeOpType::store;
            ad.m_loadOp = view.m_targetTextures[i].m_loadOp;
            ad.m_stencilLoadOp = gfx::e_loadOpType::load;
            spd.m_colorAttachments.push_back(ad);

            if (view.m_resolve == e_resolveType::after)
            {
                gfx::TextureDesc* rtd = memNew<gfx::TextureDesc>();
                *rtd = *td;
                rtd->m_samples = 1;
                ad.m_texture = rtd;
                tds.push_back(rtd);

                spd.m_resolveAttachments.push_back(ad);
            }
        }

        if (view.m_depthTarget.m_texture)
        {
            gfx::TextureDesc* td = memNew<gfx::TextureDesc>();
            *td = ctx->getTextureDesc(view.m_depthTarget.m_texture);
            tds.push_back(td);

            gfx::AttachmentDesc ad = {};
            ad.m_texture = td;
            ad.m_loadOp = view.m_depthTarget.m_loadOp;
            ad.m_storeOp = gfx::e_storeOpType::store;
            ad.m_stencilStoreOp = gfx::e_storeOpType::store;
            ad.m_stencilLoadOp = view.m_depthTarget.m_loadOp;
            spd.m_depthAttachments.push_back(ad);

            // use depth stencil as the viewport dimensions if no other textures are used
            if (view.m_numTargets == 0)
            {
                view.m_viewport.x = 0.0f;
                view.m_viewport.y = 0.0f;
                view.m_viewport.m_width = (float)td->m_width;
                view.m_viewport.m_height = (float)td->m_height;
                view.m_viewport.m_minDepth = 0.0f;
                view.m_viewport.m_maxDepth = 1.0f;
            }
        }

        rpd.m_subpasses.push_back(spd);

        gfx::RenderPass rp = ctx->createRenderPass(rpd);

        for (auto td : tds)
            memDelete(td);
        tds.clear();

        return rp;
    }

    gfx::Shader loadShader(const rapidjson::Value& sj, gfx::ShaderStageFlags stage, gfx::VertexInputDesc* inputDesc, size_t* numOutputs)
    {
        auto ctx = gfx::getMainContext();

        gfx::ShaderDesc shaderDesc;
        size_t offset = 0;
        for (u32 i = 0, n = sj["Push Constants"].GetArray().Size(); i < n; ++i)
        {
            const rapidjson::Value& pushConstant = sj["Push Constants"][i];
            gfx::PushConstantDesc pushDesc;
            pushDesc.m_size = 0;
            for (u32 j = 0; j < pushConstant["Members"].GetArray().Size(); ++j)
            {
                const rapidjson::Value& member = pushConstant["Members"][j];
                size_t size = 0;
                std::string typeString = member["Type"].GetString();
                if (typeString == "uint")
                    size = sizeof(uint32_t);
                else if (typeString == "int")
                    size = sizeof(int);
                else if (typeString == "float")
                    size = sizeof(float);
                size = size * member["Count"].GetUint();
                pushDesc.m_size += (u32)size;
            }
            pushDesc.m_offset = (u32)offset;
            offset += pushDesc.m_size;
            pushDesc.m_binding = pushConstant["Binding"].GetUint();
            shaderDesc.m_pushConstants.push_back(pushDesc);
        }

        for (int i = 0, n = sj["Uniforms"].GetArray().Size(); i < n; ++i)
        {
            const rapidjson::Value& uniform = sj["Uniforms"][i];
            gfx::DescriptorDesc descriptor;
            std::string typeString = uniform["Type"].GetString();
            if (typeString == "Combined Sampler")
                descriptor.m_type = gfx::e_descriptorType::combinedImageSampler;
            else if (typeString == "Sampler")
                descriptor.m_type = gfx::e_descriptorType::sampler;
            else if (typeString == "Texture")
                descriptor.m_type = gfx::e_descriptorType::sampledImage;
            else if (typeString == "Storage Image")
                descriptor.m_type = gfx::e_descriptorType::storageImage;
            else if (typeString == "Uniform Buffer")
                descriptor.m_type = gfx::e_descriptorType::uniformBuffer;
            else if (typeString == "Storage Buffer")
                descriptor.m_type = gfx::e_descriptorType::storageBuffer;
            descriptor.m_count = uniform["Count"].GetUint();
            descriptor.m_binding = uniform["Binding"].GetUint();
            if (uniform.FindMember("Mapping") != uniform.MemberEnd())
                descriptor.m_mapping = uniform["Mapping"].GetUint();
            descriptor.m_name = uniform["Name"].GetString();
            shaderDesc.m_descriptors.push_back(descriptor);
        }

        std::string api = gfx::getAPI();
        ASSERT(sj["Shader Files"].HasMember(api.c_str()));
        auto& shaderJson = sj["Shader Files"][api.c_str()];
        std::string shaderData = std::string(shaderJson.GetString(), shaderJson.GetStringLength());
        shaderDesc.m_data.m_data = (void*)shaderData.c_str();
        shaderDesc.m_data.m_size = sj["Shader Files"][api.c_str()].GetStringLength();
        shaderDesc.m_entry = "main";
        shaderDesc.m_stage = stage;
        gfx::Shader shader = ctx->createShader(shaderDesc);

        if (stage & gfx::e_shaderStageFlags::vertex && inputDesc)
        {
            // read vertex layout from vertex shader input
            size_t perVertexOffset = 0;
            size_t perInstanceOffset = 0;
            for (int i = 0, n = sj["Inputs"].GetArray().Size(); i < n; ++i)
            {
                const rapidjson::Value& input = sj["Inputs"][i];
                bool instanced = input["Instanced"].GetBool();
                u32 count = input["Count"].GetUint();
                u32 numIndices = count > 4 ? count / 4 : 1;
                count = count > 4 ? 4 : count;
                gfx::VertexAttributeDesc desc = {};
                desc.m_location = i;
                desc.m_binding = instanced ? 1 : 0;
                desc.m_format.m_channels = (1 << count) - 1;
                desc.m_format.m_bitDepth = 32;
                desc.m_format.m_type = enumNameToValue<gfx::FormatType>(k_typenameNames, ARRAY_SIZE(k_typenameNames), getJsonValueOrDefault<const char*>(input, "Type", "unorm"));
                
                for (u32 j = 0; j < numIndices; ++j)
                {
                    desc.m_offset = instanced ? (u32)perInstanceOffset : (u32)perVertexOffset;
                    if (instanced)
                        perInstanceOffset += (size_t)((desc.m_format.m_bitDepth / 8) * input["Count"].GetUint() * (1.0f / numIndices));
                    else
                        perVertexOffset += (size_t)((desc.m_format.m_bitDepth / 8) * input["Count"].GetUint() * (1.0f / numIndices));
                    desc.m_location = i + j;
                    inputDesc->m_attributes.push_back(desc);
                }
            }
            inputDesc->m_bindings.push_back({ 0, perVertexOffset, gfx::e_inputRateType::vertex });
            if (perInstanceOffset > 0)
                inputDesc->m_bindings.push_back({ 1, perInstanceOffset, gfx::e_inputRateType::instance });
        }
        else if (stage & gfx::e_shaderStageFlags::fragment && numOutputs)
        {
            *numOutputs = sj["Outputs"].GetArray().Size();
        }
        return shader;
    }

    gfx::BlendDesc loadBlendDesc(const rapidjson::Value& blend)
    {
        gfx::BlendDesc bd = {};
        bd.m_blendEnable = getJsonValueOrDefault<bool>(blend, "enabled", false);
        bd.m_srcColorBlendFactor = enumNameToValue<gfx::BlendFactor>(k_blendFactorNames, sizeof(k_blendFactorNames) / sizeof(k_blendFactorNames[0]), getJsonValueOrDefault<const char*>(blend, "srcColorBlendFactor", "one"));
        bd.m_dstColorBlendFactor = enumNameToValue<gfx::BlendFactor>(k_blendFactorNames, sizeof(k_blendFactorNames) / sizeof(k_blendFactorNames[0]), getJsonValueOrDefault<const char*>(blend, "dstColorBlendFactor", "one"));
        bd.m_colorBlendOp = enumNameToValue<gfx::BlendOp>(k_blendOpNames, sizeof(k_blendOpNames) / sizeof(k_blendOpNames[0]), getJsonValueOrDefault<const char*>(blend, "colorBlendOp", "add"));
        bd.m_srcAlphaBlendFactor = enumNameToValue<gfx::BlendFactor>(k_blendFactorNames, sizeof(k_blendFactorNames) / sizeof(k_blendFactorNames[0]), getJsonValueOrDefault<const char*>(blend, "srcAlphaBlendFactor", "one"));
        bd.m_dstAlphaBlendFactor = enumNameToValue<gfx::BlendFactor>(k_blendFactorNames, sizeof(k_blendFactorNames) / sizeof(k_blendFactorNames[0]), getJsonValueOrDefault<const char*>(blend, "dstAlphaBlendFactor", "one"));
        bd.m_alphaBlendOp = enumNameToValue<gfx::BlendOp>(k_blendOpNames, sizeof(k_blendOpNames) / sizeof(k_blendOpNames[0]), getJsonValueOrDefault<const char*>(blend, "alphaBlendOp", "add"));
        return bd;
    }

    gfx::DepthStencilDesc loadDepthStencilDesc(const rapidjson::Value& depthStencil)
    {
        gfx::DepthStencilDesc dsd = {};
        dsd.m_depthTest = getJsonValueOrDefault<bool>(depthStencil, "depthTest", false);
        dsd.m_depthWrite = getJsonValueOrDefault<bool>(depthStencil, "depthWrite", true);
        dsd.m_depthCompareOp = enumNameToValue<gfx::CompareOp>(k_compareOpNames, sizeof(k_compareOpNames) / sizeof(k_compareOpNames[0]), getJsonValueOrDefault<const char*>(depthStencil, "compareOp", "always"));
        dsd.m_depthBounds = getJsonValueOrDefault<bool>(depthStencil, "depthBounds", false);
        dsd.m_stencilTest = getJsonValueOrDefault<bool>(depthStencil, "stencilTest", false);
        dsd.m_front.m_failOp = enumNameToValue<gfx::StencilOp>(k_stencilOpNames, sizeof(k_stencilOpNames) / sizeof(k_stencilOpNames[0]), getJsonValueOrDefault<const char*>(depthStencil, "frontFailOp", "keep"));
        dsd.m_front.m_passOp = enumNameToValue<gfx::StencilOp>(k_stencilOpNames, sizeof(k_stencilOpNames) / sizeof(k_stencilOpNames[0]), getJsonValueOrDefault<const char*>(depthStencil, "frontPassOp", "keep"));
        dsd.m_front.m_depthFailOp = enumNameToValue<gfx::StencilOp>(k_stencilOpNames, sizeof(k_stencilOpNames) / sizeof(k_stencilOpNames[0]), getJsonValueOrDefault<const char*>(depthStencil, "frontDepthFailOp", "keep"));
        dsd.m_front.m_compareMask = getJsonValueOrDefault<u32>(depthStencil, "frontCompareMask", 0);
        dsd.m_front.m_writeMask = getJsonValueOrDefault<u32>(depthStencil, "frontWriteMask", 0);
        dsd.m_front.m_reference = getJsonValueOrDefault<u32>(depthStencil, "frontReference", 0);
        dsd.m_back.m_failOp = enumNameToValue<gfx::StencilOp>(k_stencilOpNames, sizeof(k_stencilOpNames) / sizeof(k_stencilOpNames[0]), getJsonValueOrDefault<const char*>(depthStencil, "backFailOp", "keep"));
        dsd.m_back.m_passOp = enumNameToValue<gfx::StencilOp>(k_stencilOpNames, sizeof(k_stencilOpNames) / sizeof(k_stencilOpNames[0]), getJsonValueOrDefault<const char*>(depthStencil, "backPassOp", "keep"));
        dsd.m_back.m_depthFailOp = enumNameToValue<gfx::StencilOp>(k_stencilOpNames, sizeof(k_stencilOpNames) / sizeof(k_stencilOpNames[0]), getJsonValueOrDefault<const char*>(depthStencil, "backDepthFailOp", "keep"));
        dsd.m_back.m_compareMask = getJsonValueOrDefault<u32>(depthStencil, "backCompareMask", 0);
        dsd.m_back.m_writeMask = getJsonValueOrDefault<u32>(depthStencil, "backWriteMask", 0);
        dsd.m_back.m_reference = getJsonValueOrDefault<u32>(depthStencil, "backReference", 0);
        dsd.m_minDepthBounds = getJsonValueOrDefault<float>(depthStencil, "minDepthBounds", 0.0f);
        dsd.m_maxDepthBounds = getJsonValueOrDefault<float>(depthStencil, "maxDepthBounds", 1.0f);
        return dsd;
    }

    void Renderer::loadConfig(const char* filename)
    {
        if(!utils::fileExists(filename))
            return;

        void* buf = nullptr;
        size_t len = 0;
        utils::readFileToBuffer(filename, buf, len);
        bool result = m_cfg.Parse<rapidjson::kParseStopWhenDoneFlag>((const char*)buf).HasParseError();
        if (result)
        {
            auto error = m_cfg.GetParseError();
            printf("Error!");
            return;
        }

        auto ctx = gfx::getMainContext();

        m_msaaSamples = getJsonValueOrDefault<u32>(m_cfg, "msaa_samples", 1);

        if(m_cfg.HasMember("pipelines"))
        {
            for(auto it = m_cfg["pipelines"].MemberBegin(); it != m_cfg["pipelines"].MemberEnd(); ++it)
            {
                const rapidjson::Value& j = it->value;
                std::string pipelineName = it->name.GetString();
                ID id = hash(pipelineName);

                // pipeline shaders are loaded from a single combined file
                std::string shaderFilepath = "data/shaders/" + pipelineName + ".shader";
                void* sbuf = nullptr;
                size_t slen = 0;
                utils::readFileToBuffer(shaderFilepath.c_str(), sbuf, slen);
                if (!sbuf)
                    continue;

                rapidjson::Document sd;
                sd.Parse<rapidjson::kParseStopWhenDoneFlag>((const char*)sbuf, slen);
                if (sd.HasParseError())
                    continue;

                utils::freeFileBuffer(sbuf);

                PartialPipeline pp = {};
                pp.m_raster.m_cullMode = enumNameToFlagValue(k_cullModeNames, ARRAY_SIZE(k_cullModeNames), getJsonValueOrDefault<const char*>(j, "cull_mode", "back"));
                pp.m_raster.m_frontFace = gfx::e_frontFaceType::counterClockwise;

                // blend desc
                std::vector<std::string> blendNames;
                if(j.HasMember("blends"))
                {
                    for(auto bit = j["blends"].Begin(); bit != j["blends"].End(); ++bit)
                    {
                        blendNames.push_back((*bit).GetString());
                    }
                }
                else
                {
                    blendNames.push_back("disabled");
                }

                std::vector<gfx::BlendDesc> blendDescs;
                for(const auto& blendName : blendNames)
                {
                    const rapidjson::Value& blend = m_cfg["blend_states"][blendName.c_str()];
                    blendDescs.push_back(loadBlendDesc(blend));
                }

                // depth stencil desc
                std::string depthStencilName = getJsonValueOrDefault<const char*>(j, "depth_stencil", "disabled");
                const rapidjson::Value& depthStencil = m_cfg["depth_stencil_states"][depthStencilName.c_str()];
                pp.m_depthStencil = loadDepthStencilDesc(depthStencil);

                pp.m_dynamic.m_dynamicState = (gfx::e_dynamicStateFlags::viewport | gfx::e_dynamicStateFlags::scissor);

                const rapidjson::Value& jvs = sd["variations"].GetArray();
                size_t numVariations = jvs.Size();
                for (size_t i = 0; i < numVariations; ++i)
                {
                    memset(pp.m_shaders, 0, sizeof(gfx::Shader) * 6);
                    pp.m_vertexInput = gfx::VertexInputDesc();

                    const rapidjson::Value& sj = jvs[i];
                    if (j.HasMember("vertex_shader") && sj.HasMember("vert"))
                    {
                        pp.m_shaders[0] = loadShader(sj["vert"], gfx::e_shaderStageFlags::vertex, &pp.m_vertexInput, nullptr);
                        pp.m_shaderStageFlags |= gfx::e_shaderStageFlags::vertex;
                    }

                    if (j.HasMember("tessellation_control_shader") && sj.HasMember("tesc"))
                    {
                        pp.m_shaders[1] = loadShader(sj["tesc"], gfx::e_shaderStageFlags::tessellationControl, nullptr, nullptr);
                        pp.m_shaderStageFlags |= gfx::e_shaderStageFlags::tessellationControl;
                    }

                    if (j.HasMember("tessellation_evaluation_shader") && sj.HasMember("tese"))
                    {
                        pp.m_shaders[2] = loadShader(sj["tese"], gfx::e_shaderStageFlags::tessellationEvaluation, nullptr, nullptr);
                        pp.m_shaderStageFlags |= gfx::e_shaderStageFlags::tessellationEvaluation;
                    }

                    if (j.HasMember("geometry_shader") && sj.HasMember("geom"))
                    {
                        pp.m_shaders[3] = loadShader(sj["geom"], gfx::e_shaderStageFlags::geometry, nullptr, nullptr);
                        pp.m_shaderStageFlags |= gfx::e_shaderStageFlags::geometry;
                    }

                    size_t numOutputs = 0;
                    if (j.HasMember("fragment_shader") && sj.HasMember("frag"))
                    {
                        pp.m_shaders[4] = loadShader(sj["frag"], gfx::e_shaderStageFlags::fragment, nullptr, &numOutputs);
                        pp.m_shaderStageFlags |= gfx::e_shaderStageFlags::fragment;
                    }

                    if (sj.HasMember("blend_state"))
                    {
                        const rapidjson::Value& blend = sj["blend_state"];
                        blendDescs.clear();
                        blendDescs.push_back(loadBlendDesc(blend));
                    }
                    
                    // splat last blend desc if fewer blend descs have been specified than number of outputs
                    pp.m_blends.resize(numOutputs);
                    for (size_t k = 0; k < numOutputs; ++k)
                    {
                        if (k < blendDescs.size())
                            pp.m_blends[k] = blendDescs[k];
                        else
                            pp.m_blends[k] = blendDescs.back();
                    }
                    
                    ID completeID = id;
                    if (sj.HasMember("variation"))
                    {
                        ID variationID = 0;
                        const rapidjson::Value& jvn = sj["variation"].GetArray();
                        size_t numVariationNames = jvn.Size();
                        for (size_t k = 0; k < numVariationNames; ++k)
                        {
                            std::string variationName = jvn[k].GetString();
                            if (variationID)
                                variationID = hashCombine(variationID, variationName);
                            else
                                variationID = hash(variationName);
                        }
                        completeID = completeID ^ variationID;
                    }
                    m_partialPipelines[completeID] = pp;

                    // if this is the default variation of the pipeline also add the partial pipeline with no variation id
                    // TODO: this leads to duplicate pipelines when using defaults so fix that
                    if (getJsonValueOrDefault(sj, "default_variation", false))
                        m_partialPipelines[id] = pp;
                }
            }
        }

        if(m_cfg.HasMember("render_targets"))
        {
            for (auto it = m_cfg["render_targets"].MemberBegin(); it != m_cfg["render_targets"].MemberEnd(); ++it)
            {
                const rapidjson::Value& j = it->value;
                ID id = hash<std::string>(it->name.GetString());

                RenderTarget rt = {};

                gfx::TextureDesc td = {};
                td.m_usage = gfx::e_textureUsageFlags::shaderWrite | gfx::e_textureUsageFlags::shaderRead;

                bool isOutputRes = false;
                if (j["size"].IsArray())
                {
                    td.m_width = j["size"][0].GetUint();
                    td.m_height = j["size"][1].GetUint();
                }
                else if (j["size"] == "output_resolution")
                {
                    td.m_width = utils::getWindowWidth();
                    td.m_height = utils::getWindowHeight();
                    isOutputRes = true;
                }

                td.m_samples = 1;
                if (j.HasMember("samples") && j["samples"].IsString())
                {
                    std::string samplesName = getJsonValueOrDefault(j, "samples", "");
                    if (samplesName == "user_defined")
                    {
                        td.m_samples = m_msaaSamples;
                        rt.m_variableSamples = true;
                    }
                }
                else
                {
                    td.m_samples = getJsonValueOrDefault<u32>(j, "samples", 1);
                }

                // parse render target texture format
                std::string format = j["format"].GetString();
                td.m_format.m_channels = 0;
                if (format[0] != 'd')
                {
                    char channels[4] = { 'r', 'g', 'b', 'a' };
                    for (int i = 0; i < 4; ++i)
                    {
                        if (format.find(channels[i]) != std::string::npos)
                        {
                            td.m_format.m_channels |= 1 << i;
                        }
                    }

                    if (format[0] == 's')
                    {
                        td.m_format.m_standard = true;
                    }

                    if (pystring::endswith(format, "f"))
                        td.m_format.m_type = gfx::e_formatType::signedFloat;
                }
                else
                {
                    td.m_format.m_channels |= gfx::e_formatChannelFlags::d;
                    if(format.find('s') != std::string::npos)
                    {
                        td.m_format.m_channels |= gfx::e_formatChannelFlags::ds;
                    }
                    td.m_usage &= ~gfx::e_textureUsageFlags::shaderWrite;
                    td.m_usage |= gfx::e_textureUsageFlags::depthStencil;
                }

                size_t ds = -1;
                size_t de = format.size();
                for (size_t i = 0; i < format.size(); ++i)
                {
                    if (format[i] >= '0' && format[i] <= '9')
                    {
                        if (ds == (size_t)-1)
                        {
                            ds = i;
                        }
                    }
                    else if (ds != (size_t)-1)
                    {
                        de = i;
                        break;
                    }
                }
                td.m_format.m_bitDepth = std::stoi(format.substr(ds, de - ds));

                rt.m_texture[0] = ctx->createTexture(td);
                rt.m_texture[1] = nullptr;
                rt.m_outputRes = isOutputRes;
                m_renderTargets[id] = rt;
            }
        }

        if (m_cfg.HasMember("samplers"))
        {
            for (auto it = m_cfg["samplers"].MemberBegin(); it != m_cfg["samplers"].MemberEnd(); ++it)
            {
                const rapidjson::Value& j = it->value;
                ID id = hash<std::string>(it->name.GetString());

                gfx::SamplerDesc sd = {};
                sd.m_magFilter = enumNameToValue<gfx::Filter>(k_filterNames, ARRAY_SIZE(k_filterNames), getJsonValueOrDefault<const char*>(j, "mag", "nearest"));
                sd.m_minFilter = enumNameToValue<gfx::Filter>(k_filterNames, ARRAY_SIZE(k_filterNames), getJsonValueOrDefault<const char*>(j, "min", "nearest"));
                sd.m_mipmapMode = enumNameToValue<gfx::MipmapMode>(k_mipmapModes, ARRAY_SIZE(k_mipmapModes), getJsonValueOrDefault<const char*>(j, "mipmap", "nearest"));
                sd.m_addressModeU = enumNameToValue<gfx::AddressMode>(k_addressModeNames, ARRAY_SIZE(k_addressModeNames), getJsonValueOrDefault<const char*>(j, "addressModeU", "clampToEdge"));
                sd.m_addressModeV = enumNameToValue<gfx::AddressMode>(k_addressModeNames, ARRAY_SIZE(k_addressModeNames), getJsonValueOrDefault<const char*>(j, "addressModeV", "clampToEdge"));
                sd.m_addressModeW = enumNameToValue<gfx::AddressMode>(k_addressModeNames, ARRAY_SIZE(k_addressModeNames), getJsonValueOrDefault<const char*>(j, "addressModeW", "clampToEdge"));
                sd.m_mipLodBias = getJsonValueOrDefault(j, "mipLodBias", 0.0f);
                sd.m_maxAnisotropy = getJsonValueOrDefault(j, "maxAnisotropy", 1.0f);
                sd.m_anisotropyEnable = sd.m_maxAnisotropy > 1.0f;
                if (j.HasMember("compareOp"))
                {
                    sd.m_compareEnable = true;
                    sd.m_compareOp = enumNameToValue<gfx::CompareOp>(k_compareOpNames, ARRAY_SIZE(k_compareOpNames), getJsonValueOrDefault<const char*>(j, "compareOp", "always"));
                }
                sd.m_minLod = getJsonValueOrDefault(j, "minLod", -1000.0f);
                sd.m_maxLod = getJsonValueOrDefault(j, "maxLod", 1000.0f);
                sd.m_borderColor = enumNameToValue<gfx::BorderColor>(k_borderColourNames, ARRAY_SIZE(k_borderColourNames), getJsonValueOrDefault<const char*>(j, "borderColour", "transparentBlack"));
                sd.m_unnormalizedCoordinates = getJsonValueOrDefault(j, "unnormalizedCoordinates", false);
                m_samplers[id] = ctx->createSampler(sd);
            }
        }        
    }

    View Renderer::setupView(const View& v)
    {
        auto ctx = gfx::getMainContext();

        View view = v;
        view.m_idName = hash(view.m_name);

        const rapidjson::Value& j = m_cfg["views"][view.m_name.c_str()];

        if (j.HasMember("pipeline"))
        {
            std::string pipelineName = j["pipeline"].GetString();
            view.m_pipeline = hash(pipelineName);
        }

        if (j.HasMember("pipeline_variation"))
        {
            std::string joinedVariationNames = j["pipeline_variation"].GetString();
            std::vector<std::string> variationNames;
            pystring::split(joinedVariationNames, variationNames, "/");
            view.m_pipelineVariation = 0;
            for (const std::string& s : variationNames)
            {
                if (view.m_pipelineVariation == 0)
                    view.m_pipelineVariation = hash(s);
                else
                    view.m_pipelineVariation = hashCombine(view.m_pipelineVariation, s);
            }
        }

        if (j.HasMember("components"))
        {
            view.m_components = j["components"].GetUint();
        }

        bool resolve = getJsonValueOrDefault(j, "resolve", false);
        if (resolve)
            view.m_resolve = e_resolveType::standalone;

        // read view sampled texture ids
        if (j.HasMember("textures"))
        {
            size_t idx = 0;
            for (auto tit = j["textures"].Begin(); tit != j["textures"].End(); ++tit)
            {
                const rapidjson::Value& tVal = *tit;
                std::string textureName = tVal["texture"].GetString();
                view.m_textures.m_textureID[idx] = hash(textureName);

                std::string samplerName = tVal["sampler"].GetString();
                view.m_textures.m_samplerID[idx] = hash(samplerName);

                
                ++idx;
            }
        }

        // assign sampled textures from ids and find any dependancies
        std::set<View*> potentialDependancies;
        for (size_t i = 0; i < 4; ++i)
        {
            ID tid = view.m_textures.m_textureID[i];
            ID sid = view.m_textures.m_samplerID[i];

            if (tid == 0)
                continue;

            gfx::Texture tex = getRenderTarget(tid, e_renderTargetUsage::read);
            view.m_textures.m_texture[i] = tex;
            view.m_textures.m_sampler[i] = m_samplers[sid];

            // find the view that we're relying on to draw to this texture for synchronising
            bool found = false;
            for (auto rit = m_views.rbegin(); rit != m_views.rend(); ++rit)
            {
                for (size_t i = 0; i < (*rit)->m_numTargets; ++i)
                {
                    if ((*rit)->m_targetTextures[i].m_texture == tex)
                    {
                        // don't include duplicate dependancies
                        potentialDependancies.insert(*rit);
                        found = true;
                        break;
                    }
                }
                if (found)
                    break;

                if ((*rit)->m_depthTarget.m_texture == tex)
                {
                    found = true;
                    break;
                }
            }
            ASSERT(found); // couldn't find dependancy for this render target!

        }

        // filter dependancies
        for (auto& a : potentialDependancies)
        {
            bool found = false;
            for (auto& b : potentialDependancies)
            {
                if (a == b)
                    continue;

                if (findExistingDependanciesRecursive(b, a))
                {
                    found = true;
                    break;
                }
            }
            if (!found)
                view.m_dependancies.insert(a);
        }

        // uniform buffer
        if (!view.m_uniformBufferMem.m_data && j.HasMember("uniforms"))
        {
            size_t uniformCount = j["uniforms"].Size();
            size_t uniformSize = uniformCount * sizeof(float);
            view.m_uniformBufferMem.m_size = uniformSize;
            view.m_uniformBufferMem.m_data = memAlloc(uniformSize);
            for (size_t i = 0; i < uniformCount; ++i)
            {
                ((float*)view.m_uniformBufferMem.m_data)[i] = j["uniforms"][i].GetFloat();
            }
        }

        if (view.m_uniformBufferMem.m_data)
        {
            gfx::BufferDesc bd = {};
            bd.m_data.m_data = view.m_uniformBufferMem.m_data;
            bd.m_data.m_size = view.m_uniformBufferMem.m_size;
            bd.m_usage = gfx::e_bufferUsageFlags::uniform;
            view.m_uniformBuffer = ctx->createBuffer(bd);
        }

        // colour attachments
        if (j.HasMember("target"))
        {
            if (j["target"].IsArray())
            {
                for (size_t i = 0; i < j["target"].Size();++i)
                {
                    // don't overwrite texture ids set elsewhere
                    if (i < view.m_numTargets)
                        continue;

                    ID targetID = hash(std::string(j["target"][i].GetString()));
                    view.m_targetTextures[view.m_numTargets++].m_textureID = targetID;
                }
            }
            else if(view.m_numTargets == 0)
            {
                view.m_targetTextures[view.m_numTargets++].m_textureID = hash(std::string(j["target"].GetString()));
            }
        }

        gfx::Clear clear = {};
        gfx::LoadOp loadOp = gfx::e_loadOpType::load;
        if (j.HasMember("clear"))
        {
            const rapidjson::Value& clearVals = j["clear"];
            loadOp = gfx::e_loadOpType::clear;
            clear.r = clearVals[0].GetFloat();
            clear.g = clearVals[1].GetFloat();
            clear.b = clearVals[2].GetFloat();
            clear.a = clearVals[3].GetFloat();
        }

        ID backbufferID = hash(std::string("backbuffer"));
        for(size_t i = 0; i < view.m_numTargets; ++i)
        {
            view.m_targetTextures[i].m_loadOp = loadOp;
            view.m_targetTextures[i].m_clear = clear;

            // don't overwrite textures set elsewhere
            if (view.m_targetTextures[i].m_texture)
                continue;

            gfx::Texture target = nullptr;
            if (view.m_targetTextures[i].m_textureID != backbufferID)
            {
                target = getRenderTarget(view.m_targetTextures[i].m_textureID, resolve ? e_renderTargetUsage::resolve : e_renderTargetUsage::write);
            }
            view.m_targetTextures[i].m_texture = target;
        }

        // depth attachment
        gfx::Clear depthClear = {};
        gfx::LoadOp depthLoadOp = gfx::e_loadOpType::load;
        if (j.HasMember("clear_depth"))
        {
            depthLoadOp = gfx::e_loadOpType::clear;
            depthClear.depth = j["clear_depth"]["depth"].GetFloat();
            depthClear.stencil = j["clear_depth"]["stencil"].GetUint();
        }

        if (view.m_depthTarget.m_textureID == 0 && j.HasMember("depth_target"))
        {
            view.m_depthTarget.m_textureID = hash<std::string>(j["depth_target"].GetString());
        }

        if (!view.m_depthTarget.m_texture && view.m_depthTarget.m_textureID != 0)
        {
            view.m_depthTarget.m_texture = getRenderTarget(view.m_depthTarget.m_textureID, e_renderTargetUsage::write);
        }
        view.m_depthTarget.m_loadOp = depthLoadOp;
        view.m_depthTarget.m_clear = depthClear;

        view.m_pass = createViewRenderPass(view);

        // create the generic view uniforms buffer
        GenericViewUniforms g = {};
        g.m_dimensions[0] = view.m_viewport.m_width;
        g.m_dimensions[1] = view.m_viewport.m_height;
        g.m_dimensions[2] = 1.0f / view.m_viewport.m_width; // texel width
        g.m_dimensions[3] = 1.0f / view.m_viewport.m_height; // texel height
        
        gfx::BufferDesc bd = {};
        bd.m_access = gfx::e_accessFlags::write;
        bd.m_usage = gfx::e_bufferUsageFlags::uniform;
        bd.m_data.m_data = &g;
        bd.m_data.m_size = sizeof(GenericViewUniforms);
        view.m_genericViewUniformBuffer = ctx->createBuffer(bd);

        // find view render function
        ASSERT(j.HasMember("render_function"));
        ID funcid = hash<std::string>(j["render_function"].GetString());
        view.m_renderFunction = m_renderFuncs.at(funcid);

        // reset the 'in-view' state for all render targets back to unused
        for (auto& rt : m_renderTargets)
        {
            rt.second.m_viewUsage = e_renderTargetUsage::unused;
        }

        return view;
    }

    void Renderer::createViews()
    {
        auto ctx = gfx::getMainContext();

        if (m_cfg.HasMember("view_order"))
        {
            // delete old views
            for (auto& view : m_views)
            {
                if (view->m_uniformBuffer)
                {
                    ctx->destroyBuffer(view->m_uniformBuffer);
                    view->m_uniformBuffer = nullptr;
                }

                if (view->m_genericViewUniformBuffer)
                {
                    ctx->destroyBuffer(view->m_genericViewUniformBuffer);
                    view->m_genericViewUniformBuffer = nullptr;
                }

                if (view->m_uniformBufferMem.m_data)
                {
                    memFree(view->m_uniformBufferMem.m_data);
                    view->m_uniformBufferMem.m_data = nullptr;
                }

                if (view->m_pass)
                {
                    ctx->destroyRenderPass(view->m_pass);
                    view->m_pass = nullptr;
                }

                memDelete(view);
            }
            m_views.clear();

            // lazily delete all the created pipeline+pass combinations
            for (auto& p : m_pipelines)
            {
                gfx::RenderPipeline pipeline = p.second;
                ctx->destroyRenderPipeline(pipeline);
            }
            m_pipelines.clear();

            // reset render targets
            for (auto& rtit : m_renderTargets)
            {
                RenderTarget& rt = rtit.second;
                rt.m_usage = e_renderTargetUsage::unused;
                rt.m_viewUsage = e_renderTargetUsage::unused;
                
                // delete backup ping pong texture to free up mem
                if (rt.m_texture[1])
                {
                    ctx->destroyTexture(rt.m_texture[1]);
                    rt.m_texture[1] = nullptr;
                }
            }

            // create new views
            for (auto it = m_cfg["view_order"].Begin(); it != m_cfg["view_order"].End(); ++it)
            {
                const rapidjson::Value& val = *it;
                std::string viewName = val.GetString();

                ASSERT(m_cfg["views"].HasMember(viewName.c_str()));

                if (m_cfg["views"][viewName.c_str()].HasMember("setup_function"))
                {
                    ID sid = hash<std::string>(m_cfg["views"][viewName.c_str()]["setup_function"].GetString());
                    if (m_setupFuncs.find(sid) != m_setupFuncs.end())
                    {
                        ViewSetupFunc func = m_setupFuncs.at(sid);
                        func(m_views);
                    }
                    continue;
                }
                
                View* view = memNew<View>();
                view->m_name = viewName;
                *view = setupView(*view);
                                
                m_views.push_back(view);
            }
        }
    }

    void Renderer::update(float dt)
    {
    }

    void Renderer::render()
    {
        auto ctx = gfx::getMainContext();

        for (auto& v : m_views)
        {
            // get framebuffer and begin pass
            std::vector<gfx::Clear> clears;
            gfx::FramebufferDesc fbd = {};
            fbd.m_pass = v->m_pass;
            bool anyClears = false;
            for (size_t i = 0; i < v->m_numTargets; ++i)
            {
                fbd.m_textures.push_back(v->m_targetTextures[i].m_texture);
                fbd.m_loadOps.push_back(v->m_targetTextures[i].m_loadOp);
                if (v->m_targetTextures[i].m_loadOp == gfx::e_loadOpType::clear)
                {
                    anyClears = true;
                    clears.push_back(v->m_targetTextures[i].m_clear);
                }
                else
                {
                    clears.push_back({});
                }
                if (v->m_resolve)
                    clears.push_back({});
            }

            if (v->m_depthTarget.m_texture)
            {
                fbd.m_textures.push_back(v->m_depthTarget.m_texture);
                fbd.m_loadOps.push_back(v->m_depthTarget.m_loadOp);
                if (v->m_depthTarget.m_loadOp == gfx::e_loadOpType::clear)
                {
                    anyClears = true;
                    clears.push_back(v->m_depthTarget.m_clear);
                }
            }

            // clears need to be padded to the highest target index that is being cleared
            if (!anyClears)
                clears.clear();

            ctx->beginRenderPass(v->m_pass, ctx->obtainFramebuffer(fbd), clears.data(), clears.size());

            gfx::Viewport vp = v->m_viewport;
            ctx->setViewport(vp);
            ctx->setScissorRect({ (int)vp.x, (int)vp.y,(u32)vp.m_width, (u32)vp.m_height });

            // bind view pipeline if one is set
            if(v->m_pipeline != 0)
            {
                gfx::RenderPipeline pipeline = getPipelineForPass(v->m_pipeline, v->m_pass, v->m_pipelineVariation);
                ctx->bindRenderPipeline(pipeline);
            }

            // bind any view textures
            for(u32 i = 0; i < 4; ++i)
                if(v->m_textures.m_texture[i])
                    ctx->bindTexture(v->m_textures.m_texture[i], v->m_textures.m_sampler[i], i, gfx::e_bindFlags::uniform);

            // bind the generic view uniform buffer
            if (v->m_genericViewUniformBuffer)
            {
                ctx->bindBuffer(v->m_genericViewUniformBuffer, BUFFER_SLOT_GENERIC_VIEW, gfx::e_bindFlags::uniform);
            }

            // bind any view uniform data
            if(v->m_uniformBuffer)
            {
                ctx->bindBuffer(v->m_uniformBuffer, BUFFER_SLOT_VIEW_UNIFORMS, gfx::e_bindFlags::uniform);
            }

            v->m_renderFunction(v);

            // submit gpu work
            ctx->endRenderEncoding();

            // get sync
            
            std::vector<gfx::Semaphore> syncs;
            for(auto& dependancy : v->m_dependancies)
                syncs.push_back(dependancy->m_sync);

            // only enqueue one command buffer per commit for sync reasons
            ctx->enqueue();

            // commit command buffer, grabbing sync for work and passing in any dependant syncs
            v->m_sync = ctx->commit((syncs.size() > 0) ? syncs.data() : nullptr, syncs.size(), gfx::e_pipelineStageType::top);
        }
    }

    void Renderer::setMSAASamples(u32 samples)
    {
        auto ctx = gfx::getMainContext();

        m_msaaSamples = samples;

        for (auto& rt : m_renderTargets)
        {
            if (!rt.second.m_variableSamples)
                continue;

            // do for both ping pong textures
            for (size_t i = 0; i < 2; ++i)
            {
                if (!rt.second.m_texture[i])
                    continue;

                gfx::Texture oldTex = rt.second.m_texture[i];

                gfx::TextureDesc td = ctx->getTextureDesc(oldTex);
                td.m_samples = m_msaaSamples;

                gfx::Texture newTex = ctx->createTexture(td);
                ctx->destroyTexture(oldTex);

                rt.second.m_texture[i] = newTex;
            }
        }

        createViews();
    }

    void Renderer::resize(u32 width, u32 height)
    {
        auto ctx = gfx::getMainContext();

        std::set<View*> replaceOutputs, replaceInputs;
        
        // recreate any views that render to the back buffer
        for (auto& v : m_views)
        {
            for(size_t i = 0; i < v->m_numTargets; ++i)
            {
                if (v->m_targetTextures[i].m_texture == nullptr)
                {
                    replaceOutputs.insert(v);
                    break;
                }
            }
        }

        // create new render targets for output res targets
        std::unordered_map<gfx::Texture, gfx::Texture> textureReplacements;
        for (auto& rt : m_renderTargets)
        {
            if (!rt.second.m_outputRes)
                continue;

            // loop both possible ping-pong textures
            for (size_t i = 0; i < 2; ++i)
            {
                if (!rt.second.m_texture[i])
                    continue;

                gfx::Texture oldTex = rt.second.m_texture[i];

                gfx::TextureDesc td = ctx->getTextureDesc(oldTex);
                td.m_width = width;
                td.m_height = height;

                gfx::Texture newTex = ctx->createTexture(td);
                ctx->destroyTexture(oldTex);

                rt.second.m_texture[i] = newTex;

                textureReplacements[oldTex] = newTex;
                for (auto& v : m_views)
                {
                    for (size_t i = 0; i < v->m_numTargets; ++i)
                    {
                        if (v->m_targetTextures[i].m_texture == oldTex)
                        {
                            replaceOutputs.insert(v);
                            break;
                        }
                    }

                    if (v->m_depthTarget.m_texture == oldTex)
                    {
                        replaceOutputs.insert(v);
                    }

                    for (auto& t : v->m_textures.m_texture)
                    {
                        if (t == oldTex)
                        {
                            replaceInputs.insert(v);
                            break;
                        }
                    }
                }
            }
        }

        // recreate any view render passes if the output render targets have changed
        for (auto& v : replaceOutputs)
        {
            if (v->m_pass)
            {
                ctx->destroyRenderPass(v->m_pass);
            }

            for (size_t i = 0; i < v->m_numTargets; ++i)
            {
                gfx::Texture& t = v->m_targetTextures[i].m_texture;
                if (textureReplacements.find(t) != textureReplacements.end())
                {
                    t = textureReplacements.at(t);
                }
            }

            if (v->m_depthTarget.m_texture)
            {
                gfx::Texture& t = v->m_depthTarget.m_texture;
                if (textureReplacements.find(t) != textureReplacements.end())
                {
                    t = textureReplacements.at(t);
                }
            }

            v->m_pass = createViewRenderPass(*v);

            // update the generic view uniforms buffer
            GenericViewUniforms g = {};
            g.m_dimensions[0] = v->m_viewport.m_width;
            g.m_dimensions[1] = v->m_viewport.m_height;
            g.m_dimensions[2] = 1.0f / v->m_viewport.m_width; // texel width
            g.m_dimensions[3] = 1.0f / v->m_viewport.m_height; // texel height

            Mem mem = {};
            mem.m_size = sizeof(GenericViewUniforms);
            mem.m_data = &g;
            ctx->updateBuffer(v->m_genericViewUniformBuffer, mem);
        }

        // update texture pointers for view inputs
        for (auto& v : replaceInputs)
        {
            for (auto& t : v->m_textures.m_texture)
            {
                if (textureReplacements.find(t) != textureReplacements.end())
                {
                    t = textureReplacements.at(t);
                }
            }
        }

        // lazily delete all the created pipeline+pass combinations
        for (auto& p : m_pipelines)
        {
            gfx::RenderPipeline pipeline = p.second;
            ctx->destroyRenderPipeline(pipeline);
        }
        m_pipelines.clear();
    }

    View* Renderer::getView(ID idName, ID linkID)
    {
        for (auto& v : m_views)
            if (v->m_idName == idName && v->m_linkID == linkID)
                return v;
        return nullptr;
    }

    void Renderer::getViews(ID idName, ID linkID, std::vector<View*>& views)
    {
        for (auto& v : m_views)
            if (v->m_idName == idName && v->m_linkID == linkID)
                views.push_back(v);
    }

    gfx::Texture Renderer::getRenderTarget(ID id, RenderTargetUsage usage)
    {
        ASSERT(m_renderTargets.find(id) != m_renderTargets.end());

        RenderTarget& rt = m_renderTargets.at(id);
        ASSERT((usage != e_renderTargetUsage::read && usage != e_renderTargetUsage::resolve) || rt.m_usage != e_renderTargetUsage::unused); // can't read or resolve an unwritten texture
        if (usage == e_renderTargetUsage::write && rt.m_viewUsage == e_renderTargetUsage::read)
        {
            // ping pong the render target textures
            gfx::Texture prev = rt.m_texture[0];
            rt.m_texture[0] = rt.m_texture[1];
            rt.m_texture[1] = prev;
            if (!rt.m_texture[0])
            {
                auto ctx = gfx::getMainContext();

                // create a new render target texture
                gfx::TextureDesc td = ctx->getTextureDesc(rt.m_texture[1]);
                rt.m_texture[0] = ctx->createTexture(td);
            }
        }
        rt.m_usage = usage; 
        rt.m_viewUsage = usage;
        return rt.m_texture[0];
    }

    gfx::RenderPipeline Renderer::getPipelineForPass(ID id, gfx::RenderPass pass, ID variation)
    {
        ID fullID = (variation != 0) ? id ^ variation : id;
        ASSERT(m_partialPipelines.find(fullID) != m_partialPipelines.end());

        PartialPipeline pp = m_partialPipelines.at(fullID);

        // search existing pipelines for a matching pipeline/pass combo
        fullID = hashCombine(fullID, (intptr_t)pass);
        if (m_pipelines.find(fullID) == m_pipelines.end())
        {
            // create a new pipeline using this pipeline/pass combo
            auto ctx = gfx::getMainContext();

            // build multisample desc

            gfx::RenderPipelineDesc rpd = {};

            for (size_t i = 0; i < 6; ++i)
                if (pp.m_shaders[i])
                    rpd.m_shaders.push_back(pp.m_shaders[i]);

            gfx::MrtBlendDesc mrtBlend;
            mrtBlend.m_blends = pp.m_blends;

            rpd.m_vertexInput = pp.m_vertexInput;
            rpd.m_inputAssembly = pp.m_input;
            rpd.m_viewports = { { 0, 0, 1, 1, 0, 1 } };
            rpd.m_scissors = { {} };
            rpd.m_rasterizer = pp.m_raster;
            rpd.m_multisample = pp.m_multisample;
            rpd.m_blendDesc = mrtBlend;
            rpd.m_depthStencil = pp.m_depthStencil;
            rpd.m_dynamic = pp.m_dynamic;
            rpd.m_renderPass = pass;
            rpd.m_subpass = 0;

            m_pipelines[fullID] = ctx->createRenderPipeline(rpd);
        }
        return m_pipelines.at(fullID);
    }

    gfx::Sampler Renderer::getSampler(ID id)
    {
        ASSERT(m_samplers.find(id) != m_samplers.end());
        return m_samplers.at(id);
    }

    void Renderer::registerRenderFunc(ViewRenderFunc func, const std::string& name)
    {
        ID id = hash(name);
        m_renderFuncs[id] = func;
    }

    void Renderer::registerSetupFunc(ViewSetupFunc func, const std::string& name)
    {
        ID id = hash(name);
        m_setupFuncs[id] = func;
    }

    void Renderer::unregisterRenderFunc(const std::string& name)
    {
        ID id = hash(name);
        m_renderFuncs.erase(id);
    }

    void Renderer::unregisterSetupFunc(const std::string& name)
    {
        ID id = hash(name);
        m_setupFuncs.erase(id);
    }
}
