#include "resources.h"

#include "gfx.h"
#include "audio.h"

#include <stb_image.h>

namespace engine
{
    ResourceManager* ResourceManager::get()
    {
        static ResourceManager* s_self = nullptr;
        if (!s_self)
            s_self = memNew<ResourceManager>();
        return s_self;
    }

    //
    // textures
    //

    struct TextureAsyncData
    {
        Mem	m_textureMem;
        int	m_width = 0;
        int	m_height = 0;
    };

    template<>
    bool ResourceManager::loadAsync<TextureLoadDesc>(const TextureLoadDesc& desc, Resource* resource)
    {
        int width, height, channels;
        stbi_uc* pixels = stbi_load(desc.m_filename.c_str(), &width, &height, &channels, STBI_rgb_alpha);
        if (!pixels)
            return false;

        resource->m_asyncData = memNew<TextureAsyncData>();
        TextureAsyncData& asyncData = *(TextureAsyncData*)resource->m_asyncData;
        asyncData.m_textureMem.m_data = pixels;
        asyncData.m_textureMem.m_size = width * height * 4;
        asyncData.m_width = width;
        asyncData.m_height = height;

        return true;
    }

    template<>
    bool ResourceManager::load<TextureLoadDesc>(const TextureLoadDesc& desc, Resource* resource)
    {
        auto ctx = gfx::getMainContext();

        TextureAsyncData& async = *(TextureAsyncData*)(resource->m_asyncData);
        
        gfx::TextureDesc td = {};
        td.m_width = (u32)async.m_width;
        td.m_height = (u32)async.m_height;
        td.m_format.m_channels = gfx::e_formatChannelFlags::rgba;
        td.m_format.m_bitDepth = 8;
        td.m_format.m_type = gfx::e_formatType::unsignedNorm;
        td.m_usage = gfx::e_textureUsageFlags::shaderRead;
        td.m_data.m_data = async.m_textureMem.m_data;
        td.m_data.m_size = async.m_textureMem.m_size;

        // don't downsample too much for mips
        td.m_mips = std::max(1, std::min((int)log2(std::min(td.m_width, td.m_height)) - 6, (int)desc.m_mips));

        gfx::Texture tex = ctx->createTexture(td);
        if (td.m_mips > 1)
            ctx->generateMips(tex);
        resource->m_internal = tex;

        // clean up async data
        memDelete(async.m_textureMem.m_data);
        memDelete(resource->m_asyncData);

        return true;
    }

    template<>
    void ResourceManager::unload<gfx::Texture>(Resource* resource)
    {
        auto ctx = gfx::getMainContext();
        ctx->destroyTexture((gfx::Texture)resource->m_internal);
        resource->m_internal = nullptr;
    }

    template<>
    ID ResourceManager::getIDFromDesc<TextureLoadDesc>(const TextureLoadDesc& desc)
    {
        return hash(desc.m_filename);
    }

    //
    // audio
    //

    template<>
    bool ResourceManager::loadAsync<SoundLoadDesc>(const SoundLoadDesc& desc, Resource* resource)
    {
        void* adata = nullptr;
        size_t asize = 0;
        utils::readFileToBuffer(desc.m_filename.c_str(), adata, asize);
        if (!adata)
            return false;

        resource->m_asyncData = memNew<audio::SoundDesc>();
        audio::SoundDesc& sd = *(audio::SoundDesc*)resource->m_asyncData;
        sd.m_streamingType = desc.m_streamingType;
        sd.m_loopType = desc.m_loopType;
        sd.m_startTime = desc.m_startTime;
        sd.m_loopStartTime = desc.m_loopStartTime;
        sd.m_loopEndTime = desc.m_loopEndTime;

        // read the wav properties from the file header
        u8* header = (u8*)adata;
        header += 22;

        sd.m_channels = *(u16*)header;
        header += 2;

        sd.m_samplesPerSecond = *(u32*)header;
        header += 10;

        sd.m_bitsPerSample = *(u16*)header;
        header += 10;

        sd.m_data.m_size = asize - 44;
        sd.m_data.m_data = memAlloc(sd.m_data.m_size);
        memcpy(sd.m_data.m_data, header, sd.m_data.m_size);

        utils::freeFileBuffer(adata);

        return true;
    }

    template<>
    bool ResourceManager::load<SoundLoadDesc>(const SoundLoadDesc& desc, Resource* resource)
    {
        audio::AudioManager* am = audio::AudioManager::get();

        audio::SoundDesc& sd = *(audio::SoundDesc*)resource->m_asyncData;
        
        resource->m_internal = am->loadSound(sd);

        // clean up async data
        memDelete(sd.m_data.m_data);
        memDelete(resource->m_asyncData);

        return true;
    }

    template<>
    void ResourceManager::unload<audio::Sound>(Resource* resource)
    {
        audio::AudioManager* am = audio::AudioManager::get();
        am->unloadSound((audio::Sound)resource->m_internal);
        resource->m_internal = nullptr;        
    }

    template<>
    ID ResourceManager::getIDFromDesc<SoundLoadDesc>(const SoundLoadDesc& desc)
    {
        return hash(desc.m_filename);
    }

    //
    // fonts
    //
    template<>
    bool ResourceManager::loadAsync<gfx::FontDesc>(const gfx::FontDesc& desc, Resource* resource)
    {
        gfx::Font* font = gfx::createFont();
        font->load(desc);
        resource->m_internal = font;
        return true;
    }

    template<>
    bool ResourceManager::load<gfx::FontDesc>(const gfx::FontDesc& desc, Resource* resource)
    {       
        gfx::Font* font = (gfx::Font*)resource->m_internal;
        font->createTexture();
        return true;
    }

    template<>
    void ResourceManager::unload<gfx::Font*>(Resource* resource)
    {
        gfx::Font* font = (gfx::Font*)resource->m_internal;
        font->unload();
        gfx::destroyFont(font);
        resource->m_internal = nullptr;
    }

    template<>
    ID ResourceManager::getIDFromDesc<gfx::FontDesc>(const gfx::FontDesc& desc)
    {
        ID id = hash(desc.m_filename);
        id = hashCombine(id, desc.m_scale);
        return id;
    }
}
