#include "save.h"

#include "platform.h"
#include <atomic>
#include <pystring.h>

namespace engine
{
    SaveManager* SaveManager::get()
    {
        static SaveManager* s_self = nullptr;
        if (!s_self)
            s_self = memNew<SaveManager>();
        return s_self;
    }

	void* saveThread(void* data);

	namespace e_saveThreadRequestType
	{
		enum saveThreadRequestType
		{
			save,
			load,
			shutdown
		};
	}
	typedef e_saveThreadRequestType::saveThreadRequestType SaveThreadRequestType;

	struct SaveThreadRequest
	{
		SaveThreadRequestType	m_type = e_saveThreadRequestType::save;
	};

	class SaveManagerInternal : public SaveManager 
	{
	public:
		RingBuffer<SaveThreadRequest>	m_threadRequests;
		std::atomic<SaveState>			m_saveThreadState = { e_saveState::waiting };
		std::string						m_payload = "";
		std::string						m_location = "data/saves";

		void setup() override
		{
			utils::createThread(saveThread, this, utils::e_threadFlags::detached);
		}

		void shutdown() override
		{
			SaveThreadRequest r = {};
			r.m_type = e_saveThreadRequestType::shutdown;
			m_threadRequests.put(r);
		}

		void update(float dt) override
		{
			SaveState state = m_saveThreadState.load();

			if (state == e_saveState::saved)
			{
				m_saveThreadState = e_saveState::waiting;
			}
			else if (state == e_saveState::loaded)
			{
				m_saveThreadState = e_saveState::waiting;
			}
		}

		std::string dumpGameState(const GameState& s, int indent)
		{
			std::string out = "";

			std::string indentStr = "";
			for (int i = 0; i < indent; ++i)
				indentStr += '\t';

			switch (s.type())
			{
			case e_gameStateType::object:
			{
				out = out + "\n"+ indentStr + "{";
				for (auto it = s.begin(); it != s.end(); ++it)
				{
					std::string name = (*it).name();
					out = out + "\n";
					out = out + indentStr + "\t" + name + ":";
					out = out + dumpGameState(*it, indent + 1);
				}
				out = out + "\n" + indentStr + "}";
			}
			break;
			case e_gameStateType::array:
			{
				size_t index = 0;
				out = out + "\n" + indentStr + "[";
				for (auto it = s.begin(); it != s.end(); ++it)
				{
					out = out + "\n";
					out = out + dumpGameState(*it, indent + 1);
					index++;
				}
				out = out + "\n" + indentStr + "]";
			}
			break;
			case e_gameStateType::value:
			{
				size_t count = s.count();
				if (s.name().empty())
					out = out + indentStr;
				switch (s.valueType())
				{
				case e_gameStateValueType::uint:
				{
					out = out + " " + "(uint)" + std::to_string(s.value<u64>());
				}
				break;
				case e_gameStateValueType::sint:
				{
					out = out + " " + "(int)" + std::to_string(s.value<int>());
					
				}
				break;
				case e_gameStateValueType::sfloat:
				{
					out = out + " " + "(float)" + std::to_string(s.value<float>());
				}
				break;
				case e_gameStateValueType::string:
				{
					out = out + "(string)" + s.string();
				}
				break;
				}
			}
			break;
			case e_gameStateType::unset:
				break;
			}

			return out;
		}

		void save(const GameState& s) override
		{
			// convert game state to save payload
			m_payload = pystring::strip(dumpGameState(s, 0));

			SaveThreadRequest r = {};
			r.m_type = e_saveThreadRequestType::save;
			m_threadRequests.put(r);
		}

		void parseValue(const std::string& str, GameState& s)
		{
			// determine type
			if (pystring::find(str, "(uint)") != -1)
			{
				s.setUints(std::stoul(pystring::replace(str, "(uint)", "")));
			}
			else if (pystring::find(str, "(int)") != -1)
			{
				s.setInts(std::stoi(pystring::replace(str, "(int)", "")));
			}
			else if (pystring::find(str, "(float)") != -1)
			{
				s.setFloats(std::stof(pystring::replace(str, "(float)", "")));
			}
			else if (pystring::find(str, "(string)") != -1)
			{
				s.setString(pystring::replace(str, "(string)", ""));
			}
		}

		void parseGameState(const std::vector<std::string>& lines, GameState& s, size_t& currentLine)
		{
			GameStateType type = e_gameStateType::unset;

			size_t arrayIndex = 0;
			while(currentLine < lines.size())
			{
				const std::string& l = pystring::strip(lines[currentLine]);
				if (l.empty())
				{
					++currentLine;
					continue;
				}

				if (type == e_gameStateType::array)
				{
					if (pystring::startswith(l, "]"))
						break;

					GameState arrMember = {};
					s.push(arrMember);

					parseGameState(lines, s[arrayIndex++], currentLine);
				}
				else if (type == e_gameStateType::object)
				{
					if (pystring::startswith(l, "}"))
						break;

					std::vector<std::string> keyValue;
					pystring::split(l, keyValue, ":");

					for (size_t i = 0; i < keyValue.size(); ++i)
					{
						keyValue[i] = pystring::strip(keyValue[i]);
						if (keyValue[i].empty())
							keyValue.erase(keyValue.begin() + i);
					}

					if (keyValue.size() == 1)
					{
						// child is an object itself
						++currentLine;
						parseGameState(lines, s[keyValue[0]], currentLine);
					}
					else
					{
						// child is a raw value
						GameState& val = s[keyValue[0]];
						parseValue(keyValue[1], val);
					}
				}
				else if (pystring::startswith(l, "{"))
				{
					type = e_gameStateType::object;
				}
				else if (pystring::startswith(l, "["))
				{
					type = e_gameStateType::array;
				}
				else
				{
					parseValue(l, s);
					break;
				}
				currentLine++;
			}
		}

		void load(GameState& s) override
		{
			SaveThreadRequest r = {};
			r.m_type = e_saveThreadRequestType::load;
			m_threadRequests.put(r);

			while (m_saveThreadState.load() != e_saveState::loaded)
			{
				utils::sleepMs(1);
			}

			// clear existing saved state
			s.clear();

			// convert save payload to game state
			std::vector<std::string> lines;
			pystring::splitlines(m_payload, lines);
			size_t currentLine = 0;
			parseGameState(lines, s, currentLine);
		}

		void setSaveLocation(const std::string& location) override
		{
			m_location = location;
		}
	};

	void SaveManager::setup()
	{
		m_internal = memNew<SaveManagerInternal>();
		m_internal->setup();
	}

	void* saveThread(void* data)
	{
		SaveManagerInternal* sm = (SaveManagerInternal*)data;

		bool shutdown = false;
		while (!shutdown)
		{
			if (sm->m_saveThreadState.load() == e_saveState::waiting)
			{
				SaveThreadRequest request = {};
				if (sm->m_threadRequests.get(request))
				{
					switch (request.m_type)
					{
					case e_saveThreadRequestType::save:
					{
						sm->m_saveThreadState = e_saveState::saving;

						std::string payload = sm->m_payload;

						std::string location = sm->m_location;
						std::string savePath = pystring::os::path::join(location, std::string("001.save"));
						utils::createDirectory(location.c_str());
						utils::writeFileFromBuffer(savePath.c_str(), (void*)payload.c_str(), payload.size());				

						sm->m_saveThreadState = e_saveState::saved;
					}
					break;
					case e_saveThreadRequestType::load:
					{
						sm->m_saveThreadState = e_saveState::loading;

						std::string loadPath = pystring::os::path::join(sm->m_location, std::string("001.save"));

						sm->m_payload = "";
						if (utils::fileExists(loadPath.c_str()))
						{
							void* buf = nullptr;
							size_t size = 0;
							utils::readFileToBuffer(loadPath.c_str(), buf, size);
							sm->m_payload = std::string((char*)buf, size);
							utils::freeFileBuffer(buf);
						}

						sm->m_saveThreadState = e_saveState::loaded;
					}
					break;
					case e_saveThreadRequestType::shutdown:
					{
						shutdown = true;
					}
					break;
					}
				}
			}
			utils::sleepMs(1);
		}

		return nullptr;
	}
}