#include "scene.h"

#include "platform.h"
#include "gfx.h"
#include "renderer.h"
#include "model.h"
#include "json.h"
#include "engine.h"

#include <algorithm>
#include <pystring.h>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtx/rotate_vector.hpp>
#include <glm/gtx/transform.hpp>

namespace
{

    const char* k_audioStreamingTypeNames[] = {
        "nonStreaming",
        "streaming"
    };

    const char* k_audioLoopTypeNames[] = {
        "none",
        "looping"
    };
    
    const char* k_layerFlagNames[] = {
        "normalised"
    };

    namespace e_cullingType
    {
        enum cullingType
        {
            none,
            sphere,
            frustum
        };
    }
    typedef e_cullingType::cullingType CullingType;

    struct LayerCameraData
    {
        glm::mat4 m_inverseVP = glm::mat4(1.0f);
        glm::vec4 m_eye = glm::vec4(0.0f);
    };
}

namespace engine
{
    rapidjson::Value mergeEntityAndPrefab(const rapidjson::Value& entity, const rapidjson::Value& prefab, rapidjson::Document::AllocatorType& allocator)
    {
        rapidjson::Value result;
        result.CopyFrom(prefab, allocator, true);
        if (!mergeObjects(result, entity, allocator))
            ASSERT(false);
        return result;
    }

    rapidjson::Value loadPrefabs(const rapidjson::Value& prefabsJson, rapidjson::Document::AllocatorType& alloc)
    {
        rapidjson::Value prefabs;
        prefabs.SetObject();
        for (auto p = prefabsJson.Begin(); p != prefabsJson.End(); ++p)
        {
            std::string prefabFilename = p->GetString();
            std::string prefabName = pystring::replace(pystring::os::path::basename(prefabFilename), ".prefab", "");

            rapidjson::Value name(prefabName.c_str(), alloc);
            prefabs.AddMember(name, "", alloc);
            prefabs[prefabName.c_str()].SetObject();
            rapidjson::Value& prefabCollection = prefabs[prefabName.c_str()];

            // load the prefabs file
            void* pbuf = nullptr;
            size_t plen = 0;
            utils::readFileToBuffer(prefabFilename.c_str(), pbuf, plen);
            ASSERT(pbuf != nullptr);
            rapidjson::Document pd;
            bool result = pd.Parse<rapidjson::kParseStopWhenDoneFlag>((const char*)pbuf).HasParseError();
            if (result)
            {
                auto error = pd.GetParseError();
                printf("Error!");
                ASSERT(false);
            }
            utils::freeFileBuffer(pbuf);

            // read in entity prefabs
            if (pd.FindMember("entities") != pd.MemberEnd())
            {
                prefabCollection.AddMember("entities", "", alloc);
                prefabCollection["entities"].SetObject();

                rapidjson::Value& prefabEntities = prefabCollection["entities"];

                for (auto pe = pd["entities"].MemberBegin(); pe != pd["entities"].MemberEnd(); ++pe)
                {
                    rapidjson::Value name;
                    name.CopyFrom(pe->name, alloc);
                    rapidjson::Value value;
                    value.CopyFrom(pe->value, alloc);
                    prefabEntities.AddMember(name, value, alloc);
                }
            }

            // read in layer prefabs
            if (pd.HasMember("layer_order"))
            {
                prefabCollection.AddMember("layer_order", "", alloc);
                prefabCollection["layer_order"].SetArray();

                for (size_t i = 0; i < pd["layer_order"].Size(); ++i)
                {
                    prefabCollection["layer_order"].PushBack(pd["layer_order"][i], alloc);
                }
            }

            if (pd.HasMember("layers"))
            {
                prefabCollection.AddMember("layers", "", alloc);
                prefabCollection["layers"].SetObject();

                rapidjson::Value& prefabLayers = prefabCollection["layers"];

                for (auto pl = pd["layers"].MemberBegin(); pl != pd["layers"].MemberEnd(); ++pl)
                {
                    rapidjson::Value name;
                    name.CopyFrom(pl->name, alloc);
                    rapidjson::Value value;
                    value.CopyFrom(pl->value, alloc);
                    prefabLayers.AddMember(name, value, alloc);
                }
            }
        }

        return prefabs;
    }

    Layer loadLayer(const rapidjson::Value& layerJson)
    {
        Layer layer = {};
        layer.m_cameraEntity = 0;
        if (layerJson["camera_entity"].IsString())
        {
            std::string cameraName = getJsonValueOrDefault<const char*>(layerJson, "camera_entity", "");
            layer.m_cameraID = hash(cameraName);
        }
        else
        {
            layer.m_cameraEntity = getJsonValueOrDefault(layerJson, "camera_entity", 0);
        }
        if (layerJson.HasMember("flags"))
        {
            const rapidjson::Value& flagsJson = layerJson["flags"].GetArray();
            for (size_t i = 0; i < flagsJson.Size(); ++i)
            {
                layer.m_flags |= enumNameToFlagValue(k_layerFlagNames, ARRAY_SIZE(k_layerFlagNames), flagsJson[i].GetString());
            }
        }
        layer.m_width = utils::getWindowWidth();
        layer.m_height = utils::getWindowHeight();

        if (layerJson.HasMember("post_processing"))
        {
            const rapidjson::Value& postProcessingJson = layerJson["post_processing"].GetArray();
            for (size_t i = 0; i < postProcessingJson.Size(); ++i)
            {
                const rapidjson::Value& postProcessJson = postProcessingJson[i];

                bool preBlend = getJsonValueOrDefault(postProcessJson, "pre_blend", false);

                View view = {};
                view.m_name = preBlend ? "layer_post_process_separated" : "layer_post_process";
                view.m_pipeline = hash<std::string>(getJsonValueOrDefault<const char*>(postProcessJson, "pipeline", ""));

                // uniform buffer members
                if (postProcessJson.HasMember("uniforms"))
                {
                    std::vector<float> floats;
                    for (size_t i = 0; i < postProcessJson["uniforms"].Size(); ++i)
                    {
                        const rapidjson::Value& uniform = postProcessJson["uniforms"][i];
                        size_t count = 1;
                        if (uniform.HasMember("dimensions"))
                        {
                            if (uniform["dimensions"].IsArray())
                            {
                                for (size_t j = 0; j < uniform["dimensions"].Size(); ++j)
                                {
                                    count *= (size_t)uniform["dimensions"][j].GetUint();
                                }
                            }
                            else
                            {
                                count = (size_t)(uniform["dimensions"].GetUint());
                            }
                        }
                        else
                        {
                            count = uniform["value"].Size();
                        }

                        bool hasValues = uniform.HasMember("value");
                        for (size_t j = 0; j < count; ++j)
                        {
                            if (hasValues)
                                floats.push_back(uniform["value"][j].GetFloat());
                            else
                                floats.push_back(0.0f);
                        }
                    }
                    size_t uniformSize = floats.size() * sizeof(float);
                    view.m_uniformBufferMem.m_size = uniformSize;
                    view.m_uniformBufferMem.m_data = memAlloc(uniformSize);
                    memcpy(view.m_uniformBufferMem.m_data, floats.data(), uniformSize);
                }

                // textures
                if (postProcessJson.HasMember("textures"))
                {
                    size_t textureCount = postProcessJson["textures"].Size();
                    for (size_t i = 0; i < textureCount; ++i)
                    {
                        std::string textureName = postProcessJson["textures"][i]["texture"].GetString();
                        std::string samplerName = postProcessJson["textures"][i]["sampler"].GetString();
                        view.m_textures.m_textureID[i + 1] = hash(textureName);
                        view.m_textures.m_samplerID[i + 1] = hash(samplerName);
                    }
                }

                if (preBlend)
                    layer.m_preBlendPostProcessViews.push_back(view);
                else
                    layer.m_postBlendPostProcessViews.push_back(view);

            }
        }

        std::string blendName = getJsonValueOrDefault<const char*>(layerJson, "blend_name", "");
        if (!blendName.empty())
            layer.m_blendID = hash(blendName);

        layer.m_deferredRender = getJsonValueOrDefault(layerJson, "deferred_render", false);

        if (layerJson.HasMember("ambient_light"))
        {
            getJsonArrayOrDefault(layerJson, "ambient_light", &layer.m_ambientLight.x, 3, 0.0f);
            layer.m_ambientLight.a = 1.0f;
        }

        return layer;
    }

    TriggerLogic createTriggerLogic(const rapidjson::Value& triggerJson, Engine* engine, Scene* s, size_t index = -1)
    {
        TriggerLogic tl = {};

        bool fireOnce = getJsonValueOrDefault(triggerJson, "fire_once", false);
        tl.m_fire = (fireOnce) ? e_triggerLogicFireType::fireOnce : e_triggerLogicFireType::canFire;
        tl.m_simultaneousActions = getJsonValueOrDefault(triggerJson, "simultaneous_actions", true);

        if (triggerJson.FindMember("actions") != triggerJson.MemberEnd())
        {
            const rapidjson::Value& actions = triggerJson["actions"].GetArray();
            size_t numActions = (size_t)actions.Size();
            for (size_t j = 0; j < numActions; ++j)
            {
                std::string name = actions[j]["name"].GetString();
                tl.m_actions.push_back(engine->createTriggerAction(hash(name), actions[j], s, index));
            }
        }

        if (triggerJson.FindMember("else_actions") != triggerJson.MemberEnd())
        {
            const rapidjson::Value& actions = triggerJson["else_actions"].GetArray();
            size_t numActions = (size_t)actions.Size();
            for (size_t j = 0; j < numActions; ++j)
            {
                std::string name = actions[j]["name"].GetString();
                tl.m_elseActions.push_back(engine->createTriggerAction(hash(name), actions[j], s, index));
            }
        }

        // don't create conditions if trigger won't do anything
        if ((!tl.m_actions.empty() || !tl.m_elseActions.empty()) && triggerJson.FindMember("conditions") != triggerJson.MemberEnd())
        {
            const rapidjson::Value& conditions = triggerJson["conditions"].GetArray();
            size_t numConditions = (size_t)conditions.Size();
            for (size_t j = 0; j < numConditions; ++j)
            {
                std::string name = conditions[j]["name"].GetString();
                tl.m_conditions.push_back(engine->createTriggerCondition(hash(name), conditions[j], s, index));
            }
        }


        return tl;
    }

    void Scene::load(const char* filename)
    {
        if (m_flags & e_sceneFlags::loading)
            return;

        m_flags |= e_sceneFlags::loading;

        if (!utils::fileExists(filename))
        {
            m_flags &= ~e_sceneFlags::loading;
            return;
        }

        void* buf = nullptr;
        size_t len = 0;
        utils::readFileToBuffer(filename, buf, len);
        rapidjson::Document d;
        bool result = d.Parse<rapidjson::kParseStopWhenDoneFlag>((const char*)buf).HasParseError();
        if (result)
        {
            auto error = d.GetParseError();
            printf("Error!");
            m_flags &= ~e_sceneFlags::loading;
            ASSERT(false);
            return;
        }
        utils::freeFileBuffer(buf);

        Engine* engine = Engine::get();
        ResourceManager* rm = ResourceManager::get();

        // load prefabs for this scene into memory
        rapidjson::Value prefabs;
        prefabs.SetObject();
        if (d.HasMember("prefabs"))
        {
            prefabs = loadPrefabs(d["prefabs"], d.GetAllocator());
        }

        // create layers
        bool findDefaultCamera = false;
        if (d.HasMember("layers"))
        {
            if (d.FindMember("layer_order") != d.MemberEnd())
            {
                const rapidjson::Value& loj = d["layer_order"].GetArray();
                for (size_t i = 0; i < loj.Size(); ++i)
                {
                    m_layerOrder.push_back(hash<std::string>(loj[i].GetString()));
                }
            }
            bool orderSpecified = !m_layerOrder.empty();

            const rapidjson::Value& lj = d["layers"];
            for (auto& mit = lj.MemberBegin(); mit != lj.MemberEnd(); ++mit)
            {
                const rapidjson::Value& layerJson = mit->value;
                std::string layerName = mit->name.GetString();
                ID layerID = hash(layerName);             

                m_layers[layerID] = loadLayer(layerJson);
                if (!orderSpecified)
                {
                    m_layerOrder.push_back(layerID);
                }
            }
        }
        else
        {
            // create a single default layer that every entity will be added to
            ID defaultLayerID = hash<std::string>("default");

            Layer defaultLayer = {};
            defaultLayer.m_width = utils::getWindowWidth();
            defaultLayer.m_height = utils::getWindowHeight();
            m_layers[defaultLayerID] = defaultLayer;
            m_layerOrder.push_back(defaultLayerID);
            findDefaultCamera = true;
        }

        // create entities
        rapidjson::Value& entities = d["entities"].GetArray();
        size_t numEntities = (size_t)entities.Size();

        m_world = memNew<World>();
        m_world->grow(numEntities);

        m_animManager = memNew<AnimationManager>();
        m_animManager->setup(m_world);

        auto& w = *m_world;

        for(size_t i = 0; i < numEntities; ++i)
        {
            rapidjson::Value& entity = entities[i];

            std::string prefabName = getJsonValueOrDefault<const char*>(entity, "prefab", "");
            if (!prefabName.empty())
            {
                std::vector<std::string> prefabNames;
                pystring::split(prefabName, prefabNames, "/");
                const rapidjson::Value& prefabCollectionJson = prefabs[prefabNames[0].c_str()];
                const rapidjson::Value& prefabJson = prefabCollectionJson["entities"][prefabNames[1].c_str()];
                entity = mergeEntityAndPrefab(entities[i], prefabJson, d.GetAllocator());
            }

            size_t e = m_world->newEntity();
            loadEntity(e, entity);
        }

        // find first camera entity if a default camera is needed
        if (findDefaultCamera)
        {
            for (size_t i = 0, n = w.size(); i < n; ++i)
            {
                if (!(w.m_components[i] & e_cmp::camera))
                    continue;
                
                m_layers[hash<std::string>("default")].m_cameraEntity = i;
                break;
            }
        }

        // now entities are loaded find layer cameras
        for (auto& l : m_layers)
        {
            if (l.second.m_cameraID == 0)
                continue;

            for (size_t k = 0, n = w.size(); k < n; ++k)
            {
                if (!(w.m_components[k] & e_cmp::camera))
                    continue;

                if (w.m_ulid[k] == (ULID)l.second.m_cameraID)
                {
                    l.second.m_cameraEntity = k;
                    break;
                }
            }
        }

        // create triggers
        if (d.FindMember("triggers") != d.MemberEnd())
        {
            const rapidjson::Value& triggers = d["triggers"].GetArray();
            size_t numTriggers = (size_t)triggers.Size();
            m_triggers.resize(numTriggers);

            for(size_t i = 0; i < numTriggers; ++i)
            {
                const rapidjson::Value& t = triggers[i];
                m_triggers.push_back(createTriggerLogic(t, engine, this));
            }
        }

        m_flags &= ~e_sceneFlags::loading;
        m_flags |= e_sceneFlags::loaded;
    }

    void Scene::loadEntity(size_t e, const rapidjson::Value& entity)
    {
        World& w = *m_world;
        Engine* engine = Engine::get();
        ResourceManager* rm = ResourceManager::get();
        
        // don't overwrite layer
        if (w.m_layer[e] == 0)
        {
            std::string layerName = getJsonValueOrDefault<const char*>(entity, "layer", "default");
            w.m_layer[e] = hash(layerName);
        }

        if (entity.HasMember("flags"))
        {
            const rapidjson::Value& flags = entity["flags"];
            for (size_t j = 0; j < flags.Size(); ++j)
            {
                w.m_entityFlags[e] |= e_entityFlags::nameToValue(flags[j].GetString());
            }
        }

        if (entity.HasMember("name"))
        {
            std::string name = getJsonValueOrDefault<const char*>(entity, "name", "");
            w.m_ulid[e] = (ULID)hash(name);
        }
        else
        {
            std::string name = "entity_" + std::to_string(e);
            w.m_ulid[e] = (ULID)hash(name);
        }

        if (w.m_parent[e] == (size_t)-1 && entity.HasMember("parent"))
        {
            std::string parentName = getJsonValueOrDefault<const char*>(entity, "parent", "");
            ULID parentULID = (ULID)hash(parentName);
            for (size_t i = 0; i < e; ++i)
            {
                if (w.m_ulid[i] == parentULID)
                {
                    w.m_parent[e] = i;
                    break;
                }
            }
        }

        if (entity.HasMember("triggers"))
        {
            std::vector<TriggerLogic> entityTriggers;
            for (auto triggerIt = entity["triggers"].MemberBegin(); triggerIt != entity["triggers"].MemberEnd(); ++triggerIt)
            {
                TriggerLogic tl = createTriggerLogic(triggerIt->value, engine, this, e);
                // ignore triggers that don't do anything
                if (!tl.m_actions.empty() || !tl.m_elseActions.empty())
                    entityTriggers.push_back(tl);
            }
            if (!entityTriggers.empty())
            {
                w.m_triggerID[e] = (ID)m_entityTriggers.size();
                m_entityTriggers.push_back(entityTriggers);
                w.m_components[e] |= e_cmp::triggers;
            }
        }

        if (entity.HasMember("transform"))
        {
            const rapidjson::Value& transform = entity["transform"];
            getJsonArrayOrDefault<float>(transform, "pos", &w.m_pos[e].x, 3, 0.0f);
            getJsonArrayOrDefault<float>(transform, "scale", &w.m_scale[e].x, 3, 1.0f);
            getJsonArrayOrDefault<float>(transform, "rot", &w.m_rot[e].x, 3, 0.0f);
            w.m_components[e] |= e_cmp::transform;

            w.m_localAabb[e] = {};
        }

        if (entity.HasMember("pipeline"))
        {
            std::string pipeline = getJsonValueOrDefault<const char*>(entity, "pipeline", "");
            w.m_pipeline[e] = hash(pipeline);
        }

        if (entity.HasMember("geometry"))
        {
            const rapidjson::Value& geometry = entity["geometry"];

            ModelLoadDesc ld = {};

            std::string gType = getJsonValueOrDefault<const char*>(geometry, "type", "file");
            if (gType == "plane")
            {
                ld.m_genFunc = generatePlaneGeometry;

                ld.m_genData.m_data = memNew<PlaneGenData>();
                ld.m_genData.m_size = sizeof(PlaneGenData);

                PlaneGenData& desc = *(PlaneGenData*)ld.m_genData.m_data;
                desc = {};
                getJsonArrayOrDefault<float>(geometry, "normal", desc.m_normal, 3, 1.0f);
                getJsonArrayOrDefault<float>(geometry, "offset", desc.m_offset, 3, 0.0f);
                desc.m_subdivisions = getJsonValueOrDefault<float>(geometry, "subdivisions", 1.0f);

            }
            else if (gType == "cube")
            {
                ld.m_genFunc = generateCubeGeometry;
                ld.m_genData.m_data = nullptr;
                ld.m_genData.m_size = 0;
            }
            else if (gType == "file")
            {
                ld.m_filename = getJsonValueOrDefault<const char*>(geometry, "filepath", "");
            }

            ModelRef ref = rm->load<ModelData, ModelLoadDesc>(ld, e_resourceLoadFlags::async);

            bool createChildren = getJsonValueOrDefault(geometry, "create_children", false);
            if (createChildren)
            {
                ModelData mdl = ref.getTypedResource();
                for (size_t i = 0; i < mdl.m_numUniques; ++i)
                {
                    size_t c = w.newEntity();
                    w.m_parent[c] = e;
                    w.m_layer[c] = w.m_layer[e];
                    w.m_components[c] = (e_cmp::transform | e_cmp::modelRef | e_cmp::renderable);
                    w.m_pos[c] = glm::vec3(0.0f);
                    w.m_rot[c] = glm::vec3(0.0f);
                    w.m_scale[c] = glm::vec3(1.0f);
                    w.m_modelRef[c] = ref;
                    w.m_meshIndex[c] = i;
                    w.m_materialIndex[c] = mdl.m_materialIndices[i];
                    w.m_pipeline[c] = w.m_pipeline[e];
                }
                w.m_pipeline[e] = 0;
            }
            else
            {
                w.m_components[e] |= (e_cmp::modelRef | e_cmp::renderable);
                w.m_modelRef[e] = ref;
            }

        }

        if (entity.HasMember("texture"))
        {
            TextureLoadDesc td = {};
            td.m_filename = getJsonValueOrDefault<const char*>(entity["texture"], "filename", "");

            w.m_textureRef[e] = rm->load<gfx::Texture, TextureLoadDesc>(td, e_resourceLoadFlags::async);
            w.m_components[e] |= e_cmp::textureRef;
        }

        if (entity.HasMember("camera"))
        {
            const rapidjson::Value& camera = entity["camera"];

            w.m_camera[e] = memNew<Camera>();
            w.m_camera[e]->m_type = e_cameraType::nameToValue(getJsonValueOrDefault<const char*>(camera, "type", "orthographic"));
            w.m_camera[e]->m_near = getJsonValueOrDefault<float>(camera, "near", 0.0f);
            w.m_camera[e]->m_far = getJsonValueOrDefault<float>(camera, "far", 1.0f);

            if (w.m_camera[e]->m_type == e_cameraType::orthographic)
            {
                w.m_camera[e]->m_ortho.m_matchOutput = getJsonValueOrDefault<bool>(camera, "match_output", true);
                w.m_camera[e]->m_ortho.m_left = getJsonValueOrDefault<float>(camera, "left", -1.0f);
                w.m_camera[e]->m_ortho.m_right = getJsonValueOrDefault<float>(camera, "right", 1.0f);
                w.m_camera[e]->m_ortho.m_bottom = getJsonValueOrDefault<float>(camera, "bottom", 1.0f);
                w.m_camera[e]->m_ortho.m_top = getJsonValueOrDefault<float>(camera, "top", -1.0f);
                w.m_camera[e]->m_ortho.m_invertY = getJsonValueOrDefault<bool>(camera, "invert_y", false);
                w.m_camera[e]->m_cullingFrustum.m_nearPlane = w.m_camera[e]->m_near;
                w.m_camera[e]->m_cullingFrustum.m_farPlane = w.m_camera[e]->m_far;
                w.m_camera[e]->m_cullingFrustum.m_nearRight = w.m_camera[e]->m_ortho.m_right;
                w.m_camera[e]->m_cullingFrustum.m_nearTop = w.m_camera[e]->m_ortho.m_top;
            }
            else if (w.m_camera[e]->m_type == e_cameraType::perspective)
            {
                w.m_camera[e]->m_perspective.m_aspect = getJsonValueOrDefault<float>(camera, "aspect", 1.778f);
                w.m_camera[e]->m_perspective.m_fov = getJsonValueOrDefault<float>(camera, "fov", 90.0f);
                w.m_camera[e]->m_cullingFrustum = maths::createFrustum( w.m_camera[e]->m_near, 
                                                                        w.m_camera[e]->m_far, 
                                                                        glm::radians(w.m_camera[e]->m_perspective.m_fov), 
                                                                        w.m_camera[e]->m_perspective.m_aspect);
            }

            w.m_localAabb[e].m_min = glm::vec3(0.0f, 0.0f, 0.0f);
            w.m_localAabb[e].m_max = glm::vec3(0.0f, 0.0f, 0.0f);

            w.m_components[e] |= e_cmp::camera;

            w.m_entityFlags[e] |= e_entityFlags::passthroughCursor;
        }

        glm::vec4 colour = glm::vec4(1.0f);
        getJsonArrayOrDefault(entity, "colour", &colour.x, 4, 1.0f);
        w.m_uniformData[e].m_userData0 = colour;

        if (entity.HasMember("sound"))
        {
            const rapidjson::Value& sj = entity["sound"];

            SoundLoadDesc sd = {};
            sd.m_filename = getJsonValueOrDefault(sj, "filename", "");
            sd.m_startTime = getJsonValueOrDefault(sj, "start_time", 0.0f);
            sd.m_loopStartTime = getJsonValueOrDefault(sj, "loop_start_time", 0.0f);
            sd.m_loopEndTime = getJsonValueOrDefault(sj, "loop_end_time", -1.0f);
            sd.m_streamingType = enumNameToValue<audio::AudioStreamingType>(k_audioStreamingTypeNames, ARRAY_SIZE(k_audioStreamingTypeNames), getJsonValueOrDefault<const char*>(sj, "streaming_type", "nonStreaming"));
            sd.m_loopType = enumNameToValue<audio::AudioLoopType>(k_audioLoopTypeNames, ARRAY_SIZE(k_audioLoopTypeNames), getJsonValueOrDefault<const char*>(sj, "loop_type", "none"));

            w.m_soundRef[e] = rm->load<audio::Sound, SoundLoadDesc>(sd, e_resourceLoadFlags::async);
            w.m_components[e] |= e_cmp::soundRef;
        }

        if (entity.HasMember("text"))
        {
            const rapidjson::Value& tj = entity["text"];

            gfx::FontDesc fd = {};
            fd.m_filename = getJsonValueOrDefault(tj, "font_filename", "");
            fd.m_scale = getJsonValueOrDefault(tj, "font_scale", 32.0f);
            fd.m_codepoints = { {' ', '~'} };
            w.m_fontRef[e] = rm->load<gfx::Font*, gfx::FontDesc>(fd, e_resourceLoadFlags::async);

            w.m_text[e] = memNew<Text>();
            w.m_text[e]->m_text = getJsonValueOrDefault<const char*>(tj, "text", "");

            float alphaCutoff = getJsonValueOrDefault(tj, "alpha_cutoff", 0.0f);

            w.m_uniformData[e].m_userData0.w = alphaCutoff;

            w.m_components[e] |= (e_cmp::fontRef | e_cmp::text | e_cmp::renderable);
            w.m_entityFlags[e] |= e_entityFlags::recreateText;
        }

        if (entity.HasMember("light"))
        {
            const rapidjson::Value& lj = entity["light"];

            Light* light = memNew<Light>();
            light->m_type = e_lightType::nameToValue(getJsonValueOrDefault(lj, "type", "directional"));
            getJsonArrayOrDefault<float>(lj, "diffuse", &light->m_renderCpuData.m_diffuse.x, 3, 1.0f);

            if (light->m_type == e_lightType::point || light->m_type == e_lightType::spot)
            {
                light->m_renderCpuData.m_user.x = getJsonValueOrDefault(lj, "falloff", 1.0f);
                if (light->m_type == e_lightType::spot)
                {
                    light->m_renderCpuData.m_user.y = getJsonValueOrDefault(lj, "spot_falloff", 0.0f);
                }
            }

            if (light->m_type == e_lightType::spot)
            {
                light->m_cullingFrustum = maths::createFrustum(0.1f, light->m_renderCpuData.m_user.x, (1.0f - light->m_renderCpuData.m_user.y) * glm::pi<float>(), 1.0f);
            }

            w.m_light[e] = light;

            if (w.m_ulid[e] == 0)
            {
                std::string name = "light_" + std::to_string(e);
                w.m_ulid[e] = hash(name);
            }

            w.m_components[e] |= e_cmp::light;
        }

        if (entity.HasMember("animations"))
        {
            for (auto& ait = entity["animations"].MemberBegin(); ait != entity["animations"].MemberEnd(); ++ait)
            {
                ID animID = hash<std::string>(ait->name.GetString());
                m_animManager->addAnimation(e, animID, ait->value);
            }
        }

        if (entity.HasMember("children"))
        {
            size_t numChildren = entity["children"].Size();
            for (size_t i = 0; i < numChildren; ++i)
            {
                size_t c = w.newEntity();
                w.m_parent[c] = e;
                w.m_layer[c] = w.m_layer[e];
                loadEntity(c, entity["children"][i]);
            }
        }
    }

    void Scene::resize(u32 width, u32 height)
    {
        // if already active normalise necessary layers
        bool active = isActive();
        if(active)
            normalise(false);

        // update layer dimensions to new resolution
        for (auto& lid : m_layerOrder)
        {
            Layer& l = m_layers[lid];
            l.m_width = width;
            l.m_height = height;
        }

        // if already active unnormalise necessary layers
        if (active)
            normalise(true);
    }

    void Scene::normalise(bool inverse)
    {
        auto& w = *m_world;

        for (auto& lid : m_layerOrder)
        {
            Layer& l = m_layers[lid];

            if (!(l.m_flags & e_layerFlags::normalised))
                continue;

            glm::vec2 dims = (inverse) ? glm::vec2(l.m_width, l.m_height) : glm::vec2(1.0f / l.m_width, 1.0f / l.m_height);
            
            for (size_t i = 0, n = w.size(); i < n; ++i)
            {
                if (w.m_layer[i] != lid)
                    continue;

                // skip for child entities as they will be already normalised by their parents
                if (w.m_parent[i] != (size_t)-1)
                    continue;

                w.m_pos[i].x *= dims.x;
                w.m_pos[i].y *= dims.y;

                // TODO: handle entities that don't need normalised scaling more gracefully
                if (!(w.m_components[i] & e_cmp::camera))
                {
                    w.m_scale[i].x *= dims.x;
                    w.m_scale[i].y *= dims.y;
                }
            }
        }
        
    }

    void Scene::activate()
    {
        ResourceManager* rm = ResourceManager::get();
        auto& w = *m_world;
        auto ctx = gfx::getMainContext();

        // create ndc fullscreen quad for fullscreen draws
        GeometryCpuData* d = memNew<GeometryCpuData>();
        GeometryCpuData& ndcData = *d;
        ndcData = {};
        float vertexData[] =
        {
            -1.0f, -1.0f, 0.0f, 1.0f, 0.0f, 0.0f,
             1.0f,  1.0f, 0.0f, 1.0f, 1.0f, 1.0f,
             1.0f, -1.0f, 0.0f, 1.0f, 1.0f, 0.0f,
            -1.0f, -1.0f, 0.0f, 1.0f, 0.0f, 0.0f,
            -1.0f,  1.0f, 0.0f, 1.0f, 0.0f, 1.0f,
             1.0f,  1.0f, 0.0f, 1.0f, 1.0f, 1.0f
        };
        ndcData.m_numVertices = 6;
        ndcData.m_vertexData.m_size = sizeof(vertexData);
        ndcData.m_vertexData.m_data = memAlloc(sizeof(vertexData));
        memcpy(ndcData.m_vertexData.m_data, vertexData, sizeof(vertexData));

        u32 indexData[] = { 0, 1, 2, 3, 4, 5 };
        ndcData.m_numIndices = 6;
        ndcData.m_indexData.m_size = sizeof(indexData);
        ndcData.m_indexData.m_data = memAlloc(sizeof(indexData));
        memcpy(ndcData.m_indexData.m_data, indexData, sizeof(indexData));

        ModelLoadDesc md = {};
        md.m_genData.m_size = sizeof(GeometryCpuData);
        md.m_genData.m_data = d;
        m_ndcQuadRef = rm->load<ModelData, ModelLoadDesc>(md, e_resourceLoadFlags::none);

        u64 off = 0;
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_components[i] & (e_cmp::camera | e_cmp::light)))
                continue;

            w.m_cullingMask[i] = 1 << (off++);
        }

        // create uniform buffer for ambient light data
        for(auto& l : m_layers)
        {
            if (l.second.m_ambientLight.a <= 0.0f)
                continue;

            LightUniformData ambient = {};
            ambient.m_diffuse = l.second.m_ambientLight;

            gfx::BufferDesc bd = {};
            bd.m_access = gfx::e_accessFlags::write;
            bd.m_usage = gfx::e_bufferUsageFlags::uniform;
            bd.m_data.m_size = sizeof(LightUniformData);
            bd.m_data.m_data = &ambient;
            l.second.m_ambientLightUniformBuffer = ctx->createBuffer(bd);
        }

        // finish loading resources on main thread
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (w.m_components[i] & e_cmp::modelRef)
            {
                rm->load<ModelData, ModelLoadDesc>(w.m_modelRef[i], e_resourceLoadFlags::none);

                // update local aabb
                ModelData mdl = w.m_modelRef[i].getTypedResource();
                gfx::Geometry g = mdl.m_meshes[w.m_meshIndex[i]];
                w.m_localAabb[i].m_min = glm::vec3(g.m_minAABB[0], g.m_minAABB[1], g.m_minAABB[2]);
                w.m_localAabb[i].m_max = glm::vec3(g.m_maxAABB[0], g.m_maxAABB[1], g.m_maxAABB[2]);

                if (w.m_materialIndex[i] < mdl.m_materials.size())
                {
                    MaterialData mat = mdl.m_materials[w.m_materialIndex[i]].getTypedResource();

                    if (mat.m_flags != (u32)-1)
                    {
                        const char* variationNames[4] =
                        {
                            "diffuse_map",
                            "specular_map",
                            "bump_map",
                            "alpha_map"
                        };

                        size_t idx = 0;
                        for (u32 j = 0; j < 4; ++j)
                        {
                            if (mat.m_flags & (1 << j))
                            {
                                size_t k = idx++;
                                w.m_textures[i].m_texture[j] = mat.m_textures[k].getTypedResource();
                                w.m_textures[i].m_sampler[j] = m_renderer->getSampler(mat.m_defaultSamplerID);
                                w.m_pipelineVariations[i].m_variations[k] = hash<std::string>(variationNames[j]);
                            }
                        }
                    }
                    else
                    {
                        for (size_t j = 0; j < mat.m_textures.size(); ++j)
                        {
                            w.m_textures[i].m_texture[j] = mat.m_textures[j].getTypedResource();
                            w.m_textures[i].m_sampler[j] = m_renderer->getSampler(mat.m_defaultSamplerID);
                        }
                    }

                    w.m_material[i] = memNew<UniformBuffer>();
                    w.m_material[i]->m_buffer = mat.m_buffer;
                }
            }

            if (w.m_components[i] & e_cmp::textureRef)
            {
                rm->load<gfx::Texture, TextureLoadDesc>(w.m_textureRef[i], e_resourceLoadFlags::none);
            }

            if (w.m_components[i] & e_cmp::soundRef)
            {
                rm->load<audio::Sound, SoundLoadDesc>(w.m_soundRef[i], e_resourceLoadFlags::none);
            }

            if (w.m_components[i] & e_cmp::fontRef)
            {
                rm->load<gfx::Font*, gfx::FontDesc>(w.m_fontRef[i], e_resourceLoadFlags::none);
            }
        }

        // create gpu buffers for uniforms
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            w.m_uniforms[i] = memNew<UniformBuffer>();
            w.m_uniforms[i]->m_size = sizeof(UniformData);

            gfx::BufferDesc bd = {};
            bd.m_access = gfx::e_accessFlags::write;
            bd.m_usage = gfx::e_bufferUsageFlags::uniform;
            bd.m_data.m_data = &w.m_uniformData[i];
            bd.m_data.m_size = sizeof(UniformData);
            w.m_uniforms[i]->m_buffer = ctx->createBuffer(bd);
        }

        // create gpu resources for lights
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_components[i] & e_cmp::light))
                continue;

            LightType type = w.m_light[i]->m_type;

            // create light render buffer
            {
                gfx::BufferDesc bd = {};
                bd.m_access = gfx::e_accessFlags::write;
                bd.m_usage = gfx::e_bufferUsageFlags::uniform;
                bd.m_data.m_size = sizeof(LightUniformData);
                bd.m_data.m_data = nullptr;
                w.m_light[i]->m_renderBuffer = ctx->createBuffer(bd);
            }

            // create buffer for rendering shadow maps
            {
                gfx::BufferDesc bd = {};
                bd.m_access = gfx::e_accessFlags::write;
                bd.m_usage = gfx::e_bufferUsageFlags::uniform;
                bd.m_data.m_data = nullptr;

                switch (type)
                {
                case e_lightType::directional:
                case e_lightType::spot:
                    bd.m_data.m_size = sizeof(glm::mat4);
                    break;
                case e_lightType::point:
                    bd.m_data.m_size = sizeof(glm::mat4) * 6;
                    break;
                }
                w.m_light[i]->m_shadowMapBuffer = ctx->createBuffer(bd);
            }

            // create shadow maps
            {
                gfx::TextureDesc td = {};
                td.m_usage = gfx::e_textureUsageFlags::depthStencil | gfx::e_textureUsageFlags::shaderRead;
                td.m_format.m_channels = gfx::e_formatChannelFlags::d;
                td.m_format.m_type = gfx::e_formatType::signedFloat;
                td.m_format.m_bitDepth = 32;
                td.m_width = 2048;
                td.m_height = 2048;

                switch(type)
                {
                case e_lightType::directional:
                case e_lightType::spot:
                    td.m_layers = 1;
                    break;
                case e_lightType::point:
                    td.m_layers = 6;
                }

                w.m_light[i]->m_shadowMap = ctx->createTexture(td);
            }
        }

        // create camera matrix buffers
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_components[i] & e_cmp::camera))
                continue;
            
            gfx::BufferDesc bd = {};
            bd.m_access = gfx::e_accessFlags::write;
            bd.m_usage = gfx::e_bufferUsageFlags::uniform;
            bd.m_data.m_data = nullptr;
            bd.m_data.m_size = sizeof(glm::mat4);
            w.m_camera[i]->m_matrixBuffer = ctx->createBuffer(bd);
        }

        // unnormalise layers
        normalise(true);

        m_flags |= e_sceneFlags::activated;
    }

    void Scene::shutdown()
    {
        auto ctx = gfx::getMainContext();

        m_world->clear();
        memDelete(m_world);

        ResourceManager* rm = ResourceManager::get();
        rm->unload(m_ndcQuadRef);

        for (auto& l : m_layers)
        {
            if (l.second.m_ambientLightUniformBuffer)
            {
                ctx->destroyBuffer(l.second.m_ambientLightUniformBuffer);
            }
        }
        m_layers.clear();
        m_layerOrder.clear();

        m_animManager->shutdown();
        memDelete(m_animManager);
    }

    void Scene::processInput(std::vector<utils::InputEvent>& events)
    {
        Engine* engine = Engine::get();
        GameState& s = engine->getGameState();


        // reset flag state for all entities
        auto& w = *m_world;
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            w.m_entityFlags[i] &= ~e_entityFlags::wasHovered;
            if (w.m_entityFlags[i] & e_entityFlags::hovered)
                w.m_entityFlags[i] |= e_entityFlags::wasHovered;
            w.m_entityFlags[i] &= ~e_entityFlags::hovered;

            w.m_entityFlags[i] &= ~e_entityFlags::wasSelected;
            if (w.m_entityFlags[i] & e_entityFlags::selected)
                w.m_entityFlags[i] |= e_entityFlags::wasSelected;
            w.m_entityFlags[i] &= ~e_entityFlags::selected;
        }

        for (auto& e : events)
        {
            if (e.m_type == utils::e_inputEventType::mouse)
            {
                float ww = (float)utils::getWindowWidth();
                float wh = (float)utils::getWindowHeight();

                // calculate mouse hovers
                if (e.m_mouse.m_type == utils::e_mouseEventType::move)
                {
                    maths::Ray mr = {};
                    glm::vec3 mouse = glm::vec3(e.m_mouse.m_pos.m_posX, e.m_mouse.m_pos.m_posY, 1.0f);
                    mouse.x = (mouse.x / ww) * 2.0f - 1.0f;
                    mouse.y = (mouse.y / wh) * 2.0f - 1.0f;

                    float closestT = FLT_MAX;
                    size_t closest = -1;
                    maths::Ray closestRay = {};
                    for (auto& lid : m_layerOrder)
                    {
                        Layer& l = m_layers[lid];
                        Camera* cam = w.m_camera[l.m_cameraEntity];

                        // transform mouse into world coords
                        if (cam->m_type == e_cameraType::perspective)
                        {
                            glm::mat4 inverseMatrix = glm::inverse(cam->m_matrix);
                            glm::vec4 unproj = inverseMatrix * glm::vec4(mouse, 1.0f);
                            mr.m_origin = glm::vec3(w.m_worldMatrix[l.m_cameraEntity][3]);
                            mr.m_ray = glm::vec3(unproj) / unproj.w;
                            mr.m_ray = mr.m_ray - mr.m_origin;
                            mr.m_ray = glm::normalize(mr.m_ray);
                        }
                        else if (cam->m_type == e_cameraType::orthographic)
                        {
                            float x = cam->m_ortho.m_left + (cam->m_ortho.m_right - cam->m_ortho.m_left) * (mouse.x * 0.5f + 0.5f);
                            float y = cam->m_ortho.m_top + (cam->m_ortho.m_bottom - cam->m_ortho.m_top) * (mouse.y * 0.5f + 0.5f);
                            mr.m_origin = glm::vec3(x, y, 0.0f);
                            glm::vec4 origin = w.m_worldMatrix[l.m_cameraEntity] * glm::vec4(mr.m_origin, 1.0f);
                            mr.m_origin = glm::vec3(origin) / origin.w;
                            mr.m_ray.x = w.m_worldMatrix[l.m_cameraEntity][0][2];
                            mr.m_ray.y = w.m_worldMatrix[l.m_cameraEntity][1][2];
                            mr.m_ray.z = w.m_worldMatrix[l.m_cameraEntity][2][2];
                        }
                        mr.compute();

                        for (size_t i = 0, n = w.size(); i < n; ++i)
                        {
                            if (w.m_layer[i] != lid)
                                continue;

                            if (!(w.m_components[i] & e_cmp::transform))
                                continue;

                            // we only care if this particular entity is set to pass through cursor and not parents
                            if (w.m_entityFlags[i] & e_entityFlags::passthroughCursor)
                                continue;

                            if (w.m_inheritedEntityFlags[i] & e_entityFlags::hidden)
                                continue;

                            float t = 0.0f;
                            if (maths::rayAABBIntersection(mr, { w.m_worldObb[i].m_min, w.m_worldObb[i].m_max }, t) && t < closestT && t >= 0.0f)
                            {
                                // unhover previous closest
                                if (closest != (size_t)-1)
                                    w.m_entityFlags[closest] &= ~e_entityFlags::hovered;

                                w.m_entityFlags[i] |= e_entityFlags::hovered;
                                closestT = t;
                                closest = i;
                                closestRay = mr;
                            }
                        }

                        if (closest == (size_t)-1)
                        {
                            closestT = cam->m_far;
                            closestRay = mr;
                        }

                    }

                    // store the intersection point as game state
                    glm::vec3 intersection = closestRay.m_origin + closestRay.m_ray * closestT;
                    s["mouse_intersection_world"].setFloats(intersection);

                    ID lid = 0;
                    if (closest != (size_t)-1)
                        lid = w.m_layer[closest];
                    else
                        lid = m_layerOrder.back();
                    Layer& l = m_layers[lid];
                    glm::vec4 is = w.m_camera[l.m_cameraEntity]->m_matrix * glm::vec4(intersection, 1.0f);
                    glm::vec3 screen = glm::vec3(is) / is.w;
                    screen.x = (screen.x * 0.5f + 0.5f) * ww;
                    screen.y = (screen.y * 0.5f + 0.5f) * wh;
                    s["mouse_intersection_screen"].setFloats(screen);
                }

                // calculate mouse selection
                if (e.m_mouse.m_type == utils::e_mouseEventType::button)
                {
                    if (e.m_mouse.m_button.m_button == utils::e_mouseButton::left && e.m_mouse.m_button.m_buttonState == utils::e_mouseButtonState::clicked)
                    {
                        size_t selected = (size_t)-1;
                        for (size_t i = 0, n = w.size(); i < n; ++i)
                        {
                            if (w.m_entityFlags[i] & e_entityFlags::hovered)
                            {
                                w.m_entityFlags[i] |= e_entityFlags::selected;
                                selected = i;
                            }
                        }
                        s["selected_entity"].setUints(selected);
                    }
                }
            }
        }

        for (auto& t : m_triggers)
        {
            for (auto& c : t.m_conditions)
            {
                c->processInput(events);
            }
        }

        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_components[i] & e_cmp::triggers))
                continue;

            std::vector<TriggerLogic>& entityTriggers = m_entityTriggers[w.m_triggerID[i]];
            for (auto& t : entityTriggers)
            {
                for (auto& c : t.m_conditions)
                {
                    c->processInput(events);
                }
            }
        }
    }

    void updateTriggerLogic(float dt, TriggerLogic& t, std::vector<TriggerAction*>& actions)
    {
        // update non-simultaneous actions
        if (!t.m_simultaneousActions && t.m_actionIndex != (size_t)-1)
        {
            if (!t.m_actions[t.m_actionIndex]->m_firing)
            {
               // push next actions
                t.m_actionIndex++;
                if (t.m_actionIndex >= t.m_actions.size())
                {
                    t.m_actionIndex = (size_t)-1;
                }
                else
                {
                    TriggerAction* a = t.m_actions[t.m_actionIndex];
                    a->m_firing = true;
                    actions.push_back(a);
                }
            }
        }

        if (t.m_fire == e_triggerLogicFireType::cantFire)
            return;


        bool fire = true;
        for (auto& c : t.m_conditions)
        {
            bool met = c->update(dt);
            if (c->m_not)
                met = !met;
            fire = fire && met;
        }

        if (fire)
        {
            if (t.m_simultaneousActions)
            {
                for (auto& a : t.m_actions)
                {
                    a->m_firing = true;
                    actions.push_back(a);
                }
            }
            else
            {
                TriggerAction* a = t.m_actions[0];
                t.m_actionIndex = 0;
                a->m_firing = true;
                actions.push_back(a);
            }

            if (t.m_fire == e_triggerLogicFireType::fireOnce)
                t.m_fire = e_triggerLogicFireType::cantFire;
        }
        else
        {
            for (auto& a : t.m_elseActions)
            {
                a->m_firing = true;
                actions.push_back(a);
            }
        }
    }

    void Scene::updateLogic(float dt)
    {
        auto& w = *m_world;

        for (auto& t : m_triggers)
        {
            updateTriggerLogic(dt, t, m_currentActions);
        }

        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_components[i] & e_cmp::triggers))
                continue;

            std::vector<TriggerLogic>& entityTriggers = m_entityTriggers[w.m_triggerID[i]];
            for (auto& t : entityTriggers)
            {
                updateTriggerLogic(dt, t, m_currentActions);
            }
        }

        // sort current updating actions by priority
        std::sort(m_currentActions.begin(), m_currentActions.end(), [](TriggerAction* a, TriggerAction* b) { return a->m_priority < b->m_priority; });

        std::vector<TriggerAction*> completedActions;
        for (auto& a : m_currentActions)
        {
            if (a->update(dt))
            {
                completedActions.push_back(a);
            }
        }

        for (auto& a : completedActions)
        {
            a->m_firing = false;
            a->reset();
            m_currentActions.erase(std::find(m_currentActions.begin(), m_currentActions.end(), a));
        }
    }

    void Scene::updateTransforms(float dt)
    {
        auto& w = *m_world;
        auto ctx = gfx::getMainContext();

        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_components[i] & e_cmp::transform))
                continue;
            
            glm::mat4 t = glm::translate(glm::vec3(w.m_pos[i].x, w.m_pos[i].y, w.m_pos[i].z));
            glm::mat4 r = maths::createRotationMatrix(w.m_rot[i].y, w.m_rot[i].x, w.m_rot[i].z);
            glm::mat4 s = glm::scale(w.m_scale[i]);

            w.m_localMatrix[i] = t * r * s;

            glm::mat4 p = glm::mat4(1.0f);
            if (w.m_parent[i] != (size_t)-1)
            {
                p = w.m_worldMatrix[w.m_parent[i]];
            }
            w.m_worldMatrix[i] = p * w.m_localMatrix[i];

            // update world obb
            glm::vec4 min = w.m_worldMatrix[i] * glm::vec4(w.m_localAabb[i].m_min, 1.0f);
            glm::vec4 max =  w.m_worldMatrix[i] * glm::vec4(w.m_localAabb[i].m_max, 1.0f);

            maths::OBB& obb = w.m_worldObb[i];

            obb.m_min = min;
            obb.m_max = max;

            // precalculate useful obb data
            glm::vec3 corners[4] =
            {
                { obb.m_min.x, obb.m_min.y, obb.m_min.z },
                { obb.m_max.x, obb.m_min.y, obb.m_min.z },
                { obb.m_min.x, obb.m_max.y, obb.m_min.z },
                { obb.m_min.x, obb.m_min.y, obb.m_max.z }
            };

            obb.m_axes[0] = corners[1] - corners[0];
            obb.m_axes[1] = corners[2] - corners[0];
            obb.m_axes[2] = corners[3] - corners[0];

            obb.m_centre = corners[0] + 0.5f * (obb.m_axes[0] + obb.m_axes[1] + obb.m_axes[2]);
            obb.m_extents = glm::vec3(glm::length(obb.m_axes[0]), glm::length(obb.m_axes[1]), glm::length(obb.m_axes[2]));
            obb.m_axes[0] = obb.m_axes[0] / obb.m_extents.x;
            obb.m_axes[1] = obb.m_axes[1] / obb.m_extents.y;
            obb.m_axes[2] = obb.m_axes[2] / obb.m_extents.z;
            obb.m_extents *= 0.5f;

            // precalculate forward vector for ease of use
            glm::vec3 forward = glm::vec3(0.0, 0.0, 1.0);
            w.m_forward[i] = glm::normalize(glm::mat3(w.m_worldMatrix[i]) * forward);

            glm::vec3 right = glm::vec3(1.0, 0.0, 0.0);
            w.m_right[i] = glm::normalize(glm::mat3(w.m_worldMatrix[i]) * right);
        }
    }

    void Scene::updateText(float dt)
    {
        auto& w = *m_world;

        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_entityFlags[i] & e_entityFlags::recreateText))
                continue;

            if (!w.m_geometry[i])
                w.m_geometry[i] = memNew<gfx::Geometry>();

            std::string text = w.m_text[i]->m_text;
            gfx::Font* font = w.m_fontRef[i].getTypedResource();
            *w.m_geometry[i] = font->createTextMesh(text, gfx::e_textHAlign::centre, gfx::e_textVAlign::centre);

            gfx::Geometry* geo = w.m_geometry[i];

            w.m_localAabb[i].m_min = glm::vec3(geo->m_minAABB[0], geo->m_minAABB[1], geo->m_minAABB[2]);
            w.m_localAabb[i].m_max = glm::vec3(geo->m_maxAABB[0], geo->m_maxAABB[1], geo->m_maxAABB[2]);

            w.m_entityFlags[i] &= ~e_entityFlags::recreateText;
        }
    }

    void Scene::updateRendering(float dt)
    {
        auto& w = *m_world;
        auto ctx = gfx::getMainContext();

        ID defaultSamplerID = hash<std::string>("clamp_linear");

        // update cameras
        glm::mat4 clip = glm::mat4( 1.0f, 0.0f, 0.0f, 0.0f,
                                    0.0f, -1.0f, 0.0f, 0.0f,
                                    0.0f, 0.0f, 0.5f, 0.0f,
                                    0.0f, 0.0f, 0.5f, 1.0f);

        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_components[i] & e_cmp::camera))
                continue;

            glm::vec3 eye = glm::vec3(w.m_worldMatrix[i][3]);
            glm::vec3 centre = eye + w.m_forward[i];

            glm::mat4 view = maths::createLookAtMatrix(eye, centre, glm::vec3(0.0f, 1.0f, 0.0f));
            glm::mat4 proj = glm::mat4(1.0f);

            Camera* cam = w.m_camera[i];
            if (cam->m_type == e_cameraType::orthographic)
            {
                float l, r, b, t;
                if (cam->m_ortho.m_matchOutput)
                {
                    l = ((float)utils::getWindowWidth()) * -0.5f;
                    r = -l;
                    t = ((float)utils::getWindowHeight()) * 0.5f;
                    if (cam->m_ortho.m_invertY)
                        t = -t;
                    b = -t;

                    cam->m_ortho.m_left = l;
                    cam->m_ortho.m_right = r;
                    cam->m_ortho.m_bottom = b;
                    cam->m_ortho.m_top = t;
                }
                else
                {
                    l = cam->m_ortho.m_left;
                    r = cam->m_ortho.m_right;
                    b = cam->m_ortho.m_bottom;
                    t = cam->m_ortho.m_top;
                }
                proj = clip * glm::ortho(l, r, b, t, cam->m_near, cam->m_far);
            }
            else
            {
                proj = clip * glm::perspective(glm::radians(cam->m_perspective.m_fov), cam->m_perspective.m_aspect, cam->m_near, cam->m_far);
            }

            cam->m_view = view;
            cam->m_proj = proj;
            cam->m_matrix = proj * view;

            Mem matrixMem;
            matrixMem.m_size = sizeof(glm::mat4);
            matrixMem.m_data = &cam->m_matrix;
            ctx->updateBuffer(w.m_camera[i]->m_matrixBuffer, matrixMem);

            // check if this camera is the viewing camera for any layer
            Layer& l = m_layers[w.m_layer[i]];
            if (l.m_cameraEntity == i && l.m_deferredRender)
            {
                View* v = m_renderer->getView(hash<std::string>("layer_deferred"), w.m_layer[i]);
                if (v)
                {
                    LayerCameraData cd = {};
                    cd.m_inverseVP = glm::inverse(cam->m_matrix);
                    cd.m_eye = glm::vec4(eye, 0.0f);

                    Mem inverseMem;
                    inverseMem.m_size = sizeof(LayerCameraData);
                    inverseMem.m_data = &cd;

                    ctx->updateBuffer(v->m_uniformBuffer, inverseMem);
                }
            }
        }

        // bind texture refs to texture slot
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_components[i] & e_cmp::textureRef))
                continue;

            gfx::Texture tex = w.m_textureRef[i].getTypedResource();
            if (tex)
            {
                w.m_textures[i].m_texture[0] = tex;
                w.m_textures[i].m_sampler[0] = m_renderer->getSampler(defaultSamplerID);
            }
        }

        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_components[i] & e_cmp::fontRef))
                continue;

            gfx::Font* font = w.m_fontRef[i].getTypedResource();
            gfx::Texture tex = font->getTexture();
            if (tex)
            {
                w.m_textures[i].m_texture[0] = tex;
                w.m_textures[i].m_sampler[0] = m_renderer->getSampler(defaultSamplerID);
            }
        }

        // get the bounding boxes for each layer
        std::unordered_map <ID, maths::AABB> layerAABB;
        for (auto& l : m_layers)
        {
            maths::AABB& aabb = layerAABB[l.first];
            aabb.m_min = glm::vec3(FLT_MAX);
            aabb.m_max = glm::vec3(-FLT_MAX);
            
            for (size_t i = 0, n = w.size(); i < n; ++i)
            {
                if (w.m_layer[i] != l.first)
                    continue;

                if (!(w.m_components[i] & e_cmp::renderable))
                    continue;

                maths::AABB& aabb = layerAABB[w.m_layer[i]];

                aabb.m_min.x = std::min(aabb.m_min.x, w.m_worldObb[i].m_min.x);
                aabb.m_min.y = std::min(aabb.m_min.y, w.m_worldObb[i].m_min.y);
                aabb.m_min.z = std::min(aabb.m_min.z, w.m_worldObb[i].m_min.z);
                aabb.m_max.x = std::max(aabb.m_max.x, w.m_worldObb[i].m_max.x);
                aabb.m_max.y = std::max(aabb.m_max.y, w.m_worldObb[i].m_max.y);
                aabb.m_max.z = std::max(aabb.m_max.z, w.m_worldObb[i].m_max.z);
            }
        }

        // update light render buffers
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_components[i] & e_cmp::light))
                continue;

            Light* light = w.m_light[i];

            // update world matrix
            light->m_renderCpuData.m_world = w.m_worldMatrix[i];

            // update shadow map matrix
            switch (light->m_type)
            {
            case e_lightType::spot:
            {
                glm::vec3 eye = glm::vec3(w.m_worldMatrix[i][3]);
                glm::vec3 centre = eye + w.m_forward[i];

                float range = light->m_renderCpuData.m_user.x;
                light->m_renderCpuData.m_user.z = range;

                glm::mat4 view = maths::createLookAtMatrix(eye, centre, glm::vec3(0.0f, 1.0f, 0.0f));
                glm::mat4 proj = clip * glm::perspective((1.0f - light->m_renderCpuData.m_user.y) * glm::pi<float>(), 1.0f, 0.01f, range);
                
                glm::mat4 vp = proj * view;
                
                light->m_renderCpuData.m_view = view;
                light->m_renderCpuData.m_proj = proj;
                
                Mem mem = {};
                mem.m_size = sizeof(glm::mat4);
                mem.m_data = &vp;
                ctx->updateBuffer(light->m_shadowMapBuffer, mem);
            }
            break;
            case e_lightType::point:
            {
                glm::vec3 lookAts[6] =
                {
                    glm::vec3( 1.0f,  0.0f,  0.0f),
                    glm::vec3(-1.0f,  0.0f,  0.0f),
                    glm::vec3( 0.0f,  1.0f,  0.0f),
                    glm::vec3( 0.0f, -1.0f,  0.0f),
                    glm::vec3( 0.0f,  0.0f,  1.0f),
                    glm::vec3( 0.0f,  0.0f, -1.0f)
                };

                glm::vec3 ups[6] =
                {
                    glm::vec3(0.0f, 1.0f, 0.0f),
                    glm::vec3(0.0f, 1.0f, 0.0f),
                    glm::vec3(0.0f, 0.0f, -1.0f),
                    glm::vec3(0.0f, 1.0f, 1.0f),
                    glm::vec3(0.0f, 1.0f, 0.0f),
                    glm::vec3(0.0f, 1.0f, 0.0f)
                };

                glm::mat4 matrices[6];
                for (size_t j = 0;  j < 6; ++j)
                {
                    glm::vec3 eye = glm::vec3(w.m_worldMatrix[i][3]);
                    glm::vec3 centre = eye + lookAts[j];

                    float range = light->m_renderCpuData.m_user.x;

                    glm::mat4 view = maths::createLookAtMatrix(eye, centre, glm::vec3(0.0f, 1.0f, 0.0f));
                    glm::mat4 proj = clip * glm::perspective(glm::pi<float>() * 0.5f, 1.0f, 0.01f, range);

                    light->m_renderCpuData.m_user.z = 0.01f;
                    light->m_renderCpuData.m_user.w = range;

                    matrices[j] = proj * view;
                }

                Mem mem = {};
                mem.m_size = sizeof(glm::mat4) * 6;
                mem.m_data = matrices;
                ctx->updateBuffer(light->m_shadowMapBuffer, mem);
            }
            break;
            case e_lightType::directional:
            {
                maths::AABB aabb = layerAABB[w.m_layer[i]];

                glm::vec3 eye = glm::vec3(w.m_worldMatrix[i][3]);
                glm::vec3 centre = eye + w.m_forward[i];

                glm::mat4 view = maths::createLookAtMatrix(eye, centre, glm::vec3(0.0f, 1.0f, 0.0f));

                // transform aabb corners into camera view space
                glm::vec4 corners[8] = {
                    glm::vec4(aabb.m_min.x, aabb.m_min.y, aabb.m_min.z, 1.0f),
                    glm::vec4(aabb.m_max.x, aabb.m_min.y, aabb.m_min.z, 1.0f),
                    glm::vec4(aabb.m_max.x, aabb.m_max.y, aabb.m_min.z, 1.0f),
                    glm::vec4(aabb.m_min.x, aabb.m_max.y, aabb.m_min.z, 1.0f),
                    glm::vec4(aabb.m_min.x, aabb.m_min.y, aabb.m_max.z, 1.0f),
                    glm::vec4(aabb.m_max.x, aabb.m_min.y, aabb.m_max.z, 1.0f),
                    glm::vec4(aabb.m_max.x, aabb.m_max.y, aabb.m_max.z, 1.0f),
                    glm::vec4(aabb.m_min.x, aabb.m_max.y, aabb.m_max.z, 1.0f)
                };

                aabb.m_min = glm::vec3(FLT_MAX);
                aabb.m_max = glm::vec3(-FLT_MAX);
                for (auto& a : corners)
                {
                    a = view * a;
                    a = a / glm::vec4(a.w);
                    aabb.m_min.x = std::min(aabb.m_min.x, a.x);
                    aabb.m_min.y = std::min(aabb.m_min.y, a.y);
                    aabb.m_min.z = std::min(aabb.m_min.z, a.z);
                    aabb.m_max.x = std::max(aabb.m_max.x, a.x);
                    aabb.m_max.y = std::max(aabb.m_max.y, a.y);
                    aabb.m_max.z = std::max(aabb.m_max.z, a.z);
                }
                aabb.m_min -= glm::vec3(0.1f);
                aabb.m_max += glm::vec3(0.1f);

                glm::mat4 proj = clip * glm::ortho(aabb.m_min.x, aabb.m_max.x, aabb.m_min.y, aabb.m_max.y, aabb.m_min.z, aabb.m_max.z);
                glm::mat4 vp = proj * view;

                light->m_renderCpuData.m_view = view;
                light->m_renderCpuData.m_proj = proj;

                light->m_renderCpuData.m_user.y = 0.0f;
                light->m_renderCpuData.m_user.z = aabb.m_min.z;
                light->m_renderCpuData.m_user.w = aabb.m_max.z;

                Mem mem = {};
                mem.m_size = sizeof(glm::mat4);
                mem.m_data = &vp;
                ctx->updateBuffer(light->m_shadowMapBuffer, mem);
            }
            break;
            }

            Mem mem = {};
            mem.m_size = sizeof(LightUniformData);
            mem.m_data = &light->m_renderCpuData;
            ctx->updateBuffer(light->m_renderBuffer, mem);
        }

        // bind model ref vertex/index buffers
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_components[i] & e_cmp::modelRef))
                continue;

            ModelData mdl = w.m_modelRef[i].getTypedResource();
            gfx::Geometry geometry = mdl.m_meshes[w.m_meshIndex[i]];

            if (!w.m_geometry[i])
            {
                w.m_geometry[i] = memNew<gfx::Geometry>();
            }

            w.m_geometry[i]->m_vertexBuffer = geometry.m_vertexBuffer;
            w.m_geometry[i]->m_indexBuffer = geometry.m_indexBuffer;
            w.m_geometry[i]->m_numVertices = geometry.m_numVertices;
            w.m_geometry[i]->m_numIndices = geometry.m_numIndices;
        }

        // update per layer frustum culling of entities
        for (auto& lid : m_layerOrder)
        {
            Layer& l = m_layers[lid];

            u32 cullingMask = w.m_cullingMask[l.m_cameraEntity];

            CullingType cullingType = e_cullingType::none;
            maths::Frustum frustum = {};
            glm::mat4 mat = glm::mat4(1.0f);

            if (w.m_camera[l.m_cameraEntity]->m_type == e_cameraType::perspective)
            {
                cullingType = e_cullingType::frustum;
                frustum = w.m_camera[l.m_cameraEntity]->m_cullingFrustum;
                mat = w.m_camera[l.m_cameraEntity]->m_matrix;
            }

            for (size_t i = 0, n = w.size(); i < n; ++i)
            {
                if (i == l.m_cameraEntity)
                    continue;

                if (!(w.m_components[i] & e_cmp::renderable))
                    continue;

                if (w.m_inheritedEntityFlags[i] & (e_entityFlags::hidden | e_baseEntityFlags::disabled))
                    continue;

                if (w.m_layer[i] != lid)
                    continue;

                // reset to culled state
                w.m_cullingMask[i] &= ~cullingMask;

                switch (cullingType)
                {
                case e_cullingType::frustum:
                {
                    if (maths::frustumOBBIntersectionSimple(mat, w.m_worldObb[i]))
                    {
                        w.m_cullingMask[i] |= cullingMask;
                    }
                }
                break;
                case e_cullingType::sphere:
                case e_cullingType::none:
                {
                    w.m_cullingMask[i] |= cullingMask;
                }
                break;
                }
            }
        }

        // update per light culling
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_components[i] & e_cmp::light))
                continue;

            u32 cullingMask = w.m_cullingMask[i];
            CullingType cullingType = e_cullingType::none;
            maths::Frustum frustum = {};
            glm::mat4 mat = glm::mat4(1.0f);
            
            glm::vec3 centre = glm::vec3(0.0f);
            float radius = 0.0f;
            switch (w.m_light[i]->m_type)
            {
            case e_lightType::spot:
            {
                cullingType = e_cullingType::frustum;
                frustum = w.m_light[i]->m_cullingFrustum;
                mat = w.m_light[i]->m_renderCpuData.m_proj * w.m_light[i]->m_renderCpuData.m_view;
            }
            break;
            case e_lightType::point:
            {
                cullingType = e_cullingType::sphere;
                radius = w.m_light[i]->m_renderCpuData.m_user.x;
                centre = w.m_worldMatrix[i][3];
            }
            break;
            case e_lightType::directional:
            break;
            }

            for (size_t j = 0; j < n; ++j)
            {
                if (i == j)
                    continue;

                if (!(w.m_components[j] & e_cmp::renderable))
                    continue;

                if (w.m_inheritedEntityFlags[j] & (e_entityFlags::hidden | e_baseEntityFlags::disabled))
                    continue;

                if (w.m_layer[i] != w.m_layer[j])
                    continue;

                // reset to culled state
                w.m_cullingMask[j] &= ~cullingMask;

                switch (cullingType)
                {
                case e_cullingType::frustum:
                {
                    if (maths::frustumOBBIntersectionSimple(mat, w.m_worldObb[j]))
                    {
                        w.m_cullingMask[j] |= cullingMask;
                    }
                }
                break;
                case e_cullingType::sphere:
                {
                    if (maths::sphereOBBIntersection(centre, radius, w.m_worldObb[j]))
                    {
                        w.m_cullingMask[j] |= cullingMask;
                    }
                }
                break;
                case e_cullingType::none:
                {
                    w.m_cullingMask[j] |= cullingMask;
                }
                break;
                }
            }
        }

        // update uniform gpu buffer
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!w.m_uniforms[i]->m_buffer)
                continue;

            w.m_uniformData[i].m_world = w.m_worldMatrix[i];
            w.m_uniformData[i].m_userData1.x = w.m_time[i];

            Mem mem = {};
            mem.m_size = sizeof(UniformData);
            mem.m_data = &w.m_uniformData[i];
            ctx->updateBuffer(w.m_uniforms[i]->m_buffer, mem);
        }
    }

    void Scene::update(float dt)
    {
        World& w = *m_world;

        // set up inherited flags from parents
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            w.m_inheritedEntityFlags[i] = w.m_entityFlags[i];
            if (w.m_parent[i] != (size_t)-1)
            {
                w.m_inheritedEntityFlags[i] |= w.m_inheritedEntityFlags[w.m_parent[i]];
            }
        }

        // increment time for each entity
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (w.m_inheritedEntityFlags[i] & e_baseEntityFlags::disabled)
                continue;

            w.m_time[i] += dt;
        }

        updateLogic(dt);

        if (m_animManager)
        {
            m_animManager->update(dt);
        }

        updateTransforms(dt);
        updateText(dt);
        updateRendering(dt);
    }

    void Scene::registerViewFunctions()
    {
        m_renderer->unregisterRenderFunc("render_scene");
        m_renderer->unregisterRenderFunc("render_layer");
        m_renderer->unregisterRenderFunc("render_post_process");
        m_renderer->unregisterSetupFunc("setup_layers");
        m_renderer->registerRenderFunc(BIND_FUNCTION(Scene, render), "render_scene");
        m_renderer->registerRenderFunc(BIND_FUNCTION(Scene, renderLayer), "render_layer");
        m_renderer->registerRenderFunc(BIND_FUNCTION(Scene, renderLayerPostProcess), "render_post_process");
        m_renderer->registerRenderFunc(BIND_FUNCTION(Scene, renderLayerDeferred), "render_deferred");
        m_renderer->registerSetupFunc(BIND_FUNCTION(Scene, setupLayerViews), "setup_layers");
    }

    void Scene::render(View* view)
    {
        auto ctx = gfx::getMainContext();
        World& w = *m_world;

        // bind the camera buffer from the world
        for (size_t i = 0; i < w.size(); ++i)
        {
            if (!(w.m_components[i] & e_cmp::camera))
                continue;
            ctx->bindBuffer(w.m_camera[i]->m_matrixBuffer, BUFFER_SLOT_CAMERA_UNIFORMS, gfx::e_bindFlags::uniform);
            break;
        }

        for(size_t i = 0; i < w.size(); ++i)
        {
            // skip hidden or disabled entities
            if(w.m_entityFlags[i] & (e_entityFlags::hidden | e_baseEntityFlags::disabled))
                continue;

            // check component exclusion
            if(view->m_components && !(w.m_components[i] & view->m_components))
                continue;

            // entities need geometry to draw
            if(!w.m_geometry[i])
                continue;

            // bind entity pipelines
            if(w.m_pipeline[i])
            {
                gfx::RenderPipeline pipeline = m_renderer->getPipelineForPass(w.m_pipeline[i], view->m_pass);
                ctx->bindRenderPipeline(pipeline);
            }

            // bind entity uniform buffers
            if(w.m_uniforms[i])
            {
                ctx->bindBuffer(w.m_uniforms[i]->m_buffer, BUFFER_SLOT_ENTITY_UNIFORMS, gfx::e_bindFlags::uniform);
            }

            if(w.m_material[i])
            {
                ctx->bindBuffer(w.m_material[i]->m_buffer, BUFFER_SLOT_ENTITY_MATERIAL, gfx::e_bindFlags::uniform);
            }

            // bind entity textures
            for(u32 t = 0; t < 4; ++t)
                if(w.m_textures[i].m_texture[t])
                    ctx->bindTexture(w.m_textures[i].m_texture[t], w.m_textures[i].m_sampler[t], t, gfx::e_bindFlags::uniform);

            // bind geometry and draw
            ctx->bindBuffer(w.m_geometry[i]->m_vertexBuffer, 0, gfx::e_bindFlags::vertexBuffer);

            if(w.m_geometry[i]->m_indexBuffer)
            {
                ctx->drawIndexed((u32)w.m_geometry[i]->m_numIndices, w.m_geometry[i]->m_indexBuffer, gfx::e_indexType::uint32);
            }
            else
            {
                ctx->draw((u32)w.m_geometry[i]->m_numVertices);
            }
        }
    }

    void Scene::renderLayer(View* view)
    {
        auto ctx = gfx::getMainContext();
        World& w = *m_world;

        ID layerID = 0;
        u32 cullingMask = 0;
        if (m_layers.find(view->m_linkID) != m_layers.end())
        {
            Layer& layer = m_layers.at(view->m_linkID);
            cullingMask = w.m_cullingMask[layer.m_cameraEntity];
            ctx->bindBuffer(w.m_camera[layer.m_cameraEntity]->m_matrixBuffer, BUFFER_SLOT_CAMERA_UNIFORMS, gfx::e_bindFlags::uniform);
            layerID = view->m_linkID;
        }
        else
        {
            // look for entity this view is linked to
            for (size_t i = 0, n = w.size(); i < n; ++i)
            {
                if (w.m_ulid[i] != (ULID)view->m_linkID)
                    continue;

                if (!(w.m_components[i] & e_cmp::light))
                    continue;

                ctx->bindBuffer(w.m_light[i]->m_shadowMapBuffer, BUFFER_SLOT_CAMERA_UNIFORMS, gfx::e_bindFlags::uniform);
                layerID = w.m_layer[i];
                cullingMask = w.m_cullingMask[i];
                break;
            }
        }

        for (size_t i = 0; i < w.size(); ++i)
        {
            // skip layers not on the rendering layer
            if (w.m_layer[i] != layerID)
                continue;

            // skip hidden or disabled entities
            if (w.m_inheritedEntityFlags[i] & (e_entityFlags::hidden | e_baseEntityFlags::disabled))
                continue;
            
            // skip culled entities
            if (cullingMask && !(w.m_cullingMask[i] & cullingMask))
                continue;
            
            // check component exclusion
            if (view->m_components && !(w.m_components[i] & view->m_components))
                continue;

            // entities need geometry to draw
            if (!w.m_geometry[i])
                continue;

            // bind entity pipelines
            if (w.m_pipeline[i])
            {
                ID variation = view->m_pipelineVariation;
                for (size_t j = 0; j < MAX_PIPELINE_VARIATIONS; ++j)
                {
                    if (w.m_pipelineVariations[i].m_variations[j] != 0)
                        variation = variation ^ w.m_pipelineVariations[i].m_variations[j];
                }
                gfx::RenderPipeline pipeline = m_renderer->getPipelineForPass(w.m_pipeline[i], view->m_pass, variation);
                ctx->bindRenderPipeline(pipeline);
            }

            // bind entity uniform buffers
            if (w.m_uniforms[i])
            {
                ctx->bindBuffer(w.m_uniforms[i]->m_buffer, BUFFER_SLOT_ENTITY_UNIFORMS, gfx::e_bindFlags::uniform);
            }

            if (w.m_material[i])
            {
                ctx->bindBuffer(w.m_material[i]->m_buffer, BUFFER_SLOT_ENTITY_MATERIAL, gfx::e_bindFlags::uniform);
            }

            // bind entity textures
            for (u32 t = 0; t < 4; ++t)
                if (w.m_textures[i].m_texture[t])
                    ctx->bindTexture(w.m_textures[i].m_texture[t], w.m_textures[i].m_sampler[t], t, gfx::e_bindFlags::uniform);

            // bind geometry and draw
            ctx->bindBuffer(w.m_geometry[i]->m_vertexBuffer, 0, gfx::e_bindFlags::vertexBuffer);

            if (w.m_geometry[i]->m_indexBuffer)
            {
                ctx->drawIndexed((u32)w.m_geometry[i]->m_numIndices, w.m_geometry[i]->m_indexBuffer, gfx::e_indexType::uint32);
            }
            else
            {
                ctx->draw((u32)w.m_geometry[i]->m_numVertices);
            }
        }
    }

    void Scene::renderLayerDeferred(View* view)
    {
        auto ctx = gfx::getMainContext();
        World& w = *m_world;

        Layer& l = m_layers[view->m_linkID];

        ModelData md = m_ndcQuadRef.getTypedResource();
        gfx::Geometry geo = md.m_meshes[0];

        gfx::Sampler sampler = m_renderer->getSampler(hash(std::string("clamp_linear")));
        gfx::Sampler shadowMapSampler = m_renderer->getSampler(hash(std::string("shadow_map")));

        // do deferred render for all lights in the scene
        // TODO: batch lights by type
        for (size_t i = 0, n = w.size(); i < n; ++i)
        {
            if (!(w.m_components[i] & e_cmp::light))
                continue;
            
            if (w.m_layer[i] != view->m_linkID)
                continue;

            Light* light = w.m_light[i];
            
            ID variation = 0;
            switch (light->m_type)
            {
            case e_lightType::directional:
                variation = hash(std::string("directional"));
                break;
            case e_lightType::point:
                variation = hash(std::string("point"));
                break;
            case e_lightType::spot:
                variation = hash(std::string("spot"));
                break;
            }
            
            if(view->m_pipelineVariation != 0)
                variation = variation ^ view->m_pipelineVariation;

            ctx->bindRenderPipeline(m_renderer->getPipelineForPass(hash(std::string("deferred")), view->m_pass, variation));

            // bind per light shadow map
            if (light->m_shadowMap)
                ctx->bindTexture(light->m_shadowMap, shadowMapSampler, 4, gfx::e_bindFlags::uniform);

            ctx->bindBuffer(light->m_renderBuffer, 3, gfx::e_bindFlags::uniform);
            ctx->bindBuffer(geo.m_vertexBuffer, 0, gfx::e_bindFlags::vertexBuffer);
            ctx->drawIndexed(geo.m_numIndices, geo.m_indexBuffer, gfx::e_indexType::uint32);
        }
        
        // render ambient light with max blending
        if (l.m_ambientLightUniformBuffer)
        {
            ctx->bindRenderPipeline(m_renderer->getPipelineForPass(hash(std::string("deferred_ambient")), view->m_pass, view->m_pipelineVariation));
            ctx->bindBuffer(l.m_ambientLightUniformBuffer, 3, gfx::e_bindFlags::uniform);
            ctx->bindBuffer(geo.m_vertexBuffer, 0, gfx::e_bindFlags::vertexBuffer);
            ctx->drawIndexed(geo.m_numIndices, geo.m_indexBuffer, gfx::e_indexType::uint32);
        }
    }

    void Scene::renderLayerPostProcess(View* view)
    {
        auto ctx = gfx::getMainContext();

        Layer& l = m_layers[view->m_linkID];

        ModelData md = m_ndcQuadRef.getTypedResource();
        gfx::Geometry geo = md.m_meshes[0];
        ctx->bindBuffer(geo.m_vertexBuffer, 0, gfx::e_bindFlags::vertexBuffer);
        ctx->drawIndexed(geo.m_numIndices, geo.m_indexBuffer, gfx::e_indexType::uint32);
    }

    void Scene::setupLayerViews(std::vector<View*>& views)
    {
        World& w = *m_world;
        auto ctx = gfx::getMainContext();

        u32 samples = m_renderer->getMSAASamples();

        for (auto& lid : m_layerOrder)
        {
            Layer& l = m_layers[lid];
            
            bool separateLayerRender = (!l.m_preBlendPostProcessViews.empty()) || l.m_blendID != 0 || samples > 1;

            // create the views for rendering shadow maps
            std::vector<View*> shadowmapViews;
            for (size_t i = 0, n = w.size(); i < n; ++i)
            {
                if (!(w.m_components[i] & e_cmp::light))
                    continue;

                if (w.m_layer[i] != lid)
                    continue;

                LightType type = w.m_light[i]->m_type;

                View* view = memNew<View>();
                if (type == e_lightType::spot || type == e_lightType::directional)
                {
                    view->m_name = "layer_shadow_map";
                    view->m_linkID = w.m_ulid[i];
                    view->m_depthTarget.m_texture = w.m_light[i]->m_shadowMap;
                    *view = m_renderer->setupView(*view);
                }
                else
                {
                    view->m_name = "layer_shadow_map_point";
                    view->m_linkID = w.m_ulid[i];
                    view->m_depthTarget.m_texture = w.m_light[i]->m_shadowMap;
                    *view = m_renderer->setupView(*view);
                }
                views.push_back(view);
                shadowmapViews.push_back(view);
            }

            if (l.m_deferredRender)
            {
                View* view = memNew<View>();
                view->m_name = "layer_gbuffer";
                *view = m_renderer->setupView(*view);
                view->m_linkID = lid;
                views.push_back(view);

                View* dview = memNew<View>();
                dview->m_name = "layer_deferred";

                // resolve now if there are no pre blend post processes
                dview->m_resolve = (samples > 1 && l.m_preBlendPostProcessViews.empty()) ? e_resolveType::after : e_resolveType::none;
                *dview = m_renderer->setupView(*dview);
                dview->m_linkID = lid;
                dview->m_pipelineVariation = (samples > 1) ? hash<std::string>("MSAA") : 0;

                // deferred render view relies on any shadow map views being completed beforehand
                for (auto& sv : shadowmapViews)
                {
                    dview->m_dependancies.insert(sv);
                }

                // create the inverse camera view projection matrix buffer
                gfx::BufferDesc bd = {};
                bd.m_access = gfx::e_accessFlags::write;
                bd.m_usage = gfx::e_bufferUsageFlags::uniform;
                bd.m_data.m_size = sizeof(LayerCameraData);
                dview->m_uniformBuffer = ctx->createBuffer(bd);
                views.push_back(dview);

                // flag that we need to blend the layer output back into the main buffer
                separateLayerRender = true;
            }
            else
            {
                View* view = memNew<View>();
                view->m_name = (separateLayerRender) ? "layer_separated" : "layer";

                // resolve now if there are no pre blend post processes
                view->m_resolve = (samples > 1 && l.m_preBlendPostProcessViews.empty()) ? e_resolveType::after : e_resolveType::none;
                *view = m_renderer->setupView(*view);
                view->m_linkID = lid;
                views.push_back(view);
            }

            // apply post processing on only this layers contents
            for (size_t i = 0; i < l.m_preBlendPostProcessViews.size(); ++i)
            {
                View& v = l.m_preBlendPostProcessViews[i];

                View* view = memNew<View>();
                *view = v;

                // resolve after completing last post process for a layer
                view->m_resolve = (samples > 1 && i == l.m_preBlendPostProcessViews.size() - 1) ? e_resolveType::after : e_resolveType::none;;
                *view = m_renderer->setupView(*view);
                view->m_linkID = lid;
                views.push_back(view);
            }

            // blend layer into main output
            if (separateLayerRender)
            {
                View* view = memNew<View>();
                view->m_name = "layer_blend";
                view->m_linkID = lid;
                view->m_pipeline = (l.m_blendID != 0) ? l.m_blendID : hash<std::string>("ndc_tex_alpha");
                *view = m_renderer->setupView(*view);
                views.push_back(view);
            }

            // apply post processing to the whole output
            for (auto& v : l.m_postBlendPostProcessViews)
            {
                View* view = memNew<View>();
                *view = v;
                *view = m_renderer->setupView(*view);
                view->m_linkID = lid;
                views.push_back(view);
            }
        }
    }

    bool Scene::isLoaded() const
    {
        return m_flags.load() & e_sceneFlags::loaded;
    }

    bool Scene::isActive() const
    {
        return m_flags.load() & e_sceneFlags::activated;
    }

    bool Scene::shouldTransition() const
    {
        return m_flags.load() & e_sceneFlags::transition;
    }
}
