#include "world.h"

#include "resources.h"
#include "gfx.h"

namespace engine
{
    void GenericComponentArray::grow(size_t size)
    {
        ASSERT(size > m_capacity);
        void* newData = memAlloc(size * m_elementSize);
        memset(newData, 0, size * m_elementSize);
        if (m_data)
        {
            memcpy(newData, m_data, m_capacity * m_elementSize);
            free(m_data);
        }
        m_capacity = size;
        m_data = newData;
    }

    void GenericComponentArray::destroy()
    {
        memFree(m_data);
        m_data = nullptr;
        m_capacity = 0;
    }

    void WorldBase::grow(size_t size)
    {
        // loop over all component arrays to grow
        GenericComponentArray* arr = &m_components;
        for (;;)
        {
            arr->grow(size);
            if (arr == m_last)
                break;
            arr = (GenericComponentArray*)(((intptr_t)arr) + sizeof(GenericComponentArray));
        }
        m_capacity = size;
    }

    size_t WorldBase::newEntity()
    {
        size_t newEntity = -1;

        // find unallocated in existing capacity
        for (size_t i = 0; i < m_capacity; ++i)
        {
            if (m_entityFlags[i] & e_baseEntityFlags::allocated)
                continue;

            newEntity = i;
            break;
        }

        // if no unallocated entity was found grow world
        if (newEntity == (size_t)-1)
        {
            newEntity = m_capacity;
            grow(m_capacity * 2);
        }

        m_entityFlags[newEntity] |= e_baseEntityFlags::allocated;
        m_parent[newEntity] = (size_t)-1;

        m_size = std::max(m_size, newEntity + 1);

        return newEntity;
    }

    void World::destroyEntity(size_t e)
    {
        auto ctx = gfx::getMainContext();
        ResourceManager* rm = ResourceManager::get();

        m_entityFlags[e] = 0;

        if (m_components[e] & e_cmp::textureRef)
        {
            rm->unload(m_textureRef[e]);
        }

        if (m_components[e] & e_cmp::modelRef)
        {
            rm->unload(m_modelRef[e]);
        }

        if (m_components[e] & e_cmp::soundRef)
        {
            rm->unload(m_soundRef[e]);
        }

        if (m_components[e] & e_cmp::fontRef)
        {
            rm->unload(m_fontRef[e]);
        }

        if (m_components[e] & e_cmp::light)
        {
            Light* light = m_light[e];
            
            if (light->m_renderBuffer)
                ctx->destroyBuffer(light->m_renderBuffer);

            if (light->m_shadowMapBuffer)
                ctx->destroyBuffer(light->m_shadowMapBuffer);

            if (light->m_shadowMap)
                ctx->destroyTexture(light->m_shadowMap);

            memDelete(light);
        }

        if (m_uniforms[e])
        {
            if (m_uniforms[e]->m_buffer)
                ctx->destroyBuffer(m_uniforms[e]->m_buffer);
            memDelete(m_uniforms[e]);
        }

        if (m_geometry[e])
        {
            memDelete(m_geometry[e]);
        }

        if (m_material[e])
        {
            memDelete(m_material[e]);
        }

        if (m_camera[e])
        {
            memDelete(m_camera[e]);
        }
        
        if (m_text[e])
        {
            memDelete(m_text[e]);
        }

        m_components[e] = 0;
    }

    u64 World::enumNameToComponentFlagValue(const std::string& name)
    {
        return e_cmp::nameToValue(name);
    }

    u64 World::enumNameToEntityFlagValue(const std::string& name)
    {
        return e_baseEntityFlags::nameToValue(name) | e_entityFlags::nameToValue(name);
    }

    void WorldBase::clear()
    {
        // destroy all entities
        for (size_t i = 0, n = size(); i < n; ++i)
            destroyEntity(i);

        // destroy component arrays
        GenericComponentArray* arr = (GenericComponentArray*)&m_components;
        for (;;)
        {
            arr->destroy();
            if (arr == m_last)
                break;
            arr = (GenericComponentArray*)(((intptr_t)arr) + sizeof(GenericComponentArray));
        }
        m_capacity = 0;
        m_size = 0;
    }
}
