#include <regex>
#include <iostream>
#include <map>
#include <fstream>
#include <filesystem>
#define RAPIDJSON_HAS_STDSTRING 1
#include "rapidjson\document.h"
#include "rapidjson\writer.h"
#include "rapidjson\prettywriter.h"
#include "rapidjson\stringbuffer.h"
#include "pystring.h"

bool mergeJson(rapidjson::Value& dst, const rapidjson::Value& src, rapidjson::Document::AllocatorType& allocator)
{
    for (auto srcIt = src.MemberBegin(); srcIt != src.MemberEnd(); ++srcIt)
    {
        auto dstIt = dst.FindMember(srcIt->name);
        if (dstIt == dst.MemberEnd())
        {
            rapidjson::Value dstName;
            dstName.CopyFrom(srcIt->name, allocator);
            rapidjson::Value dstValue;
            dstValue.CopyFrom(srcIt->value, allocator);

            dst.AddMember(dstName, dstValue, allocator);
        }
        else
        {
            auto srcType = srcIt->value.GetType();
            auto dstType = dstIt->value.GetType();
            if (srcType != dstType)
                return false;

            if (srcIt->value.IsArray())
            {
                for (auto arrayIt = srcIt->value.Begin(); arrayIt != srcIt->value.End(); ++arrayIt)
                {
                    rapidjson::Value dstValue;
                    dstValue.CopyFrom(*arrayIt, allocator);
                    dstIt->value.PushBack(dstValue, allocator);
                }
            }
            else if (srcIt->value.IsObject())
            {
                if (!mergeJson(dstIt->value, srcIt->value, allocator))
                    return false;
            }
            else
            {
                dstIt->value.CopyFrom(srcIt->value, allocator);
            }
        }
    }
    return true;
}

void readFileToBuffer(const char* path, void*& buf, size_t& len)
{
    std::ifstream file(path, std::ios::ate | std::ios::binary);
    if (!file.is_open())
        return;
    len = (size_t)file.tellg();
    char* c = new char[len];
    file.seekg(0);
    file.read(c, len);
    file.close();
    buf = (void*)c;
}

rapidjson::Document readJsonFromFile(const char* filename)
{
    void* buf;
    size_t len;
    readFileToBuffer(filename, buf, len);

    rapidjson::Document d;
    bool result = d.Parse<rapidjson::kParseStopWhenDoneFlag>((const char*)buf).HasParseError();
    if (result)
    {
        auto error = d.GetParseError();
        auto offset = d.GetErrorOffset();
        std::cout << "Json error " << error << " at " << offset << " in file " << filename << "\n";
    }

    delete[] buf;
    return d;
}

void writeBufferToFile(const char* filename, const char* buf, size_t len)
{
    std::ofstream file(filename, std::ios::out);
    if (!file.is_open())
    {
        std::cout << "Failed to open file " << filename << " for write!\n";
        return;
    }

    file.write(buf, len);
    file.close();
}

void writeJsonToFile(const char* filename, rapidjson::Document& d, const std::vector<std::string>& replacements)
{
    rapidjson::StringBuffer buffer;
    rapidjson::PrettyWriter<rapidjson::StringBuffer> writer(buffer);
    d.Accept(writer);

    std::string outString = buffer.GetString();
    for (auto& s : replacements)
        outString = pystring::replace(outString, s, "");

    writeBufferToFile(filename, outString.c_str(), outString.size());
}

std::string findShaderPath(const std::string& basename, const std::vector<std::string>& sources)
{
    // check for file in possible sources
    for (auto& s : sources)
    {
        std::string path = pystring::os::path::join(s, basename);

        std::ifstream file(path, std::ios::ate | std::ios::binary);
        if (!file.is_open())
            continue;
        file.close();
        return path;
    }
    return basename;
}

rapidjson::Document loadRenderConfig(const char* path)
{
    rapidjson::Document base = readJsonFromFile(path);

    if (base.HasMember("merge"))
    {
        for (size_t i = 0; i < base["merge"].Size(); ++i)
        {
            const char* mergePath = base["merge"][i].GetString();
            rapidjson::Document merge = loadRenderConfig(mergePath);
            bool merged = mergeJson(base, merge, base.GetAllocator());
            if (!merged)
            {
                std::cout << "Failed to merge config files \"" << path << "\" and \"" << mergePath << "\"!\n";
            }
        }
    }
    return base;
}

void buildRenderConfig(const rapidjson::Value& cfg)
{
    // copies render config to correct directory
    // also goes through pipelines to compile combined shaders

    std::string cfgSrc = pystring::os::path::normpath(cfg["source"].GetString());
    std::string cfgDest = pystring::os::path::normpath(cfg["destination"].GetString());
    std::string shaderDest = pystring::os::path::normpath(cfg["shader_destination"].GetString());
    std::string glslang = pystring::os::path::normpath(cfg["glslang"].GetString());
    std::string shaderTool = pystring::os::path::normpath(cfg["shader_tool"].GetString());

    std::vector<std::string> shaderSources;
    for (int i = 0; i < cfg["shader_sources"].Size(); ++i)
    {
        shaderSources.push_back(pystring::os::path::normpath(cfg["shader_sources"][i].GetString()));
    }


    // merge in other requested cfgs
    rapidjson::Document d = loadRenderConfig(cfgSrc.c_str());
        

    std::string workingDir = std::filesystem::current_path().string();

    // compile pipelines as combined shaders
    if (d.FindMember("pipelines") != d.MemberEnd())
    {
        std::string shaderOutputDir = pystring::os::path::join(workingDir, shaderDest);
        std::filesystem::create_directories(shaderOutputDir);

        for (auto it = d["pipelines"].MemberBegin(); it != d["pipelines"].MemberEnd(); ++it)
        {
            std::string pipelineName = it->name.GetString();

            const rapidjson::Value& v = it->value;

            const char* k_shaderNames[] = 
            {
                "vertex_shader",
                "tessellation_control_shader",
                "tessellation_evaluation_shader",
                "geometry_shader",
                "fragment_shader",
                "compute_shader"
            };

            const char* k_shaderFileEndings[] =
            {
                ".vert",
                ".tesc",
                ".tese",
                ".geom",
                ".frag",
                ".compute"
            };

            std::vector<std::string> shaderPaths;
            for (size_t j = 0; j < 6; ++j)
            {
                if (v.HasMember(k_shaderNames[j]))
                {
                    std::string path = v[k_shaderNames[j]].GetString();
                    path += k_shaderFileEndings[j];
                    path = findShaderPath(path, shaderSources);
                    shaderPaths.push_back(path);
                }
            }

            // create string of comma separated shader paths
            std::string pathString = "-i ";
            for (size_t i = 0; i < shaderPaths.size(); ++i)
            {
                pathString += shaderPaths[i];
                if (i != shaderPaths.size() - 1)
                    pathString += ",";
            }

            // create path for shader output
            std::string outString = pipelineName + ".shader";
            outString = pystring::os::path::join(shaderDest, outString);
            outString = "-o " + outString;

            // compile combined shader
            std::string cmd = shaderTool + " ";
            cmd += pathString + " ";
            cmd += outString + " ";
            cmd += "-glslang " + glslang;
            int res = system(cmd.c_str());
        }
    }

    // write cfg file to dest folder
    std::string cfgOutputDir = pystring::os::path::join(workingDir, pystring::os::path::dirname(cfgDest));
    std::filesystem::create_directories(cfgOutputDir);
    writeJsonToFile(cfgDest.c_str(), d, {});
}

void buildShaders(const rapidjson::Value& cfg)
{
    // compiles shaders individually
    std::string destination = pystring::os::path::normpath(cfg["destination"].GetString());
    std::string glslang = pystring::os::path::normpath(cfg["glslang"].GetString());
    std::string shaderTool = pystring::os::path::normpath(cfg["shader_tool"].GetString());

    std::string workingDir = std::filesystem::current_path().string();
    std::string outputDir = pystring::os::path::join(workingDir, destination);
    std::filesystem::create_directories(outputDir);

    for (int i = 0; i < cfg["sources"].Size(); ++i)
    {
        std::string source = pystring::os::path::normpath(cfg["sources"][i].GetString());

        std::string cmd = shaderTool + " ";
        cmd += "-i " + source + " ";
        cmd += "-o " + destination + " ";
        cmd += "-glslang " + glslang;
        int res = system(cmd.c_str());
    }
}

void buildPremake(const rapidjson::Value& cfg)
{
    // runs premake command to build project with any args
    std::string premake = pystring::os::path::normpath(cfg["premake"].GetString());
    std::string ide = cfg["ide"].GetString();
    std::string platform = cfg["platform"].GetString();
    std::string gfx = cfg["gfx"].GetString();

    std::string cmd = premake;
    cmd += " " + ide;
    cmd += " --platform=" + platform;
    cmd += " --gfxapi=" + gfx;
    int res = system(cmd.c_str());
}

rapidjson::Value parseTriggerVarString(const std::string& str, rapidjson::MemoryPoolAllocator<rapidjson::CrtAllocator> alloc, bool parsingArithmetic)
{
    rapidjson::Value out;
    out.SetObject();

    if (pystring::startswith(str, "$(") || (parsingArithmetic && pystring::startswith(str, "(")))
    {
        out.AddMember("type", "arithmetic", alloc);

        std::string inner = (pystring::startswith(str, "$(")) ? str.substr(2, str.size() - 3) : str.substr(1, str.size() - 2);

        char arithmeticSymbols[4] =
        {
            '+',
            '-',
            '*',
            '/'
        };

        size_t symbol = 0;
        size_t depth = 0;
        for (size_t i = 0; i < inner.size() - 1; ++i)
        {
            if (inner[i] == '(')
            {
                depth++;
                continue;
            }

            if (inner[i] == ')')
            {
                depth--;
                continue;
            }

            if (depth > 0)
                continue;

            for (size_t j = 0; j < 4; ++j)
            {
                if (inner[i] != arithmeticSymbols[j])
                    continue;

                if (inner[i - 1] == ' ' && inner[i + 1] == ' ')
                {
                    symbol = j;
                    break;
                }
            }
        }

        std::string operation = "";
        operation += arithmeticSymbols[symbol];

        out.AddMember("operation", operation, alloc);


        std::vector<std::string> leftRight;
        pystring::split(inner, leftRight, " " + operation + " ", 1);

        leftRight[0] = pystring::strip(leftRight[0]);
        leftRight[1] = pystring::strip(leftRight[1]);

        rapidjson::Value leftValue = parseTriggerVarString(leftRight[0], alloc, true);
        rapidjson::Value rightValue = parseTriggerVarString(leftRight[1], alloc, true);

        out.AddMember("left", leftValue, alloc);
        out.AddMember("right", rightValue, alloc);
    }
    else if (pystring::startswith(str, "WORLD/") || pystring::startswith(str, "GAME/"))
    {
        std::string sub = str;
        if (pystring::startswith(str, "GAME/"))
        {
            out.AddMember("type", "state", alloc);
            sub = pystring::replace(sub, "GAME/", "");
        }
        else
        {
            out.AddMember("type", "entity", alloc);
            sub = pystring::replace(sub, "WORLD/", "");
        }

        int indexStart = pystring::find(sub, "[");
        int indexEnd = pystring::find(sub, "]");
        if (indexStart != -1 && indexEnd != -1)
        {
            int index = std::stoi(sub.substr(indexStart + 1, indexEnd - (indexStart + 1)));
            sub = pystring::replace(sub, sub.substr(indexStart, (indexEnd - indexStart) + 1), "");
            out.AddMember("index", index, alloc);
        }
        out.AddMember("value", sub, alloc);
    }
    else if (pystring::startswith(str, "GAME/"))
    {
        out.AddMember("type", "state", alloc);

        std::string stateSub = pystring::replace(str, "GAME/", "");
        out.AddMember("value", stateSub, alloc);
    }
    else
    {
        out.AddMember("type", "raw", alloc);

        std::string val = str;
        if (parsingArithmetic)
            val = "REMOVE_ME" + str + "REMOVE_ME";
        out.AddMember("value", val, alloc);
    }

    return out;
}

void processTrigger(rapidjson::Value& trigger, rapidjson::MemoryPoolAllocator<rapidjson::CrtAllocator> alloc)
{
    const char* triggerCheckNames[3] =
    {
        "conditions",
        "actions",
        "else_actions"
    };

    for (int j = 0; j < 3; ++j)
    {
        if (trigger.HasMember(triggerCheckNames[j]))
        {
            rapidjson::Value& json = trigger[triggerCheckNames[j]];

            for (size_t k = 0; k < json.Size(); ++k)
            {
                for (auto mit = json[k].MemberBegin(); mit != json[k].MemberEnd(); ++mit)
                {
                    std::string name = mit->name.GetString();

                    if(mit->value.IsString())
                    {
                        rapidjson::Value newValue;
                        newValue.SetObject();

                        std::string val = mit->value.GetString();
                        newValue = parseTriggerVarString(val, alloc, false);

                        if (newValue["type"].GetString() != "raw")
                        {
                            mit->value.CopyFrom(newValue, alloc);
                        }
                    }


                }

            }
        }
    }
}

std::vector<std::pair<std::string, std::string>> getFiles(const rapidjson::Value& cfg, const std::string& workingPath)
{
    std::vector<std::pair<std::string, std::string>> files;

    for (auto& pairs : cfg["files"].GetArray())
    {
        auto& pair = pairs.GetArray();
        std::string src = pystring::os::path::normpath(pair[0].GetString());
        std::string dst = pystring::os::path::normpath(pair[1].GetString());

        if (std::filesystem::is_directory(src))
        {
            std::string absoluteInputPath = pystring::os::path::join(workingPath, src);
            
            std::vector<std::string> directories = { absoluteInputPath };
            while (!directories.empty())
            {
                std::string dir = directories.front();
                std::string relative = pystring::replace(dir, absoluteInputPath, "");
                directories.erase(directories.begin());
                for (auto& p : std::filesystem::directory_iterator(dir))
                {
                    if (p.is_directory())
                    {
                        directories.push_back(p.path().string());
                    }
                    else
                    {
                        files.push_back({ p.path().string(), dst + relative});
                    }
                }
            }
        }
        else
        {
            files.push_back({ src, dst });
        }
    }

    return files;
}

void buildScenes(const rapidjson::Value& cfg)
{
    bool clean = false;
    if (cfg.FindMember("clean") != cfg.MemberEnd())
        clean = cfg["clean"].GetBool();

    std::string workingPath = std::filesystem::current_path().string();

    std::vector<std::pair<std::string, std::string>> files = getFiles(cfg, workingPath);

    for (auto& pair : files)
    {
        std::string src = pair.first;
        std::string dst = pair.second;
        if (!pystring::endswith(src, ".scene"))
        {
            continue;
        }

        // create directory
        std::string absoluteDestDir = pystring::os::path::join(workingPath, dst);
        if (pystring::find(absoluteDestDir, ".") != -1)
        {
            absoluteDestDir = pystring::os::path::dirname(absoluteDestDir);
        }
        std::filesystem::create_directories(absoluteDestDir);

        std::string destPath = dst;
        if (std::filesystem::is_directory(dst))
        {
            destPath = pystring::os::path::basename(src);
            destPath = pystring::os::path::join(dst, destPath);
        }

        std::string absoluteDestPath = pystring::os::path::join(workingPath, destPath);

        std::cout << "Building " << src << " to " << absoluteDestPath << "\n";

        rapidjson::Document d = readJsonFromFile(src.c_str());
        if (d.HasParseError())
            continue;

        auto alloc = d.GetAllocator();

        if (d.HasMember("entities"))
        {
            rapidjson::Value& entitiesArray = d["entities"];
            for (size_t i = 0; i < entitiesArray.Size(); ++i)
            {
                rapidjson::Value& entity = entitiesArray[i];
                if (!entity.HasMember("triggers"))
                    continue;

                rapidjson::Value& triggers = entity["triggers"];
                for (auto tit = triggers.MemberBegin(); tit != triggers.MemberEnd(); ++tit)
                {
                    processTrigger(tit->value, alloc);
                }
            }
        }

        if (d.HasMember("triggers"))
        {
            for (size_t i = 0; i < d["triggers"].Size(); ++i)
            {
                processTrigger(d["triggers"][i], alloc);
            }
        }

        writeJsonToFile(absoluteDestPath.c_str(), d, { "\"REMOVE_ME", "REMOVE_ME\"" });
    }
}

void buildPrefabs(const rapidjson::Value& cfg)
{
    bool clean = false;
    if (cfg.FindMember("clean") != cfg.MemberEnd())
        clean = cfg["clean"].GetBool();

    std::string workingPath = std::filesystem::current_path().string();

    std::vector<std::pair<std::string, std::string>> files = getFiles(cfg, workingPath);

    for (auto& pair : files)
    {
        std::string src = pair.first;
        std::string dst = pair.second;
        if (!pystring::endswith(src, ".prefab"))
        {
            continue;
        }

        // create directory
        std::string absoluteDestDir = pystring::os::path::join(workingPath, dst);
        if (pystring::find(absoluteDestDir, ".") != -1)
        {
            absoluteDestDir = pystring::os::path::dirname(absoluteDestDir);
        }
        std::filesystem::create_directories(absoluteDestDir);

        std::string destPath = dst;
        if (std::filesystem::is_directory(dst))
        {
            destPath = pystring::os::path::basename(src);
            destPath = pystring::os::path::join(dst, destPath);
        }

        std::string absoluteDestPath = pystring::os::path::join(workingPath, destPath);

        std::cout << "Building " << src << " to " << absoluteDestPath << "\n";

        rapidjson::Document d = readJsonFromFile(src.c_str());
        if (d.HasParseError())
            continue;

        auto alloc = d.GetAllocator();

        if (d.HasMember("entities"))
        {
            for (auto mit = d["entities"].MemberBegin(); mit != d["entities"].MemberEnd(); ++mit)
            {
                rapidjson::Value& entity = mit->value;
                if (!entity.HasMember("triggers"))
                    continue;

                rapidjson::Value& triggers = entity["triggers"];
                for (auto tit = triggers.MemberBegin(); tit != triggers.MemberEnd(); ++tit)
                {
                    processTrigger(tit->value, alloc);
                }
            }
        }

        writeJsonToFile(absoluteDestPath.c_str(), d, { "\"REMOVE_ME", "REMOVE_ME\"" });
    }
}

namespace e_codegenMode
{
    enum codegenMode
    {
        none,
        enumType,
        enumFlag
    };
}
typedef e_codegenMode::codegenMode CodegenMode;

void buildCodegen(const rapidjson::Value& cfg)
{

    std::string workingPath = std::filesystem::current_path().string();

    std::vector<std::pair<std::string, std::string>> files = getFiles(cfg, workingPath);

    for (auto& pair : files)
    {
        std::string src = pair.first;
        std::string dst = pair.second;
        if (!pystring::endswith(src, ".h"))
        {
            continue;
        }

        // create directory
        std::string absoluteDestDir = pystring::os::path::join(workingPath, dst);
        std::string hPath = pystring::os::path::join({ dst, "include", "codegen", pystring::os::path::basename(src) });
        std::string cppPath = pystring::os::path::join({ dst, "source", "codegen", pystring::replace(pystring::os::path::basename(src), ".h", ".cpp")});

        hPath = pystring::os::path::normpath(pystring::os::path::join(workingPath, hPath));
        cppPath = pystring::os::path::normpath(pystring::os::path::join(workingPath, cppPath));
        std::filesystem::create_directories(pystring::os::path::dirname(hPath));
        std::filesystem::create_directories(pystring::os::path::dirname(cppPath));

        void* buf = nullptr;
        size_t size = 0;

        readFileToBuffer(src.c_str(), (void*)buf, size);
        if (!buf)
            continue;

        std::string str = std::string((char*)buf, size);

        delete[] buf;

        std::cout << "Building " << pystring::os::path::normpath(src) << " to " << hPath << " and " << cppPath << "\n";

        std::vector<std::string> lines;
        pystring::splitlines(str, lines);

        CodegenMode currentMode = e_codegenMode::none;

        std::string enumName = "";
        std::vector<std::string> enumNames;
        std::vector<int> enumValues;
        int resolveIndent = -1;
        int currentIndent = 0;
        
        std::string hOut = "// file is generated code, do not edit\n#pragma once\n#include \"defs.h\"\n#include <string>\n";
        std::string cppOut = "// file is generated code, do not edit\n#include \"json.h\"\n#include \"codegen/" + pystring::os::path::basename(src) + "\"\n";
        bool foundMasterNamespace = false;

        for (auto line : lines)
        {
            std::string stripped = pystring::strip(line);

            if (!foundMasterNamespace)
            {
                if (pystring::startswith(stripped, "namespace"))
                {
                    std::string masterNamespace = pystring::replace(stripped, "namespace ", "");
                    if (masterNamespace != "engine")
                        cppOut = cppOut + "using namespace engine;\n";
                    cppOut = cppOut + "namespace " + masterNamespace + "\n{\n";
                    foundMasterNamespace = true;
                }
            }

            bool finishEnum = false;

            switch (currentMode)
            {
            case e_codegenMode::none:
            {
                if (pystring::startswith(stripped, "enum "))
                {
                    if (pystring::endswith(stripped, " : u64"))
                    {
                        currentMode = e_codegenMode::enumFlag;
                    }
                    else
                    {
                        currentMode = e_codegenMode::enumType;
                    }

                    std::vector<std::string> splits;
                    pystring::split(stripped, splits);
                    enumName = splits[1];

                    resolveIndent = currentIndent;
                }
            }
            break;
            case e_codegenMode::enumType:
            case e_codegenMode::enumFlag:
            {
                if (!stripped.empty())
                {
                    int equalsPos = pystring::find(stripped, "=");
                    if (equalsPos != -1)
                    {
                        std::string name = pystring::strip(stripped.substr(0, equalsPos));
                        std::string def = pystring::strip(pystring::strip(stripped.substr(equalsPos + 1, stripped.size() - (equalsPos + 1))), ",");

                        enumNames.push_back(name);

                        if (currentMode == e_codegenMode::enumFlag)
                        {
                            size_t off = line.find_last_of('<');
                            if (off != (size_t)-1)
                            {
                                std::string value = pystring::strip(line.substr(off + 1, line.size() - (off + 1)));
                                int val = std::stoi(value);
                                enumValues.push_back(val);

                                // insert necessary cast if shifting more than 31 bits
                                if (val >= 32 && pystring::find(line, "(u64)") == -1)
                                    line.insert(off + 2, "(u64)");
                            }
                            else
                            {
                                enumValues.push_back(-1);
                            }
                        }
                        else
                        {
                            int val = std::stoi(def);
                            enumValues.push_back(val);
                        }
                    }
                    else
                    {
                        if (stripped[0] != '{' && stripped[0] != '}')
                        {
                            enumNames.push_back(pystring::strip(stripped, ","));
                            enumValues.push_back(enumValues.size());
                        }
                    }
                }
            }
            break;
            }

            hOut = hOut + line + "\n";

            if (stripped.size() > 0)
            {
                if (stripped[0] == '{')
                {
                    currentIndent++;
                }
                else if (stripped[0] == '}')
                {
                    currentIndent--;
                    if (currentIndent == resolveIndent)
                    {
                        switch (currentMode)
                        {
                        case e_codegenMode::enumFlag:
                        case e_codegenMode::enumType:
                        {
                            if (enumNames.size() > 0)
                            {
                                std::string lineIndent = "";
                                for (size_t i = 0; i < currentIndent; ++i)
                                {
                                    lineIndent += "\t";
                                }

                                // add definition to header for name to value conversion
                                if (currentMode == e_codegenMode::enumFlag)
                                {
                                    hOut = hOut + lineIndent + "u64 nameToValue(const std::string& name);\n";
                                }
                                else
                                {
                                    std::string typeName = "e_" + enumName + "::" + enumName;
                                    hOut = hOut + lineIndent + typeName + " nameToValue(const std::string& name);\n";
                                    
                                }
                                // add definition to cpp for name to value conversion
                                std::string namesArrayName = "k_" + enumName + "Names";
                                cppOut = cppOut +  + "\tconst char* " + namesArrayName + "[] = \n";
                                cppOut = cppOut + "\t{\n";

                                int currentValue = 0;
                                for (size_t i = 0; i < enumNames.size(); ++i)
                                {
                                    std::string name = enumNames[i];
                                    int value = enumValues[i];

                                    if (value == -1)
                                        continue;

                                    while (currentValue++ < value)
                                    {
                                        cppOut = cppOut + "\t\t\"#UNUSED#\",\n";
                                    }
                                    cppOut = cppOut + "\t\t\"" + name + "\"";
                                    if (i < enumNames.size() - 1)
                                    {
                                        cppOut = cppOut + ",";
                                    }
                                    cppOut = cppOut + "\n";
                                }

                                cppOut = cppOut + "\t};\n";

                                cppOut = cppOut + "\tnamespace e_" + enumName + "\n\t{\n";
                                if (currentMode == e_codegenMode::enumFlag)
                                {
                                    cppOut = cppOut + "\t\tu64 nameToValue(const std::string& name)\n\t\t{\n";
                                    cppOut = cppOut + "\t\t\t return enumNameToFlagValue(" + namesArrayName + ", ARRAY_SIZE(" + namesArrayName + "), name);\n";
                                }
                                else
                                {
                                    std::string typeName = "e_" + enumName + "::" + enumName;
                                    cppOut = cppOut + "\t\t" + typeName + " nameToValue(const std::string& name)\n\t\t{\n";
                                    cppOut = cppOut + "\t\t\t return enumNameToValue<" + typeName + ">(" + namesArrayName + ", ARRAY_SIZE(" + namesArrayName + "), name); \n";
                                }
                                cppOut = cppOut + "\t\t}\n\t}\n";
                            }
                            enumNames.clear();
                            enumValues.clear();
                            enumName = "";                            
                        }
                        break;
                        default:
                            break;
                        }
                        currentMode = e_codegenMode::none;
                        resolveIndent = -1;
                    }
                }
            }
        }

        cppOut = cppOut + "}";

        writeBufferToFile(hPath.c_str(), hOut.c_str(), hOut.size());
        writeBufferToFile(cppPath.c_str(), cppOut.c_str(), cppOut.size());
    }
}

void copyFiles(const rapidjson::Value& cfg)
{
    bool clean = false;
    if (cfg.FindMember("clean") != cfg.MemberEnd())
        clean = cfg["clean"].GetBool();

    std::vector<std::string> exts;
    if (cfg.FindMember("exts") != cfg.MemberEnd())
        for (auto& ext : cfg["exts"].GetArray())
            exts.push_back(ext.GetString());

    bool recursive = false;
    if (cfg.FindMember("recursive") != cfg.MemberEnd())
        recursive = cfg["recursive"].GetBool();

    std::string workingPath = std::filesystem::current_path().string();

    for (auto& pairs : cfg["files"].GetArray())
    {
        auto& pair = pairs.GetArray();
        std::string src = pystring::os::path::normpath(pair[0].GetString());
        std::string dst = pystring::os::path::normpath(pair[1].GetString());

        std::vector<std::string> files;
        if (std::filesystem::is_directory(src))
        {
            // process all the files in a directory into compiled shaders
            std::string absoluteInputPath = pystring::os::path::join(workingPath, src);
            for (auto& p : std::filesystem::directory_iterator(absoluteInputPath))
            {
                std::string path = p.path().string();
                files.push_back(path);
            }
        }
        else
        {
            files.push_back(src);
        }

        for (auto& f : files)
        {
            // check file has valid ext
            bool hasExt = exts.empty();
            for (auto& ext : exts)
            {
                if (pystring::endswith(f, ext))
                {
                    hasExt = true;
                    break;
                }
            }

            if (!hasExt)
                continue;
            
            // create directory
            std::string absoluteDestDir = pystring::os::path::join(workingPath, dst);
            if (pystring::find(absoluteDestDir, ".") != -1)
            {
                absoluteDestDir = pystring::os::path::dirname(absoluteDestDir);
            }
            std::filesystem::create_directories(absoluteDestDir);

            std::string destPath = dst;         
            if (std::filesystem::is_directory(dst))
            {
                destPath = pystring::os::path::basename(f);
                destPath = pystring::os::path::join(dst, destPath);
            }

            std::string absoluteDestPath = pystring::os::path::join(workingPath, destPath);


            std::cout << "Copying " << f << " to " << absoluteDestPath << "\n";
            std::filesystem::copy_options options = std::filesystem::copy_options::update_existing;
            std::filesystem::copy(f, absoluteDestPath, options);
        }
    }
}

typedef void(*BuildStageFunc)(const rapidjson::Value&);

const std::map<std::string, BuildStageFunc> k_buildFuncs =
{
    { "render_config", buildRenderConfig },
    { "shaders", buildShaders },
    { "premake", buildPremake },
    { "copy", copyFiles },
    { "scenes", buildScenes },
    { "prefabs", buildPrefabs },
    { "codegen", buildCodegen }
};

int main(int argc, char *argv[])
{
    std::map<std::string, std::string> args;
    for (int i = 1; i < argc; i+=2)
    {
        std::string arg1 = std::string(argv[i]);
        std::string arg2 = (i + 1 < argc) ? std::string(argv[i + 1]) : "";
        args[arg1] = arg2;
    }

    std::string cfgName = args["-config"];
    if (cfgName.empty())
    {
        std::cout << "No build configuration specified. Use \"-config <platform>\" to specify build configuration.\n";
        return 0;
    }

    rapidjson::Document d = readJsonFromFile("config.json");

    const rapidjson::Value& cfg = d[cfgName.c_str()];

    for (auto it = cfg.MemberBegin(); it != cfg.MemberEnd(); ++it)
    {
        std::string name = it->name.GetString();
        std::cout << "----------\n" << name << "\n----------\n";

        std::string fn = name;
        if (it->value.FindMember("type") != it->value.MemberEnd())
        {
            fn = it->value["type"].GetString();
        }

        if (k_buildFuncs.find(fn) != k_buildFuncs.end())
        {
            BuildStageFunc func = k_buildFuncs.at(fn);
            func(it->value);
        }
    }

    return 0;
}
